!!******************************************************************************
!!
!!  This file is part of the GODUNOV source code, a program to perform
!!  Newtonian or relativistic magnetohydrodynamical simulations.
!!
!!  Copyright (C) 2007-2013 Grzegorz Kowal <grzegorz@amuncode.org>
!!
!!  This program is free software: you can redistribute it and/or modify
!!  it under the terms of the GNU General Public License as published by
!!  the Free Software Foundation, either version 3 of the License, or
!!  (at your option) any later version.
!!
!!  This program is distributed in the hope that it will be useful,
!!  but WITHOUT ANY WARRANTY; without even the implied warranty of
!!  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!!  GNU General Public License for more details.
!!
!!  You should have received a copy of the GNU General Public License
!!  along with this program.  If not, see <http://www.gnu.org/licenses/>.
!!
!!******************************************************************************
!!
!! module: BOUNDARIES
!!
!!  This module handles the boundary synchronization.
!!
!!******************************************************************************
!
module boundaries

#ifdef PROFILE
! include external procedures
!
    use timers, only : set_timer, start_timer, stop_timer
#endif /* PROFILE */

! module variables are not implicit by default
!
  implicit none

#ifdef PROFILE
! timer indices
!
  integer                , save :: ibi, ibn
#endif /* PROFILE */

! module parameters for the boundary update order and boundary type
!
  character(len = 32), save     :: orbndry = "xyz"
  character(len = 32), save     :: xlbndry = "periodic"
  character(len = 32), save     :: xubndry = "periodic"
  character(len = 32), save     :: ylbndry = "periodic"
  character(len = 32), save     :: yubndry = "periodic"
  character(len = 32), save     :: zlbndry = "periodic"
  character(len = 32), save     :: zubndry = "periodic"

! parameters corresponding to the boundary type
!
  integer, parameter            :: bnd_default      = 0
  integer, parameter            :: bnd_periodic     = 0
  integer, parameter            :: bnd_exchange     = 1
  integer, parameter            :: bnd_open         = 2
  integer, parameter            :: bnd_outflow      = 3
  integer, parameter            :: bnd_reflective   = 4
#ifdef BFIELD
  integer, parameter            :: bnd_current_free = 5
#endif /* BFIELD */

! variable to store boundary type flags
!
  integer, dimension(3,2), save :: bnd_type = 0

! by default everything is private
!
  private

! declare public subroutines
!
  public :: initialize_boundaries, update_boundaries
#ifdef BFIELD
  public :: update_magnetic_boundaries
#endif /* BFIELD */

!- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
!
  contains
!
!===============================================================================
!
! subroutine INITIALIZE_BOUNDARIES:
! --------------------------------
!
!   Subroutine initializes the module BOUNDARIES by setting its parameters.
!
!===============================================================================
!
  subroutine initialize_boundaries()

! include external procedures and variables
!
#ifdef MPI
    use mpitools  , only : pdims, pcoords, periodic
#endif /* MPI */
    use parameters, only : get_parameter_string

! local variables are not implicit by default
!
    implicit none
!
!-------------------------------------------------------------------------------
!
#ifdef PROFILE
! set timer descriptions
!
    call set_timer('boundary initialization', ibi)
    call set_timer('boundary update'        , ibn)

! start accounting time for the boundary initialization
!
    call start_timer(ibi)
#endif /* PROFILE */

! get runtime value of the boundary order
!
    call get_parameter_string ("orbndry" , orbndry)

! get runtime values for the boundary types
!
    call get_parameter_string ("xlbndry" , xlbndry)
    call get_parameter_string ("xubndry" , xubndry)
    call get_parameter_string ("ylbndry" , ylbndry)
    call get_parameter_string ("yubndry" , yubndry)
    call get_parameter_string ("zlbndry" , zlbndry)
    call get_parameter_string ("zubndry" , zubndry)

! fill the boundary type flag
!
    select case(xlbndry)
    case("periodic")
      bnd_type(1,1) = bnd_periodic
    case("open")
      bnd_type(1,1) = bnd_open
    case("outflow", "out")
      bnd_type(1,1) = bnd_outflow
    case("reflective", "reflecting", "reflect")
      bnd_type(1,1) = bnd_reflective
#ifdef BFIELD
    case("current-free", "current")
      bnd_type(1,1) = bnd_current_free
#endif /* BFIELD */
    case default
      bnd_type(1,1) = bnd_default
    end select

    select case(xubndry)
    case("periodic")
      bnd_type(1,2) = bnd_periodic
    case("open")
      bnd_type(1,2) = bnd_open
    case("outflow", "out")
      bnd_type(1,2) = bnd_outflow
    case("reflective", "reflecting", "reflect")
      bnd_type(1,2) = bnd_reflective
#ifdef BFIELD
    case("current-free", "current")
      bnd_type(1,2) = bnd_current_free
#endif /* BFIELD */
    case default
      bnd_type(1,2) = bnd_default
    end select

    select case(ylbndry)
    case("periodic")
      bnd_type(2,1) = bnd_periodic
    case("open")
      bnd_type(2,1) = bnd_open
    case("outflow", "out")
      bnd_type(2,1) = bnd_outflow
    case("reflective", "reflecting", "reflect")
      bnd_type(2,1) = bnd_reflective
#ifdef BFIELD
    case("current-free", "current")
      bnd_type(2,1) = bnd_current_free
#endif /* BFIELD */
    case default
      bnd_type(2,1) = bnd_default
    end select

    select case(yubndry)
    case("periodic")
      bnd_type(2,2) = bnd_periodic
    case("open")
      bnd_type(2,2) = bnd_open
    case("outflow", "out")
      bnd_type(2,2) = bnd_outflow
    case("reflective", "reflecting", "reflect")
      bnd_type(2,2) = bnd_reflective
#ifdef BFIELD
    case("current-free", "current")
      bnd_type(2,2) = bnd_current_free
#endif /* BFIELD */
    case default
      bnd_type(2,2) = bnd_default
    end select

    select case(zlbndry)
#ifdef R3D
    case("periodic")
      bnd_type(3,1) = bnd_periodic
    case("open")
      bnd_type(3,1) = bnd_open
    case("outflow", "out")
      bnd_type(3,1) = bnd_outflow
    case("reflective", "reflecting", "reflect")
      bnd_type(3,1) = bnd_reflective
#ifdef BFIELD
    case("current-free", "current")
      bnd_type(3,1) = bnd_current_free
#endif /* BFIELD */
#endif /* R3D */
    case default
      bnd_type(3,1) = bnd_default
    end select

    select case(zubndry)
#ifdef R3D
    case("periodic")
      bnd_type(3,2) = bnd_periodic
    case("open")
      bnd_type(3,2) = bnd_open
    case("outflow", "out")
      bnd_type(3,2) = bnd_outflow
    case("reflective", "reflecting", "reflect")
      bnd_type(3,2) = bnd_reflective
#ifdef BFIELD
    case("current-free", "current")
      bnd_type(3,2) = bnd_current_free
#endif /* BFIELD */
#endif /* R3D */
    case default
      bnd_type(3,2) = bnd_default
    end select


#ifdef MPI
! change the internal boundaries to "exchange" type for the MPI update
!
    if (pdims(1) .gt. 1) then
      if (periodic(1)) then
        xlbndry       = "exchange"
        xubndry       = "exchange"
        bnd_type(1,:) = bnd_exchange
      else
        if (pcoords(1) .gt. 0         ) then
          xlbndry       = "exchange"
          bnd_type(1,1) = bnd_exchange
        end if
        if (pcoords(1) .lt. pdims(1)-1) then
          xubndry       = "exchange"
          bnd_type(1,2) = bnd_exchange
        end if
      end if
    end if

    if (pdims(2) .gt. 1) then
      if (periodic(2)) then
        ylbndry       = "exchange"
        yubndry       = "exchange"
        bnd_type(2,:) = bnd_exchange
      else
        if (pcoords(2) .gt. 0         ) then
          ylbndry       = "exchange"
          bnd_type(2,1) = bnd_exchange
        end if
        if (pcoords(2) .lt. pdims(2)-1) then
          yubndry       = "exchange"
          bnd_type(2,2) = bnd_exchange
        end if
      end if
    end if

    if (pdims(3) .gt. 1) then
      if (periodic(3)) then
        zlbndry       = "exchange"
        zubndry       = "exchange"
        bnd_type(3,:) = bnd_exchange
      else
        if (pcoords(3) .gt. 0         ) then
          zlbndry       = "exchange"
          bnd_type(3,1) = bnd_exchange
        end if
        if (pcoords(3) .lt. pdims(3)-1) then
          zubndry       = "exchange"
          bnd_type(3,2) = bnd_exchange
        end if
      end if
    end if
#endif /* MPI */

#ifdef PROFILE
! stop accounting time for the boundary initialization
!
    call stop_timer(ibi)
#endif /* PROFILE */

!-------------------------------------------------------------------------------
!
  end subroutine initialize_boundaries
!
!===============================================================================
!
! subroutine UPDATE_BOUNDARIES:
! ----------------------------
!
!   Subroutine updates boundaries in selected order.
!
!   Arguments:
!
!     h - a time interval for the time varying boundary conditions
!
!===============================================================================
!
  subroutine update_boundaries(h)

! declare all variables as implicit
!
    implicit none

! subroutine arguments
!
    real, intent(in) :: h
!
!-------------------------------------------------------------------------------
!
#ifdef PROFILE
! start accounting time for the boundary update
!
    call start_timer(ibn)
#endif /* PROFILE */

! depending on the parameter orbndry update boundary in a specific order
!
    select case(orbndry)

    case ("yzx")
      call yboundary(h)
#ifdef R3D
      call zboundary(h)
#endif /* R3D */
      call xboundary(h)

    case ("zxy")
#ifdef R3D
      call zboundary(h)
#endif /* R3D */
      call xboundary(h)
      call yboundary(h)

    case ("yxz")
      call yboundary(h)
      call xboundary(h)
#ifdef R3D
      call zboundary(h)
#endif /* R3D */

    case ("xzy")
      call xboundary(h)
#ifdef R3D
      call zboundary(h)
#endif /* R3D */
      call yboundary(h)

    case ("zyx")
#ifdef R3D
      call zboundary(h)
#endif /* R3D */
      call yboundary(h)
      call xboundary(h)

   case default
      call xboundary(h)
      call yboundary(h)
#ifdef R3D
      call zboundary(h)
#endif /* R3D */

    end select

#ifdef PROFILE
! stop accounting time for the boundary update
!
    call stop_timer(ibn)
#endif /* PROFILE */

!-------------------------------------------------------------------------------
!
  end subroutine update_boundaries
#ifdef BFIELD
!
!===============================================================================
!
! subroutine UPDATE_MAGNETIC_BOUNDARIES:
! -------------------------------------
!
!   Subroutine updates magnetic boundary conditions in selected order.
!
!   Arguments:
!
!     h - a time interval for the time varying boundary conditions
!
!===============================================================================
!
  subroutine update_magnetic_boundaries(h)

! declare all variables as implicit
!
    implicit none

! subroutine arguments
!
    real, intent(in) :: h
!
!-------------------------------------------------------------------------------
!
#ifdef PROFILE
! start accounting time for the boundary update
!
    call start_timer(ibn)
#endif /* PROFILE */

#ifdef VPOT
! depending on the parameter orbndry update boundary in a specific order
!
    select case(orbndry)

    case ("yzx")
      call yboundary_magnetic(h)
#ifdef R3D
      call zboundary_magnetic(h)
#endif /* R3D */
      call xboundary_magnetic(h)

    case ("zxy")
#ifdef R3D
      call zboundary_magnetic(h)
#endif /* R3D */
      call xboundary_magnetic(h)
      call yboundary_magnetic(h)

    case ("yxz")
      call yboundary_magnetic(h)
      call xboundary_magnetic(h)
#ifdef R3D
      call zboundary_magnetic(h)
#endif /* R3D */

    case ("xzy")
      call xboundary_magnetic(h)
#ifdef R3D
      call zboundary_magnetic(h)
#endif /* R3D */
      call yboundary_magnetic(h)

    case ("zyx")
#ifdef R3D
      call zboundary_magnetic(h)
#endif /* R3D */
      call yboundary_magnetic(h)
      call xboundary_magnetic(h)

   case default
      call xboundary_magnetic(h)
      call yboundary_magnetic(h)
#ifdef R3D
      call zboundary_magnetic(h)
#endif /* R3D */

    end select
#endif /* VPOT */

#ifdef PROFILE
! stop accounting time for the boundary update
!
    call stop_timer(ibn)
#endif /* PROFILE */

!-------------------------------------------------------------------------------
!
  end subroutine update_magnetic_boundaries
#endif /* BFIELD */
!
!===============================================================================
!
! subroutine XBOUNDARY:
! --------------------
!
!   Subroutine updates array boundaries along the X direction.
!
!   Arguments:
!
!     h - a time interval for the time varying boundary conditions
!
!===============================================================================
!
  subroutine xboundary(h)

! include external subroutines and variables used this subroutine
!
    use equations , only : prim2cons
    use mesh      , only : ng, im, jm, km
    use mesh      , only : ib, ie, ibl, ibu, iel, ieu
#ifdef GLM
    use mesh      , only : dx, dyih, dzih
#endif /* GLM */
#ifdef MPI
    use mpitools  , only : pcoords, pparity, pneighs, periodic
    use mpitools  , only : send_real_array, receive_real_array
#endif /* MPI */
    use variables , only : nt, nu, nb, nk, nl
    use variables , only : uu, qp, qp_bnd
#ifdef VPOT
    use variables , only : bi
#endif /* VPOT */
    use variables , only : idn, ivx, ivy, ivz
#ifdef ADI
    use variables , only : ipr
#endif /* ADI */
#ifdef GLM
    use variables , only : ibx, iby, ibz, iph
#endif /* GLM */

! local variables are not implicit by default
!
    implicit none

! subroutine arguments
!
    real, intent(in)             :: h

! local indices
!
    integer                      :: iret
    integer                      :: i, j, k
    integer                      :: is, it
    integer                      :: itm1, itp1
#ifdef GLM
    integer                      :: itm2, itp2
    integer                      :: jm1, jp1, km1, kp1
    real                         :: dbx, dby, dbz
#endif /* GLM */

! variables saved between calls
!
    logical, save                :: first = .true.
#ifdef MPI
    integer, save                :: nn    = 1
#endif /* MPI */

! temporary array for variable conversion
!
    real, dimension(nt,jm)       :: u

#ifdef MPI
! an array for the MPI exchange
!
    real, dimension(nl,ng,jm,km) :: qb
#endif /* MPI */
!
!-------------------------------------------------------------------------------
!
! prepare some variables in the first execution
!
    if (first) then

#ifdef MPI
! calculate the size of array exchange
!
      nn    = nl * ng * jm * km
#endif /* MPI */

! change the flag first
!
      first = .false.

    end if

! === lower boundary ===
!
    select case(bnd_type(1,1))

#ifdef MPI
!-   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -
!
    case(bnd_exchange)

! "exchange" boundary conditions are used only between MPI nodes and use
! different schemes for periodic and other types of conditions
!
      if (periodic(1)) then

! this scheme sends at the same time data from even to odd nodes, and then in
! the second step, from odd to even nodes
!
        if (pparity(1) .eq. 0) then

! copy data to the exchange array
!
          qb( 1:nt,   :   ,:,:) = qp( 1:nt,iel:ie ,:,:)
#ifdef VPOT
          qb(nk:nl,   :   ,:,:) = bi( 1:nb,iel:ie ,:,:)
#endif /* VPOT */

! send data to and then receive new data from the right node
!
          call send_real_array   (nn, pneighs(1,2), 111, qb(:,:,:,:), iret)
          call receive_real_array(nn, pneighs(1,2), 112, qb(:,:,:,:), iret)

! update ghost zones
!
          qp( 1:nt,ieu:im ,:,:) = qb( 1:nt,   :   ,:,:)
#ifdef VPOT
          bi( 1:nb,ieu:im ,:,:) = qb(nk:nl,   :   ,:,:)
#endif /* VPOT */

        else

! receive data from the left node
!
          call receive_real_array(nn, pneighs(1,1), 111, qb(:,:,:,:), iret)

! update ghost zones
!
          qp( 1:nt,  1:ibl,:,:) = qb( 1:nt,   :   ,:,:)
#ifdef VPOT
          bi( 1:nb,  1:ibl,:,:) = qb(nk:nl,   :   ,:,:)
#endif /* VPOT */

! copy data to the exchange array
!
          qb( 1:nt,   :   ,:,:) = qp( 1:nt, ib:ibu,:,:)
#ifdef VPOT
          qb(nk:nl,   :   ,:,:) = bi( 1:nb, ib:ibu,:,:)
#endif /* VPOT */

! send data to the left node
!
          call send_real_array   (nn, pneighs(1,1), 112, qb(:,:,:,:), iret)

        end if

      else

! receive data from the left node
!
        call receive_real_array(nn, pneighs(1,1), 121, qb(:,:,:,:), iret)

! update ghost zones
!
        qp( 1:nt,  1:ibl,:,:) = qb( 1:nt,   :   ,:,:)
#ifdef VPOT
        bi( 1:nb,  1:ibl,:,:) = qb(nk:nl,   :   ,:,:)
#endif /* VPOT */

! copy data to the exchange array
!
        qb( 1:nt,   :   ,:,:) = qp( 1:nt, ib:ibu,:,:)
#ifdef VPOT
        qb(nk:nl,   :   ,:,:) = bi( 1:nb, ib:ibu,:,:)
#endif /* VPOT */

! send data to the left node
!
        call send_real_array   (nn, pneighs(1,1), 122, qb(:,:,:,:), iret)

      end if
#endif /* MPI */

!-   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -
!
    case(bnd_open)

! "open" boundary conditions set the derivative normal to boundary to zero
!
      do i = 1, ng

! prepare indices
!
        it   = ib - i
        itp1 = it + 1

! update the boundary
!
#ifdef GLM
        qp(1:nt,it,:,:) = qp(1:nt,itp1,:,:)
        qp( iph,it,:,:) = 0.0d+00

! update normal component of magnetic field
!
        do k = 1, km
#ifdef R3D
          km1 = max( 1, k - 1)
          kp1 = min(km, k + 1)
#endif /* R3D */
          do j = 1, jm
            jm1 = max( 1, j - 1)
            jp1 = min(jm, j + 1)

! calculate derivatives of transverse components of magnetic field
!
            dby = dyih * (qp(iby,itp1,jp1,k  ) - qp(iby,itp1,jm1,k  ))
#ifdef R3D
            dbz = dzih * (qp(ibz,itp1,j  ,kp1) - qp(ibz,itp1,j  ,km1))
#endif /* R3D */

! update normal component of magnetic field from the divergence free condition
!
#ifdef R3D
            qp(ibx,it,j,k) = qp(ibx,itp1,j,k) + dx * (dby + dbz)
#else /* R3D */
            qp(ibx,it,j,k) = qp(ibx,itp1,j,k) + dx *  dby
#endif /* R3D */
          end do
        end do
#else /* GLM */
        qp(1:nu,it,:,:) = qp(1:nu,itp1,:,:)
#endif /* GLM */

      end do

!-   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -
!
    case(bnd_outflow)

! "outflow" boundary conditions are the same conditions as "open" but the normal
! component of velocity is set to zero in inflow locations
!
      do i = 1, ng

! prepare indices
!
        it   = ib - i
        itp1 = it + 1

! update the boundary
!
#ifdef GLM
        qp(1:nt,it,:,:) = qp(1:nt,itp1,:,:)
        qp( iph,it,:,:) = 0.0d+00

! update normal component of magnetic field
!
        do k = 1, km
#ifdef R3D
          km1 = max( 1, k - 1)
          kp1 = min(km, k + 1)
#endif /* R3D */
          do j = 1, jm
            jm1 = max( 1, j - 1)
            jp1 = min(jm, j + 1)

! calculate derivatives of transverse components of magnetic field
!
            dby = dyih * (qp(iby,itp1,jp1,k  ) - qp(iby,itp1,jm1,k  ))
#ifdef R3D
            dbz = dzih * (qp(ibz,itp1,j  ,kp1) - qp(ibz,itp1,j  ,km1))
#endif /* R3D */

! update normal component of magnetic field from the divergence free condition
!
#ifdef R3D
            qp(ibx,it,j,k) = qp(ibx,itp1,j,k) + dx * (dby + dbz)
#else /* R3D */
            qp(ibx,it,j,k) = qp(ibx,itp1,j,k) + dx *  dby
#endif /* R3D */
          end do
        end do
#else /* GLM */
        qp(1:nu,it,:,:) = qp(1:nu,itp1,:,:)
#endif /* GLM */

! limit the normal component of velocity
!
        qp(ivx,it,:,:) = min(0.0d+00, qp(ivx,itp1,:,:))

      end do

!-   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -
!
    case(bnd_reflective)

! "reflective" boundary conditions reverse layers inverting the normal component
! of velocity;
!
      do i = 1, ng

! prepare the source and target indices
!
        is   = ibl + i
        it   = ib  - i

! update the boundary
!
#ifdef GLM
        qp(1:nt,it,:,:) = qp(1:nt,is,:,:)
#else /* GLM */
        qp(1:nu,it,:,:) = qp(1:nu,is,:,:)
#endif /* GLM */

! change the sign of the normal component of velocity
!
        qp(ivx,it,:,:) = - qp(ivx,is,:,:)
#ifdef GLM
        qp(ibx,it,:,:) = - qp(ibx,is,:,:)
#endif /* GLM */

      end do

#ifdef BFIELD
!-   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -
!
    case(bnd_current_free)

! "current-free" boundary conditions are the same as "open", but the magnetic
! field is extrapolated to be free of current density
!
      do i = 1, ng

! prepare indices
!
        it   = ib - i
        itp1 = it + 1

! update the boundary
!
#ifdef GLM
        qp(1:nt,it,:,:) = qp(1:nt,itp1,:,:)
#else /* GLM */
        qp(1:nu,it,:,:) = qp(1:nu,itp1,:,:)
#endif /* GLM */

      end do
#endif /* BFIELD */

!-   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -
!
    case default

! "periodic" boundary conditions simply copy one side to another
!
      do i = 1, ng

! prepare source and target indices
!
        is = ieu - i
        it = ib  - i

! update the boundary
!
#ifdef GLM
        qp(1:nt,it,:,:) = qp(1:nt,is,:,:)
#else /* GLM */
        qp(1:nu,it,:,:) = qp(1:nu,is,:,:)
#ifdef VPOT
        bi(1:nb,it,:,:) = bi(1:nb,is,:,:)
#endif /* VPOT */
#endif /* GLM */

      end do

    end select

! === upper boundary ===
!
    select case(bnd_type(1,2))

#ifdef MPI
!-   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -
!
    case(bnd_exchange)

! "exchange" boundary conditions are used only between MPI nodes and use
! different schemes for periodic and other types of conditions
!
      if (periodic(1)) then

! this scheme sends at the same time data from even to odd nodes, and then in
! the second step, from odd to even nodes
!
        if (pparity(1) .eq. 0) then

! copy data to the exchange array
!
          qb( 1:nt,   :   ,:,:) = qp( 1:nt, ib:ibu,:,:)
#ifdef VPOT
          qb(nk:nl,   :   ,:,:) = bi( 1:nb, ib:ibu,:,:)
#endif /* VPOT */

! send data to and then receive new data from the left node
!
          call send_real_array   (nn, pneighs(1,1), 113, qb(:,:,:,:), iret)
          call receive_real_array(nn, pneighs(1,1), 114, qb(:,:,:,:), iret)

! update ghost zones
!
          qp( 1:nt,  1:ibl,:,:) = qb( 1:nt,   :   ,:,:)
#ifdef VPOT
          bi( 1:nb,  1:ibl,:,:) = qb(nk:nl,   :   ,:,:)
#endif /* VPOT */

        else

! receive data from the right node
!
          call receive_real_array(nn, pneighs(1,2), 113, qb(:,:,:,:), iret)

! update ghost zones
!
          qp( 1:nt,ieu:im ,:,:) = qb( 1:nt,   :   ,:,:)
#ifdef VPOT
          bi( 1:nb,ieu:im ,:,:) = qb(nk:nl,   :   ,:,:)
#endif /* VPOT */

! copy data to the exchange array
!
          qb( 1:nt,   :   ,:,:) = qp( 1:nt,iel:ie ,:,:)
#ifdef VPOT
          qb(nk:nl,   :   ,:,:) = bi( 1:nb,iel:ie ,:,:)
#endif /* VPOT */

! send data to the right node
!
          call send_real_array   (nn, pneighs(1,2), 114, qb(:,:,:,:), iret)

        end if

      else

! copy data to the exchange array
!
        qb( 1:nt,   :   ,:,:) = qp( 1:nt,iel:ie ,:,:)
#ifdef VPOT
        qb(nk:nl,   :   ,:,:) = bi( 1:nb,iel:ie ,:,:)
#endif /* VPOT */

! send data to the right node
!
        call send_real_array   (nn, pneighs(1,2), 121, qb(:,:,:,:), iret)

! receive data from the right node
!
        call receive_real_array(nn, pneighs(1,2), 122, qb(:,:,:,:), iret)

! update ghost zones
!
        qp( 1:nt,ieu:im ,:,:) = qb( 1:nt,   :   ,:,:)
#ifdef VPOT
        bi( 1:nb,ieu:im ,:,:) = qb(nk:nl,   :   ,:,:)
#endif /* VPOT */

      end if
#endif /* MPI */

!-   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -
!
    case(bnd_open)

! "open" boundary conditions set the derivative normal to boundary to zero
!
      do i = 1, ng

! prepare indices
!
        it   = ie + i
        itm1 = it - 1

! update the boundary
!
#ifdef GLM
        qp(1:nt,it,:,:) = qp(1:nt,itm1,:,:)
        qp( iph,it,:,:) = 0.0d+00

! update normal component of magnetic field
!
        do k = 1, km
#ifdef R3D
          km1 = max( 1, k - 1)
          kp1 = min(km, k + 1)
#endif /* R3D */
          do j = 1, jm
            jm1 = max( 1, j - 1)
            jp1 = min(jm, j + 1)

! calculate derivatives of transverse components of magnetic field
!
            dby = dyih * (qp(iby,itm1,jp1,k  ) - qp(iby,itm1,jm1,k  ))
#ifdef R3D
            dbz = dzih * (qp(ibz,itm1,j  ,kp1) - qp(ibz,itm1,j  ,km1))
#endif /* R3D */

! update normal component of magnetic field from the divergence free condition
!
#ifdef R3D
            qp(ibx,it,j,k) = qp(ibx,itm1,j,k) - dx * (dby + dbz)
#else /* R3D */
            qp(ibx,it,j,k) = qp(ibx,itm1,j,k) - dx *  dby
#endif /* R3D */
          end do
        end do
#else /* GLM */
        qp(1:nu,it,:,:) = qp(1:nu,itm1,:,:)
#endif /* GLM */

      end do

!-   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -
!
    case(bnd_outflow)

! "outflow" boundary conditions are the same conditions as "open" but the normal
! component of velocity is set to zero in inflow locations
!
      do i = 1, ng

! prepare indices
!
        it   = ie + i
        itm1 = it - 1

! update the boundary
!
#ifdef GLM
        qp(1:nt,it,:,:) = qp(1:nt,itm1,:,:)
        qp( iph,it,:,:) = 0.0d+00

! update normal component of magnetic field
!
        do k = 1, km
#ifdef R3D
          km1 = max( 1, k - 1)
          kp1 = min(km, k + 1)
#endif /* R3D */
          do j = 1, jm
            jm1 = max( 1, j - 1)
            jp1 = min(jm, j + 1)

! calculate derivatives of transverse components of magnetic field
!
            dby = dyih * (qp(iby,itm1,jp1,k  ) - qp(iby,itm1,jm1,k  ))
#ifdef R3D
            dbz = dzih * (qp(ibz,itm1,j  ,kp1) - qp(ibz,itm1,j  ,km1))
#endif /* R3D */

! update normal component of magnetic field from the divergence free condition
!
#ifdef R3D
            qp(ibx,it,j,k) = qp(ibx,itm1,j,k) - dx * (dby + dbz)
#else /* R3D */
            qp(ibx,it,j,k) = qp(ibx,itm1,j,k) - dx *  dby
#endif /* R3D */
          end do
        end do
#else /* GLM */
        qp(1:nu,it,:,:) = qp(1:nu,itm1,:,:)
#endif /* GLM */

! limit the normal component of velocity
!
        qp(ivx,it,:,:) = max(0.0d+00, qp(ivx,itm1,:,:))

      end do

!-   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -
!
    case(bnd_reflective)

! "reflective" boundary conditions reverse layers inverting the normal component
! of velocity;
!
      do i = 1, ng

! prepare source and target indices
!
        is   = ieu - i
        it   = ie  + i

! update the boundary
!
#ifdef GLM
        qp(1:nt,it,:,:) = qp(1:nt,is,:,:)
#else /* GLM */
        qp(1:nu,it,:,:) = qp(1:nu,is,:,:)
#endif /* GLM */

! change the sign of the normal component of velocity
!
        qp(ivx,it,:,:) = - qp(ivx,is,:,:)
#ifdef GLM
        qp(ibx,it,:,:) = - qp(ibx,is,:,:)
#endif /* GLM */

      end do

#ifdef BFIELD
!-   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -
!
    case(bnd_current_free)

! "current-free" boundary conditions are the same as "open", but the magnetic
! field is extrapolated to be free of current density
!
      do i = 1, ng

! prepare indices
!
        it   = ie + i
        itm1 = it - 1

! update the boundary
!
#ifdef GLM
        qp(1:nt,it,:,:) = qp(1:nt,itm1,:,:)

! update normal component of magnetic field
!
        do k = 1, km
#ifdef R3D
          km1 = max( 1, k - 1)
          kp1 = min(km, k + 1)
#endif /* R3D */
          do j = 1, jm
            jm1 = max( 1, j - 1)
            jp1 = min(jm, j + 1)

! calculate derivatives of transverse components of magnetic field
!
            dby = dyih * (qp(iby,itm1,jp1,k  ) - qp(iby,itm1,jm1,k  ))
#ifdef R3D
            dbz = dzih * (qp(ibz,itm1,j  ,kp1) - qp(ibz,itm1,j  ,km1))
#endif /* R3D */

! update normal component of magnetic field from the divergence free condition
!
#ifdef R3D
            qp(ibx,it,j,k) = qp(ibx,itm1,j,k) - dx * (dby + dbz)
#else /* R3D */
            qp(ibx,it,j,k) = qp(ibx,itm1,j,k) - dx *  dby
#endif /* R3D */
          end do
        end do
#else /* GLM */
        qp(1:nu,it,:,:) = qp(1:nu,itm1,:,:)
#endif /* GLM */

      end do
#endif /* BFIELD */

!-   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -
!
    case default

! "periodic" boundary conditions simply copy one side to another
!
      do i = 1, ng

! prepare source and target indices
!
        is = ibl + i
        it = ie  + i

! update the boundary
!
#ifdef GLM
        qp(1:nt,it,:,:) = qp(1:nt,is,:,:)
#else /* GLM */
        qp(1:nu,it,:,:) = qp(1:nu,is,:,:)
#ifdef VPOT
        bi(1:nb,it,:,:) = bi(1:nb,is,:,:)
#endif /* VPOT */
#endif /* GLM */

      end do

    end select

!- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
!
! convert primitive variables at boundaries to conservative representation
!
    do k = 1, km
      do i = 1, ng
        call prim2cons(jm, qp(1:nt,i,1:jm,k), u(1:nt,1:jm))

#ifdef GLM
        uu(1:nt,i,1:jm,k) = u(1:nt,1:jm)
#else /* GLM */
        uu(1:nu,i,1:jm,k) = u(1:nu,1:jm)
#endif /* GLM */
      end do

      do i = ieu, im
        call prim2cons(jm, qp(1:nt,i,1:jm,k), u(1:nt,1:jm))

#ifdef GLM
        uu(1:nt,i,1:jm,k) = u(1:nt,1:jm)
#else /* GLM */
        uu(1:nu,i,1:jm,k) = u(1:nu,1:jm)
#endif /* GLM */
      end do
    end do

!-------------------------------------------------------------------------------
!
  end subroutine xboundary
!
!===============================================================================
!
! subroutine YBOUNDARY:
! --------------------
!
!   Subroutine updates array boundaries along the Y direction.
!
!   Arguments:
!
!     h - a time interval for the time varying boundary conditions
!
!===============================================================================
!
  subroutine yboundary(h)

! include external subroutines and variables used this subroutine
!
    use equations , only : prim2cons
    use mesh      , only : ng, im, jm, km
    use mesh      , only : jb, je, jbl, jbu, jel, jeu
#ifdef GLM
    use mesh      , only : dy, dxih, dzih
#endif /* GLM */
#ifdef MPI
    use mpitools  , only : pcoords, pparity, pneighs, periodic
    use mpitools  , only : send_real_array, receive_real_array
#endif /* MPI */
    use variables , only : nt, nu, nb, nk, nl
    use variables , only : uu, qp, qp_bnd
#ifdef VPOT
    use variables , only : bi
#endif /* VPOT */
    use variables , only : idn, ivx, ivy, ivz
#ifdef ADI
    use variables , only : ipr
#endif /* ADI */
#ifdef GLM
    use variables , only : ibx, iby, ibz, iph
#endif /* GLM */

! local variables are not implicit by default
!
    implicit none

! subroutine arguments
!
    real, intent(in)             :: h

! local variables
!
    integer                      :: iret
    integer                      :: i, j, k
    integer                      :: js, jt
    integer                      :: jtm1, jtp1
#ifdef GLM
    integer                      :: jtm2, jtp2
    integer                      :: im1, ip1, km1, kp1
    real                         :: dbx, dby, dbz
#endif /* GLM */

! variables saved between calls
!
    logical, save                :: first = .true.
#ifdef MPI
    integer, save                :: nn    = 1
#endif /* MPI */

! temporary array for variable conversion
!
    real, dimension(nt,im)       :: u

#ifdef MPI
! array for the MPI exchange
!
    real, dimension(nl,im,ng,km) :: qb
#endif /* MPI */
!
!-------------------------------------------------------------------------------
!
! prepare some variables in the first execution
!
    if (first) then

#ifdef MPI
! calculate the size of array exchange
!
      nn    = nl * im * ng * km
#endif /* MPI */

! change the flag first
!
      first = .false.

    end if

! === lower boundary ===
!
    select case(bnd_type(2,1))

#ifdef MPI
!-   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -
!
    case(bnd_exchange)

! "exchange" boundary conditions are used only between MPI nodes and use
! different schemes for periodic and other types of conditions
!
      if (periodic(2)) then

! this scheme sends at the same time data from even to odd nodes, and then in
! the second step, from odd to even nodes
!
        if (pparity(2) .eq. 0) then

! copy data to the exchange array
!
          qb( 1:nt,:,   :   ,:) = qp( 1:nt,:,jel:je ,:)
#ifdef VPOT
          qb(nk:nl,:,   :   ,:) = bi( 1:nb,:,jel:je ,:)
#endif /* VPOT */

! send data to and then receive new data from the right node
!
          call send_real_array   (nn, pneighs(2,2), 211, qb(:,:,:,:), iret)
          call receive_real_array(nn, pneighs(2,2), 212, qb(:,:,:,:), iret)

! update ghost zones
!
          qp( 1:nt,:,jeu:jm ,:) = qb( 1:nt,:,   :   ,:)
#ifdef VPOT
          bi( 1:nb,:,jeu:jm ,:) = qb(nk:nl,:,   :   ,:)
#endif /* VPOT */

        else

! receive data from the left node
!
          call receive_real_array(nn, pneighs(2,1), 211, qb(:,:,:,:), iret)

! update ghost zones
!
          qp( 1:nt,:,  1:jbl,:) = qb( 1:nt,:,   :   ,:)
#ifdef VPOT
          bi( 1:nb,:,  1:jbl,:) = qb(nk:nl,:,   :   ,:)
#endif /* VPOT */

! copy data to the exchange array
!
          qb( 1:nt,:,   :   ,:) = qp( 1:nt,:, jb:jbu,:)
#ifdef VPOT
          qb(nk:nl,:,   :   ,:) = bi( 1:nb,:, jb:jbu,:)
#endif /* VPOT */

! send data to the left node
!
          call send_real_array   (nn, pneighs(2,1), 212, qb(:,:,:,:), iret)

        end if

      else

! receive data from the left node
!
        call receive_real_array(nn, pneighs(2,1), 221, qb(:,:,:,:), iret)

! update ghost zones
!
        qp( 1:nt,:,  1:jbl,:) = qb( 1:nt,:,   :   ,:)
#ifdef VPOT
        bi( 1:nb,:,  1:jbl,:) = qb(nk:nl,:,   :   ,:)
#endif /* VPOT */

! copy data to the exchange array
!
        qb( 1:nt,:,   :   ,:) = qp( 1:nt,:, jb:jbu,:)
#ifdef VPOT
        qb(nk:nl,:,   :   ,:) = bi( 1:nb,:, jb:jbu,:)
#endif /* VPOT */

! send data to the left node
!
        call send_real_array   (nn, pneighs(2,1), 222, qb(:,:,:,:), iret)

      end if
#endif /* MPI */

!-   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -
!
    case(bnd_open)

! "open" boundary conditions set the derivative normal to boundary to zero
!
      do j = 1, ng

! prepare indices
!
        jt   = jb - j
        jtp1 = jt + 1

! update the boundary
!
#ifdef GLM
        qp(1:nt,:,jt,:) = qp(1:nt,:,jtp1,:)
        qp( iph,:,jt,:) = 0.0d+00

! update normal component of magnetic field
!
        do k = 1, km
#ifdef R3D
          km1 = max( 1, k - 1)
          kp1 = min(km, k + 1)
#endif /* R3D */
          do i = 1, im
            im1 = max( 1, i - 1)
            ip1 = min(im, i + 1)

! calculate derivatives of transversal components of magnetic field
!
            dbx = dxih * (qp(ibx,ip1,jtp1,k  ) - qp(ibx,im1,jtp1,k  ))
#ifdef R3D
            dbz = dzih * (qp(ibz,i  ,jtp1,kp1) - qp(ibz,i  ,jtp1,km1))
#endif /* R3D */

! update normal component of magnetic field from the divergence free condition
!
#ifdef R3D
            qp(iby,i,jt,k) = qp(iby,i,jtp1,k) + dy * (dbx + dbz)
#else /* R3D */
            qp(iby,i,jt,k) = qp(iby,i,jtp1,k) + dy *  dbx
#endif /* R3D */
          end do
        end do
#else /* GLM */
        qp(1:nu,:,jt,:) = qp(1:nu,:,jtp1,:)
#endif /* GLM */

      end do

!-   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -
!
    case(bnd_outflow)

! "outflow" boundary conditions are the same conditions as "open" but the normal
! component of velocity is set to zero in inflow locations
!
      do j = 1, ng

! prepare indices
!
        jt   = jb - j
        jtp1 = jt + 1

! update the boundary
!
#ifdef GLM
        qp(1:nt,:,jt,:) = qp(1:nt,:,jtp1,:)
        qp( iph,:,jt,:) = 0.0d+00

! update normal component of magnetic field
!
        do k = 1, km
#ifdef R3D
          km1 = max( 1, k - 1)
          kp1 = min(km, k + 1)
#endif /* R3D */
          do i = 1, im
            im1 = max( 1, i - 1)
            ip1 = min(im, i + 1)

! calculate derivatives of transversal components of magnetic field
!
            dbx = dxih * (qp(ibx,ip1,jtp1,k  ) - qp(ibx,im1,jtp1,k  ))
#ifdef R3D
            dbz = dzih * (qp(ibz,i  ,jtp1,kp1) - qp(ibz,i  ,jtp1,km1))
#endif /* R3D */

! update normal component of magnetic field from the divergence free condition
!
#ifdef R3D
            qp(iby,i,jt,k) = qp(iby,i,jtp1,k) + dy * (dbx + dbz)
#else /* R3D */
            qp(iby,i,jt,k) = qp(iby,i,jtp1,k) + dy *  dbx
#endif /* R3D */
          end do
        end do
#else /* GLM */
        qp(1:nu,:,jt,:) = qp(1:nu,:,jtp1,:)
#endif /* GLM */

! limit the normal component of velocity
!
        qp(ivy,:,jt,:) = min(0.0d+00, qp(ivy,:,jtp1,:))

      end do

!-   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -
!
    case(bnd_reflective)

! "reflective" boundary conditions reverse layers inverting the normal component
! of velocity;
!
      do j = 1, ng

! prepare the source and target indices
!
        js   = jbl + j
        jt   = jb  - j

! update the boundary
!
#ifdef GLM
        qp(1:nt,:,jt,:) = qp(1:nt,:,js,:)
#else /* GLM */
        qp(1:nu,:,jt,:) = qp(1:nu,:,js,:)
#endif /* GLM */

! change the sign of the normal component of velocity
!
        qp(ivy,:,jt,:) = - qp(ivy,:,js,:)
#ifdef GLM
        qp(iby,:,jt,:) = - qp(iby,:,js,:)
#endif /* GLM */

      end do

#ifdef BFIELD
!-   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -
!
    case(bnd_current_free)

! "current-free" boundary conditions are the same as "open", but the magnetic
! field is extrapolated to be free of current density
!
      do j = 1, ng

! prepare indices
!
        jt   = jb - j
        jtp1 = jt + 1

! update the boundary
!
#ifdef GLM
        qp(1:nt,:,jt,:) = qp(1:nt,:,jtp1,:)

! update normal component of magnetic field
!
        do k = 1, km
#ifdef R3D
          km1 = max( 1, k - 1)
          kp1 = min(km, k + 1)
#endif /* R3D */
          do i = 1, im
            im1 = max( 1, i - 1)
            ip1 = min(im, i + 1)

! calculate derivatives of transversal components of magnetic field
!
            dbx = dxih * (qp(ibx,ip1,jtp1,k  ) - qp(ibx,im1,jtp1,k  ))
#ifdef R3D
            dbz = dzih * (qp(ibz,i  ,jtp1,kp1) - qp(ibz,i  ,jtp1,km1))
#endif /* R3D */

! update normal component of magnetic field from the divergence free condition
!
#ifdef R3D
            qp(iby,i,jt,k) = qp(iby,i,jtp1,k) + dy * (dbx + dbz)
#else /* R3D */
            qp(iby,i,jt,k) = qp(iby,i,jtp1,k) + dy *  dbx
#endif /* R3D */
          end do
        end do
#else /* GLM */
        qp(1:nu,:,jt,:) = qp(1:nu,:,jtp1,:)
#endif /* GLM */

      end do
#endif /* BFIELD */

!-   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -
!
    case default

! "periodic" boundary conditions simply copy one side to another
!
      do j = 1, ng

! prepare source and target indices
!
        js = jeu - j
        jt = jb  - j

! update the boundary
!
#ifdef GLM
        qp(1:nt,:,jt,:) = qp(1:nt,:,js,:)
#else /* GLM */
        qp(1:nu,:,jt,:) = qp(1:nu,:,js,:)
#ifdef VPOT
        bi(1:nb,:,jt,:) = bi(1:nb,:,js,:)
#endif /* VPOT */
#endif /* GLM */

      end do

    end select

! === upper boundary ===
!
    select case(bnd_type(2,2))

#ifdef MPI
!-   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -
!
    case(bnd_exchange)

! "exchange" boundary conditions are used only between MPI nodes and use
! different schemes for periodic and other types of conditions
!
      if (periodic(2)) then

! this scheme sends at the same time data from even to odd nodes, and then in
! the second step, from odd to even nodes
!
        if (pparity(2) .eq. 0) then

! copy data to the exchange array
!
          qb( 1:nt,:,   :   ,:) = qp( 1:nt,:, jb:jbu,:)
#ifdef VPOT
          qb(nk:nl,:,   :   ,:) = bi( 1:nb,:, jb:jbu,:)
#endif /* VPOT */

! send data to and then receive new data from the left node
!
          call send_real_array   (nn, pneighs(2,1), 213, qb(:,:,:,:), iret)
          call receive_real_array(nn, pneighs(2,1), 214, qb(:,:,:,:), iret)

! update ghost zones
!
          qp( 1:nt,:,  1:jbl,:) = qb( 1:nt,:,   :   ,:)
#ifdef VPOT
          bi( 1:nb,:,  1:jbl,:) = qb(nk:nl,:,   :   ,:)
#endif /* VPOT */

        else

! receive data from the right node
!
          call receive_real_array(nn, pneighs(2,2), 213, qb(:,:,:,:), iret)

! update ghost zones
!
          qp( 1:nt,:,jeu:jm ,:) = qb( 1:nt,:,   :   ,:)
#ifdef VPOT
          bi( 1:nb,:,jeu:jm ,:) = qb(nk:nl,:,   :   ,:)
#endif /* VPOT */

! copy data to the exchange array
!
          qb( 1:nt,:,   :   ,:) = qp( 1:nt,:,jel:je ,:)
#ifdef VPOT
          qb(nk:nl,:,   :   ,:) = bi( 1:nb,:,jel:je ,:)
#endif /* VPOT */

! send data to the right node
!
          call send_real_array   (nn, pneighs(2,2), 214, qb(:,:,:,:), iret)

        end if

      else

! copy data to the exchange array
!
        qb( 1:nt,:,   :   ,:) = qp( 1:nt,:,jel:je ,:)
#ifdef VPOT
        qb(nk:nl,:,   :   ,:) = bi( 1:nb,:,jel:je ,:)
#endif /* VPOT */

! send data to the right node
!
        call send_real_array   (nn, pneighs(2,2), 221, qb(:,:,:,:), iret)

! receive data from the right node
!
        call receive_real_array(nn, pneighs(2,2), 222, qb(:,:,:,:), iret)

! update ghost zones
!
        qp( 1:nt,:,jeu:jm ,:) = qb( 1:nt,:,   :   ,:)
#ifdef VPOT
        bi( 1:nb,:,jeu:jm ,:) = qb(nk:nl,:,   :   ,:)
#endif /* VPOT */

      end if
#endif /* MPI */

!-   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -
!
    case(bnd_open)

! "open" boundary conditions set the derivative normal to boundary to zero
!
      do j = 1, ng

! prepare indices
!
        jt   = je + j
        jtm1 = jt - 1

! update the boundary
!
#ifdef GLM
        qp(1:nt,:,jt,:) = qp(1:nt,:,jtm1,:)
        qp( iph,:,jt,:) = 0.0d+00

! update normal component of magnetic field
!
        do k = 1, km
#ifdef R3D
          km1 = max( 1, k - 1)
          kp1 = min(km, k + 1)
#endif /* R3D */
          do i = 1, im
            im1 = max( 1, i - 1)
            ip1 = min(im, i + 1)

! calculate derivatives of transversal components of magnetic field
!
            dbx = dxih * (qp(ibx,ip1,jtm1,k  ) - qp(ibx,im1,jtm1,k  ))
#ifdef R3D
            dbz = dzih * (qp(ibz,i  ,jtm1,kp1) - qp(ibz,i  ,jtm1,km1))
#endif /* R3D */

! update normal component of magnetic field from the divergence free condition
!
#ifdef R3D
            qp(iby,i,jt,k) = qp(iby,i,jtm1,k) - dy * (dbx + dbz)
#else /* R3D */
            qp(iby,i,jt,k) = qp(iby,i,jtm1,k) - dy *  dbx
#endif /* R3D */
          end do
        end do
#else /* GLM */
        qp(1:nu,:,jt,:) = qp(1:nu,:,jtm1,:)
#endif /* GLM */

      end do

!-   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -
!
    case(bnd_outflow)

! "outflow" boundary conditions are the same conditions as "open" but the normal
! component of velocity is set to zero in inflow locations
!
      do j = 1, ng

! prepare indices
!
        jt   = je + j
        jtm1 = jt - 1

! update the boundary
!
#ifdef GLM
        qp(1:nt,:,jt,:) = qp(1:nt,:,jtm1,:)
        qp( iph,:,jt,:) = 0.0d+00

! update normal component of magnetic field
!
        do k = 1, km
#ifdef R3D
          km1 = max( 1, k - 1)
          kp1 = min(km, k + 1)
#endif /* R3D */
          do i = 1, im
            im1 = max( 1, i - 1)
            ip1 = min(im, i + 1)

! calculate derivatives of transversal components of magnetic field
!
            dbx = dxih * (qp(ibx,ip1,jtm1,k  ) - qp(ibx,im1,jtm1,k  ))
#ifdef R3D
            dbz = dzih * (qp(ibz,i  ,jtm1,kp1) - qp(ibz,i  ,jtm1,km1))
#endif /* R3D */

! update normal component of magnetic field from the divergence free condition
!
#ifdef R3D
            qp(iby,i,jt,k) = qp(iby,i,jtm1,k) - dy * (dbx + dbz)
#else /* R3D */
            qp(iby,i,jt,k) = qp(iby,i,jtm1,k) - dy *  dbx
#endif /* R3D */
          end do
        end do
#else /* GLM */
        qp(1:nu,:,jt,:) = qp(1:nu,:,jtm1,:)
#endif /* GLM */

! limit the normal component of velocity
!
        qp(ivy,:,jt,:) = max(0.0d+00, qp(ivy,:,jtm1,:))

      end do

!-   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -
!
    case(bnd_reflective)

! "reflective" boundary conditions reverse layers inverting the normal component
! of velocity;
!
      do j = 1, ng

! prepare source and target indices
!
        js   = jeu - j
        jt   = je  + j

! update the boundary
!
#ifdef GLM
        qp(1:nt,:,jt,:) = qp(1:nt,:,js,:)
#else /* GLM */
        qp(1:nu,:,jt,:) = qp(1:nu,:,js,:)
#endif /* GLM */

! change the sign of the normal component of velocity
!
        qp(ivy,:,jt,:) = - qp(ivy,:,js,:)
#ifdef GLM
        qp(iby,:,jt,:) = - qp(iby,:,js,:)
#endif /* GLM */

      end do

#ifdef BFIELD
!-   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -
!
    case(bnd_current_free)

! "current-free" boundary conditions are the same as "open", but the magnetic
! field is extrapolated to be free of current density
!
      do j = 1, ng

! prepare indices
!
        jt   = je + j
        jtm1 = jt - 1

! update the boundary
!
#ifdef GLM
        qp(1:nt,:,jt,:) = qp(1:nt,:,jtm1,:)

! update normal component of magnetic field
!
        do k = 1, km
#ifdef R3D
          km1 = max( 1, k - 1)
          kp1 = min(km, k + 1)
#endif /* R3D */
          do i = 1, im
            im1 = max( 1, i - 1)
            ip1 = min(im, i + 1)

! calculate derivatives of transversal components of magnetic field
!
            dbx = dxih * (qp(ibx,ip1,jtm1,k  ) - qp(ibx,im1,jtm1,k  ))
#ifdef R3D
            dbz = dzih * (qp(ibz,i  ,jtm1,kp1) - qp(ibz,i  ,jtm1,km1))
#endif /* R3D */

! update normal component of magnetic field from the divergence free condition
!
#ifdef R3D
            qp(iby,i,jt,k) = qp(iby,i,jtm1,k) - dy * (dbx + dbz)
#else /* R3D */
            qp(iby,i,jt,k) = qp(iby,i,jtm1,k) - dy *  dbx
#endif /* R3D */
          end do
        end do
#else /* GLM */
        qp(1:nu,:,jt,:) = qp(1:nu,:,jtm1,:)
#endif /* GLM */

      end do
#endif /* BFIELD */

!-   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -
!
    case default

! "periodic" boundary conditions simply copy one side to another
!
      do j = 1, ng

! prepare source and target indices
!
        js = jbl + j
        jt = je  + j

! update the boundary
!
#ifdef GLM
        qp(1:nt,:,jt,:) = qp(1:nt,:,js,:)
#else /* GLM */
        qp(1:nu,:,jt,:) = qp(1:nu,:,js,:)
#ifdef VPOT
        bi(1:nb,:,jt,:) = bi(1:nb,:,js,:)
#endif /* VPOT */
#endif /* GLM */

      end do

    end select

!- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
!
! convert primitive variables at boundaries to conservative representation
!
    do k = 1, km
      do j = 1, ng
        call prim2cons(im, qp(1:nt,1:im,j,k), u(1:nt,1:im))

#ifdef GLM
        uu(1:nt,1:im,j,k) = u(1:nt,1:im)
#else /* GLM */
        uu(1:nu,1:im,j,k) = u(1:nu,1:im)
#endif /* GLM */
      end do

      do j = jeu, jm
        call prim2cons(im, qp(1:nt,1:im,j,k), u(1:nt,1:im))

#ifdef GLM
        uu(1:nt,1:im,j,k) = u(1:nt,1:im)
#else /* GLM */
        uu(1:nu,1:im,j,k) = u(1:nu,1:im)
#endif /* GLM */
      end do
    end do

!-------------------------------------------------------------------------------
!
  end subroutine yboundary
#ifdef R3D
!
!===============================================================================
!
! subroutine ZBOUNDARY:
! --------------------
!
!   Subroutine updates array boundaries along the Z direction.
!
!   Arguments:
!
!     h - a time interval for the time varying boundary conditions
!
!===============================================================================
!
  subroutine zboundary(h)

! include external subroutines and variables used this subroutine
!
    use equations , only : prim2cons
    use mesh      , only : ng, im, jm, km
    use mesh      , only : kb, ke, kbl, kbu, kel, keu
#ifdef GLM
    use mesh      , only : dz, dxih, dyih
#endif /* GLM */
#ifdef MPI
    use mpitools  , only : pcoords, pparity, pneighs, periodic
    use mpitools  , only : send_real_array, receive_real_array
#endif /* MPI */
    use variables , only : nt, nu, nb, nk, nl
    use variables , only : uu, qp, qp_bnd
#ifdef VPOT
    use variables , only : bi
#endif /* VPOT */
    use variables , only : idn, ivx, ivy, ivz
#ifdef ADI
    use variables , only : ipr
#endif /* ADI */
#ifdef GLM
    use variables , only : ibx, iby, ibz, iph
#endif /* GLM */

! local variables are not implicit by default
!
    implicit none

! subroutine arguments
!
    real, intent(in)             :: h

! local variables
!
    integer                      :: iret
    integer                      :: i, j, k
    integer                      :: ks, kt
    integer                      :: ktm1, ktp1
#ifdef GLM
    integer                      :: ktm2, ktp2
    integer                      :: im1, ip1, jm1, jp1
    real                         :: dbx, dby, dbz
#endif /* GLM */

! variables saved between calls
!
    logical, save                :: first = .true.
#ifdef MPI
    integer, save                :: nn    = 1
#endif /* MPI */

! temporary array for variable conversion
!
    real, dimension(nt,im)       :: u

#ifdef MPI
! array for the MPI exchange
!
    real, dimension(nl,im,jm,ng) :: qb
#endif /* MPI */
!
!-------------------------------------------------------------------------------
!
! prepare some variables in the first execution
!
    if (first) then

#ifdef MPI
! calculate the size of array exchange
!
      nn    = nl * im * jm * ng
#endif /* MPI */

! change the flag first
!
      first = .false.

    end if

! === lower boundary ===
!
    select case(bnd_type(3,1))

#ifdef MPI
!-   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -
!
    case(bnd_exchange)

! "exchange" boundary conditions are used only between MPI nodes and use
! different schemes for periodic and other types of conditions
!
      if (periodic(3)) then

! this scheme sends at the same time data from even to odd nodes, and then in
! the second step, from odd to even nodes
!
        if (pparity(3) .eq. 0) then

! copy data to the exchange array
!
          qb( 1:nt,:,:,   :   ) = qp( 1:nt,:,:,kel:ke )
#ifdef VPOT
          qb(nk:nl,:,:,   :   ) = bi( 1:nb,:,:,kel:ke )
#endif /* VPOT */

! send data to and then receive new data from the right node
!
          call send_real_array   (nn, pneighs(3,2), 311, qb(:,:,:,:), iret)
          call receive_real_array(nn, pneighs(3,2), 312, qb(:,:,:,:), iret)

! update ghost zones
!
          qp( 1:nt,:,:,keu:km ) = qb( 1:nt,:,:,   :   )
#ifdef VPOT
          bi( 1:nb,:,:,keu:km ) = qb(nk:nl,:,:,   :   )
#endif /* VPOT */

        else

! receive data from the left node
!
          call receive_real_array(nn, pneighs(3,1), 311, qb(:,:,:,:), iret)

! update ghost zones
!
          qp( 1:nt,:,:,  1:kbl) = qb( 1:nt,:,:,   :   )
#ifdef VPOT
          bi( 1:nb,:,:,  1:kbl) = qb(nk:nl,:,:,   :   )
#endif /* VPOT */

! copy data to the exchange array
!
          qb( 1:nt,:,:,   :   ) = qp( 1:nt,:,:, kb:kbu)
#ifdef VPOT
          qb(nk:nl,:,:,   :   ) = bi( 1:nb,:,:, kb:kbu)
#endif /* VPOT */

! send data to the left node
!
          call send_real_array(nn, pneighs(3,1), 312, qb(:,:,:,:), iret)

        end if

      else

! receive data from the left node
!
        call receive_real_array(nn, pneighs(3,1), 321, qb(:,:,:,:), iret)

! update ghost zones
!
        qp( 1:nt,:,:,  1:kbl) = qb( 1:nt,:,:,   :   )
#ifdef VPOT
        bi( 1:nb,:,:,  1:kbl) = qb(nk:nl,:,:,   :   )
#endif /* VPOT */

! copy data to the exchange array
!
        qb( 1:nt,:,:,   :   ) = qp( 1:nt,:,:, kb:kbu)
#ifdef VPOT
        qb(nk:nl,:,:,   :   ) = bi( 1:nb,:,:, kb:kbu)
#endif /* VPOT */

! send data to the left node
!
        call send_real_array   (nn, pneighs(3,1), 322, qb(:,:,:,:), iret)

      end if
#endif /* MPI */

!-   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -
!
    case(bnd_open)

! "open" boundary conditions set the derivative normal to boundary to zero
!
      do k = 1, ng

! prepare indices
!
        kt   = kb - k
        ktp1 = kt + 1

! update the boundary
!
#ifdef GLM
        qp(1:nt,:,:,kt) = qp(1:nt,:,:,ktp1)
        qp( iph,:,:,kt) = 0.0d+00

! update normal component of magnetic field
!
        do j = 1, jm
          jm1 = max( 1, j - 1)
          jp1 = min(jm, j + 1)
          do i = 1, im
            im1 = max( 1, i - 1)
            ip1 = min(im, i + 1)

! calculate derivatives of transversal components of magnetic field
!
            dbx = dxih * (qp(ibx,ip1,j  ,ktp1) - qp(ibx,im1,j  ,ktp1))
            dby = dyih * (qp(iby,i  ,jp1,ktp1) - qp(iby,i  ,jm1,ktp1))

! update normal component of magnetic field from the divergence free condition
!
            qp(ibz,i,j,kt) = qp(ibz,i,j,ktp1) + dz * (dbx + dby)
          end do
        end do
#else /* GLM */
        qp(1:nu,:,:,kt) = qp(1:nu,:,:,ktp1)
#endif /* GLM */

      end do

!-   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -
!
    case(bnd_outflow)

! "outflow" boundary conditions are the same conditions as "open" but the normal
! component of velocity is set to zero in inflow locations
!
      do k = 1, ng

! prepare indices
!
        kt   = kb - k
        ktp1 = kt + 1

! update the boundary
!
#ifdef GLM
        qp(1:nt,:,:,kt) = qp(1:nt,:,:,ktp1)
        qp( iph,:,:,kt) = 0.0d+00

! update normal component of magnetic field
!
        do j = 1, jm
          jm1 = max( 1, j - 1)
          jp1 = min(jm, j + 1)
          do i = 1, im
            im1 = max( 1, i - 1)
            ip1 = min(im, i + 1)

! calculate derivatives of transversal components of magnetic field
!
            dbx = dxih * (qp(ibx,ip1,j  ,ktp1) - qp(ibx,im1,j  ,ktp1))
            dby = dyih * (qp(iby,i  ,jp1,ktp1) - qp(iby,i  ,jm1,ktp1))

! update normal component of magnetic field from the divergence free condition
!
            qp(ibz,i,j,kt) = qp(ibz,i,j,ktp1) + dz * (dbx + dby)
          end do
        end do
#else /* GLM */
        qp(1:nu,:,:,kt) = qp(1:nu,:,:,ktp1)
#endif /* GLM */

! limit the normal component of velocity
!
        qp(ivz,:,:,kt) = min(0.0d+00, qp(ivz,:,:,ktp1))

      end do

!-   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -
!
    case(bnd_reflective)

! "reflective" boundary conditions reverse layers inverting the normal component
! of velocity;
!
      do k = 1, ng

! prepare the source and target indices
!
        ks   = kbl + k
        kt   = kb  - k

! update the boundary
!
#ifdef GLM
        qp(1:nt,:,:,kt) = qp(1:nt,:,:,ks)
#else /* GLM */
        qp(1:nu,:,:,kt) = qp(1:nu,:,:,ks)
#endif /* GLM */

! change the sign of the normal component of velocity
!
        qp(ivz,:,:,kt) = - qp(ivz,:,:,ks)
#ifdef GLM
        qp(ibz,:,:,kt) = - qp(ibz,:,:,ks)
#endif /* GLM */

      end do

#ifdef BFIELD
!-   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -
!
    case(bnd_current_free)

! "current-free" boundary conditions are the same as "open", but the magnetic
! field is extrapolated to be free of current density
!
      do k = 1, ng

! prepare indices
!
        kt   = kb - k
        ktp1 = kt + 1

! update the boundary
!
#ifdef GLM
        qp(1:nt,:,:,kt) = qp(1:nt,:,:,ktp1)

! update normal component of magnetic field
!
        do j = 1, jm
          jm1 = max( 1, j - 1)
          jp1 = min(jm, j + 1)
          do i = 1, im
            im1 = max( 1, i - 1)
            ip1 = min(im, i + 1)

! calculate derivatives of transversal components of magnetic field
!
            dbx = dxih * (qp(ibx,ip1,j  ,ktp1) - qp(ibx,im1,j  ,ktp1))
            dby = dyih * (qp(iby,i  ,jp1,ktp1) - qp(iby,i  ,jm1,ktp1))

! update normal component of magnetic field from the divergence free condition
!
            qp(ibz,i,j,kt) = qp(ibz,i,j,ktp1) + dz * (dbx + dby)
          end do
        end do
#else /* GLM */
        qp(1:nu,:,:,kt) = qp(1:nu,:,:,ktp1)
#endif /* GLM */

      end do
#endif /* BFIELD */

!-   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -
!
    case default

! "periodic" boundary conditions simply copy one side to another
!
      do k = 1, ng

! prepare source and target indices
!
        ks = keu - k
        kt = kb  - k

! update variables
!
#ifdef GLM
        qp(1:nt,:,:,kt) = qp(1:nt,:,:,ks)
#else /* GLM */
        qp(1:nu,:,:,kt) = qp(1:nu,:,:,ks)
#ifdef VPOT
        bi(1:nb,:,:,kt) = bi(1:nb,:,:,ks)
#endif /* VPOT */
#endif /* GLM */

      end do

    end select

! === upper boundary ===
!
    select case(bnd_type(3,2))

#ifdef MPI
!-   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -
!
    case(bnd_exchange)

! "exchange" boundary conditions are used only between MPI nodes and use
! different schemes for periodic and other types of conditions
!
      if (periodic(3)) then

! this scheme sends at the same time data from even to odd nodes, and then in
! the second step, from odd to even nodes
!
        if (pparity(3) .eq. 0) then

! copy data to the exchange array
!
          qb( 1:nt,:,:,   :   ) = qp( 1:nt,:,:, kb:kbu)
#ifdef VPOT
          qb(nk:nl,:,:,   :   ) = bi( 1:nb,:,:, kb:kbu)
#endif /* VPOT */

! send data to and then receive new data from the left node
!
          call send_real_array   (nn, pneighs(3,1), 313, qb(:,:,:,:), iret)
          call receive_real_array(nn, pneighs(3,1), 314, qb(:,:,:,:), iret)

! update ghost zones
!
          qp( 1:nt,:,:,  1:kbl) = qb( 1:nt,:,:,   :   )
#ifdef VPOT
          bi( 1:nb,:,:,  1:kbl) = qb(nk:nl,:,:,   :   )
#endif /* VPOT */

        else

! receive data from the right node
!
          call receive_real_array(nn, pneighs(3,2), 313, qb(:,:,:,:), iret)

! update ghost zones
!
          qp( 1:nt,:,:,keu:km ) = qb( 1:nt,:,:,   :   )
#ifdef VPOT
          bi( 1:nb,:,:,keu:km ) = qb(nk:nl,:,:,   :   )
#endif /* VPOT */

! copy data to the exchange array
!
          qb( 1:nt,:,:,   :   ) = qp( 1:nt,:,:,kel:ke )
#ifdef VPOT
          qb(nk:nl,:,:,   :   ) = bi( 1:nb,:,:,kel:ke )
#endif /* VPOT */

! send data to the right node
!
          call send_real_array(nn, pneighs(3,2), 314, qb(:,:,:,:), iret)

        end if

      else

! copy data to the exchange array
!
        qb( 1:nt,:,:,   :   ) = qp( 1:nt,:,:,kel:ke )
#ifdef VPOT
        qb(nk:nl,:,:,   :   ) = bi( 1:nb,:,:,kel:ke )
#endif /* VPOT */

! send data to the right node
!
        call send_real_array   (nn, pneighs(3,2), 321, qb(:,:,:,:), iret)

! receive data from the right node
!
        call receive_real_array(nn, pneighs(3,2), 322, qb(:,:,:,:), iret)

! update ghost zones
!
        qp( 1:nt,:,:,keu:km ) = qb( 1:nt,:,:,   :   )
#ifdef VPOT
        bi( 1:nb,:,:,keu:km ) = qb(nk:nl,:,:,   :   )
#endif /* VPOT */

      end if
#endif /* MPI */

!-   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -
!
    case(bnd_open)

! "open" boundary conditions set the derivative normal to boundary to zero
!
      do k = 1, ng

! prepare indices
!
        kt   = ke + k
        ktm1 = kt - 1

! update the boundary
!
#ifdef GLM
        qp(1:nt,:,:,kt) = qp(1:nt,:,:,ktm1)
        qp( iph,:,:,kt) = 0.0d+00

! update normal component of magnetic field
!
        do j = 1, jm
          jm1 = max( 1, j - 1)
          jp1 = min(jm, j + 1)
          do i = 1, im
            im1 = max( 1, i - 1)
            ip1 = min(im, i + 1)

! calculate derivatives of transversal components of magnetic field
!
            dbx = dxih * (qp(ibx,ip1,j  ,ktm1) - qp(ibx,im1,j  ,ktm1))
            dby = dyih * (qp(iby,i  ,jp1,ktm1) - qp(iby,i  ,jm1,ktm1))

! update normal component of magnetic field from the divergence free condition
!
            qp(ibz,i,j,kt) = qp(ibz,i,j,ktm1) - dz * (dbx + dby)
          end do
        end do
#else /* GLM */
        qp(1:nu,:,:,kt) = qp(1:nu,:,:,ktm1)
#endif /* GLM */

      end do

!-   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -
!
    case(bnd_outflow)

! "outflow" boundary conditions are the same conditions as "open" but the normal
! component of velocity is set to zero in inflow locations
!
      do k = 1, ng

! prepare indices
!
        kt   = ke + k
        ktm1 = kt - 1

! update the boundary
!
#ifdef GLM
        qp(1:nt,:,:,kt) = qp(1:nt,:,:,ktm1)
        qp( iph,:,:,kt) = 0.0d+00

! update normal component of magnetic field
!
        do j = 1, jm
          jm1 = max( 1, j - 1)
          jp1 = min(jm, j + 1)
          do i = 1, im
            im1 = max( 1, i - 1)
            ip1 = min(im, i + 1)

! calculate derivatives of transversal components of magnetic field
!
            dbx = dxih * (qp(ibx,ip1,j  ,ktm1) - qp(ibx,im1,j  ,ktm1))
            dby = dyih * (qp(iby,i  ,jp1,ktm1) - qp(iby,i  ,jm1,ktm1))

! update normal component of magnetic field from the divergence free condition
!
            qp(ibz,i,j,kt) = qp(ibz,i,j,ktm1) - dz * (dbx + dby)
          end do
        end do
#else /* GLM */
        qp(1:nu,:,:,kt) = qp(1:nu,:,:,ktm1)
#endif /* GLM */

! limit the normal component of velocity
!
        qp(ivz,:,:,kt) = max(0.0d+00, qp(ivz,:,:,ktm1))

      end do

!-   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -
!
    case(bnd_reflective)

! "reflective" boundary conditions reverse layers inverting the normal component
! of velocity;
!
      do k = 1, ng

! prepare source and target indices
!
        ks   = keu - k
        kt   = ke  + k

! update the boundary
!
#ifdef GLM
        qp(1:nt,:,:,kt) = qp(1:nt,:,:,ks)
#else /* GLM */
        qp(1:nu,:,:,kt) = qp(1:nu,:,:,ks)
#endif /* GLM */

! change the sign of the normal component of velocity
!
        qp(ivz,:,:,kt) = - qp(ivz,:,:,ks)
#ifdef GLM
        qp(ibz,:,:,kt) = - qp(ibz,:,:,ks)
#endif /* GLM */

      end do

#ifdef BFIELD
!-   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -
!
    case(bnd_current_free)

! "current-free" boundary conditions are the same as "open", but the magnetic
! field is extrapolated to be free of current density
!
      do k = 1, ng

! prepare indices
!
        kt   = ke + k
        ktm1 = kt - 1

! update the boundary
!
#ifdef GLM
        qp(1:nt,:,:,kt) = qp(1:nt,:,:,ktm1)

! update normal component of magnetic field
!
        do j = 1, jm
          jm1 = max( 1, j - 1)
          jp1 = min(jm, j + 1)
          do i = 1, im
            im1 = max( 1, i - 1)
            ip1 = min(im, i + 1)

! calculate derivatives of transversal components of magnetic field
!
            dbx = dxih * (qp(ibx,ip1,j  ,ktm1) - qp(ibx,im1,j  ,ktm1))
            dby = dyih * (qp(iby,i  ,jp1,ktm1) - qp(iby,i  ,jm1,ktm1))

! update normal component of magnetic field from the divergence free condition
!
            qp(ibz,i,j,kt) = qp(ibz,i,j,ktm1) - dz * (dbx + dby)
          end do
        end do
#else /* GLM */
        qp(1:nu,:,:,kt) = qp(1:nu,:,:,ktm1)
#endif /* GLM */

      end do
#endif /* BFIELD */

!-   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -
!
    case default

! "periodic" boundary conditions simply copy one side to another
!
      do k = 1, ng

! prepare source and target indices
!
        ks = kbl + k
        kt = ke  + k

! update the boundary
!
#ifdef GLM
        qp(1:nt,:,:,kt) = qp(1:nt,:,:,ks)
#else /* GLM */
        qp(1:nu,:,:,kt) = qp(1:nu,:,:,ks)
#ifdef VPOT
        bi(1:nb,:,:,kt) = bi(1:nb,:,:,ks)
#endif /* VPOT */
#endif /* GLM */

      end do

    end select

!- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
!
! convert primitive variables at boundaries to conservative representation
!
    do j = 1, jm
      do k = 1, ng
        call prim2cons(im, qp(1:nt,1:im,j,k), u(1:nt,1:im))

#ifdef GLM
        uu(1:nt,1:im,j,k) = u(1:nt,1:im)
#else /* GLM */
        uu(1:nu,1:im,j,k) = u(1:nu,1:im)
#endif /* GLM */
      end do

      do k = keu, km
        call prim2cons(im, qp(1:nt,1:im,j,k), u(1:nt,1:im))

#ifdef GLM
        uu(1:nt,1:im,j,k) = u(1:nt,1:im)
#else /* GLM */
        uu(1:nu,1:im,j,k) = u(1:nu,1:im)
#endif /* GLM */
      end do
    end do

!-------------------------------------------------------------------------------
!
  end subroutine zboundary
#endif /* R3D */
#ifdef VPOT
!
!===============================================================================
!
! subroutine XBOUNDARY_MAGNETIC:
! -----------------------------
!
!   Subroutine updates the boundaries of vector potential components along
!   the X direction.
!
!   Arguments:
!
!     h - a time interval for the time varying boundary conditions
!
!===============================================================================
!
  subroutine xboundary_magnetic(h)

! include external subroutines and variables used this subroutine
!
    use mesh      , only : ng, im, jm, km, ib, ie, ibl, ibu, iel, ieu
    use mesh      , only : dx, dy, dz, dxi, dyi, dzi
#ifdef MPI
    use mpitools  , only : pcoords, pparity, pneighs, periodic
    use mpitools  , only : send_real_array, receive_real_array
#endif /* MPI */
    use variables , only : nb
    use variables , only : uu, qp_bnd
#ifdef MPI
    use variables , only : inx, iny, inz
#endif /* MPI */
    use variables , only : iax, iay, iaz

! local variables are not implicit by default
!
    implicit none

! subroutine arguments
!
    real, intent(in)             :: h

! local indices
!
    integer                      :: iret
    integer                      :: i, is, it, itm1, itp1
    integer                      :: ism1, ism2, itm2, itp2
    integer                      :: j, jm1, jp1, k, km1, kp1
    real                         :: day, daz

! variables saved between calls
!
    logical, save                :: first = .true.
#ifdef MPI
    integer, save                :: nn    = 1
#endif /* MPI */
    real   , save                :: dxyi  = 1.0d0, dxzi  = 1.0d0
    real   , save                :: dxyi2 = 1.0d0, dxzi2 = 1.0d0

#ifdef MPI
! an array for the MPI exchange
!
    real, dimension(nb,ng,jm,km) :: ub
#endif /* MPI */
!
!-------------------------------------------------------------------------------
!
! prepare some variables in the first execution
!
    if (first) then

! calculate some useful factors
!
      dxyi  = dx   * dyi
      dxzi  = dx   * dzi
      dxyi2 = dxyi * dxyi
      dxzi2 = dxzi * dxzi

#ifdef MPI
! calculate the size of array exchange
!
      nn    = nb * ng * jm * km
#endif /* MPI */

! change the flag first
!
      first = .false.

    end if

! === lower boundary ===
!
    select case(bnd_type(1,1))

#ifdef MPI
!-   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -
!
    case(bnd_exchange)

! "exchange" boundary conditions are used only between MPI nodes and use
! different schemes for periodic and other types of conditions
!
      if (periodic(1)) then

! this scheme sends at the same time data from even to odd nodes, and then in
! the second step, from odd to even nodes
!
        if (pparity(1) .eq. 0) then

! copy data to the exchange array
!
          ub(inx:inz,:,:,:) = uu(iax:iaz,iel:ie,:,:)

! send data to and then receive new data from the right node
!
          call send_real_array   (nn, pneighs(1,2), 111, ub(:,:,:,:), iret)
          call receive_real_array(nn, pneighs(1,2), 112, ub(:,:,:,:), iret)

! update ghost zones
!
          uu(iax:iaz,ieu:im,:,:) = ub(inx:inz,:,:,:)

        else

! receive data from the left node
!
          call receive_real_array(nn, pneighs(1,1), 111, ub(:,:,:,:), iret)

! update ghost zones
!
          uu(iax:iaz,1:ibl,:,:) = ub(inx:inz,:,:,:)

! copy data to the exchange array
!
          ub(inx:inz,:,:,:) = uu(iax:iaz,ib:ibu,:,:)

! send data to the left node
!
          call send_real_array   (nn, pneighs(1,1), 112, ub(:,:,:,:), iret)

        end if

      else

! receive data from the left node
!
        call receive_real_array(nn, pneighs(1,1), 121, ub(:,:,:,:), iret)

! update ghost zones
!
        uu(iax:iaz,1:ibl,:,:) = ub(inx:inz,:,:,:)

! copy data to the exchange array
!
        ub(inx:inz,:,:,:) = uu(iax:iaz,ib:ibu,:,:)

! send data to the left node
!
        call send_real_array   (nn, pneighs(1,1), 122, ub(:,:,:,:), iret)

      end if
#endif /* MPI */

!-   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -
!
    case(bnd_open, bnd_outflow)

! "open" boundary conditions set the derivative normal to boundary to zero
! "outflow" boundary conditions are the same conditions as "open" but the normal
! component of velocity is set to zero in inflow locations
!
      do i = 1, ng

! prepare indices
!
        it   = ib - i
        itp1 = it + 1

! extrapolate the X component of vector potential
!
        uu(iax,it,:,:) = uu(iax,itp1,:,:)

#ifndef CD
        if (i .gt. 1) then
#endif /* !CD */

! calculate additional index
!
          itp2 = it + 2

! extrapolate Y and Z components of vector potential
!
          uu(iay,it,:,:) = 2.0d0 * uu(iay,itp1,:,:) - uu(iay,itp2,:,:)
          uu(iaz,it,:,:) = 2.0d0 * uu(iaz,itp1,:,:) - uu(iaz,itp2,:,:)

#ifndef CD
        end if
#endif /* !CD */

      end do

!-   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -
!
    case(bnd_reflective)

! "reflective" boundary conditions reverse layers inverting the normal component
! of velocity;
!
      do i = 1, ng

! prepare the source and target indices
!
        is   = ibl + i
        it   = ib  - i
#ifdef CT
        ism1 = is  - 1
        ism2 = is  - 2
        itp1 = it  + 1
#endif /* CT */

#ifdef CT
        uu(iax,it,:,:) =   uu(iax,is  ,:,:)
        if (i .gt. 1) then
          uu(iay,it,:,:) = uu(iay,itp1,:,:)                                    &
                                       + (uu(iay,ism2,:,:) - uu(iay,ism1,:,:))
          uu(iaz,it,:,:) = uu(iaz,itp1,:,:)                                    &
                                       + (uu(iaz,ism2,:,:) - uu(iaz,ism1,:,:))
        end if
#endif /* CT */
#ifdef CD
        uu(iax,it,:,:) =   uu(iax,is  ,:,:)
        uu(iay,it,:,:) =   uu(iay,is  ,:,:)
        uu(iaz,it,:,:) =   uu(iaz,is  ,:,:)
#endif /* CD */

      end do

!-   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -
!
    case(bnd_current_free)

! "current-free" boundary conditions are the same as "open", but the magnetic
! field is extrapolated to be free of current density
!
      do i = 1, ng

! prepare indices
!
        it   = ib - i
        itp1 = it + 1

! update magnetic variables
!
! extrapolate Ax
!
        uu(iax,it,:,:) = uu(iax,itp1,:,:)

        if (i .gt. 1) then

! calculate additional index
!
          itp2 = it + 2

! extrapolate Ay and Az
!
          uu(iay,it,:,:) = 2.0d0 * uu(iay,itp1,:,:) - uu(iay,itp2,:,:)
          uu(iaz,it,:,:) = 2.0d0 * uu(iaz,itp1,:,:) - uu(iaz,itp2,:,:)

! calculate Ay and Az from nabla(A) = 0
!
          do k = 1, km
#ifdef R3D
            km1 = max( 1, k - 1)
            kp1 = min(km, k + 1)
#endif /* R3D */
            do j = 1, jm
              jm1 = max( 1, j - 1)
              jp1 = min(jm, j + 1)

#ifndef R2D
              day = uu(iay,itp1,jp1,k) - 2.0d0 * uu(iay,itp1,j,k)              &
                                                          + uu(iay,itp1,jm1,k)
#ifdef R3D
              daz = uu(iay,itp1,j,kp1) - 2.0d0 * uu(iay,itp1,j,k)              &
                                                          + uu(iay,itp1,j,km1)
#endif /* R3D */

#ifdef R3D
              uu(iay,it,j,k) = 2.0d0 * uu(iay,itp1,j,k) - uu(iay,itp2,j,k)     &
                                                   - day * dxyi2 - daz * dxzi2
#else /* R3D */
              uu(iay,it,j,k) = 2.0d0 * uu(iay,itp1,j,k) - uu(iay,itp2,j,k)     &
                                                   - day * dxyi2
#endif /* R3D */
#endif /* !R2D */

              day = uu(iaz,itp1,jp1,k) - 2.0d0 * uu(iaz,itp1,j,k)              &
                                                          + uu(iaz,itp1,jm1,k)
#ifdef R3D
              daz = uu(iaz,itp1,j,kp1) - 2.0d0 * uu(iaz,itp1,j,k)              &
                                                          + uu(iaz,itp1,j,km1)
#endif /* R3D */

#ifdef R3D
              uu(iaz,it,j,k) = 2.0d0 * uu(iaz,itp1,j,k) - uu(iaz,itp2,j,k)     &
                                                   - day * dxyi2 - daz * dxzi2
#else /* R3D */
              uu(iaz,it,j,k) = 2.0d0 * uu(iaz,itp1,j,k) - uu(iaz,itp2,j,k)     &
                                                   - day * dxyi2
#endif /* R3D */
            end do
          end do

        end if
      end do

!-   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -
!
    case default

! "periodic" boundary conditions simply copy one side to another
!
      do i = 1, ng

! prepare source and target indices
!
        is = ieu - i
        it = ib  - i

! copy source layer to target
!
        uu(iax:iaz,it,:,:) = uu(iax:iaz,is,:,:)

      end do

    end select

! === upper boundary ===
!
    select case(bnd_type(1,2))

#ifdef MPI
!-   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -
!
    case(bnd_exchange)

! "exchange" boundary conditions are used only between MPI nodes and use
! different schemes for periodic and other types of conditions
!
      if (periodic(1)) then

! this scheme sends at the same time data from even to odd nodes, and then in
! the second step, from odd to even nodes
!
        if (pparity(1) .eq. 0) then

! copy data to the exchange array
!
          ub(inx:inz,:,:,:) = uu(iax:iaz,ib:ibu,:,:)

! send data to and then receive new data from the left node
!
          call send_real_array   (nn, pneighs(1,1), 113, ub(:,:,:,:), iret)
          call receive_real_array(nn, pneighs(1,1), 114, ub(:,:,:,:), iret)

! update ghost zones
!
          uu(iax:iaz,1:ibl,:,:) = ub(inx:inz,:,:,:)

        else

! receive data from the right node
!
          call receive_real_array(nn, pneighs(1,2), 113, ub(:,:,:,:), iret)

! update ghost zones
!
          uu(iax:iaz,ieu:im,:,:) = ub(inx:inz,:,:,:)

! copy data to the exchange array
!
          ub(inx:inz,:,:,:) = uu(iax:iaz,iel:ie,:,:)

! send data to the right node
!
          call send_real_array   (nn, pneighs(1,2), 114, ub(:,:,:,:), iret)

        end if

      else

! copy data to the exchange array
!
        ub(inx:inz,:,:,:) = uu(iax:iaz,iel:ie,:,:)

! send data to the right node
!
        call send_real_array   (nn, pneighs(1,2), 121, ub(:,:,:,:), iret)

! receive data from the right node
!
        call receive_real_array(nn, pneighs(1,2), 122, ub(:,:,:,:), iret)

! update ghost zones
!
        uu(iax:iaz,ieu:im,:,:) = ub(inx:inz,:,:,:)

      end if
#endif /* MPI */

!-   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -
!
    case(bnd_open, bnd_outflow)

! "open" boundary conditions set the derivative normal to boundary to zero
! "outflow" boundary conditions are the same conditions as "open" but the normal
! component of velocity is set to zero in inflow locations
!
      do i = 1, ng

! prepare indices
!
        it   = ie + i
        itm1 = it - 1
        itm2 = it - 2

! exptrapolate vector potential components
!
        uu(iax,it,:,:) =         uu(iax,itm1,:,:)
        uu(iay,it,:,:) = 2.0d0 * uu(iay,itm1,:,:) - uu(iay,itm2,:,:)
        uu(iaz,it,:,:) = 2.0d0 * uu(iaz,itm1,:,:) - uu(iaz,itm2,:,:)

      end do

!-   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -
!
    case(bnd_reflective)

! "reflective" boundary conditions reverse layers inverting the normal component
! of velocity;
!
      do i = 1, ng

! prepare source and target indices
!
        is   = ieu - i
        it   = ie  + i
#ifdef CT
        ism1 = is  - 1
        itm1 = it  - 1
#endif /* CT */

#ifdef CT
        uu(iax,it,:,:) =   uu(iax,is  ,:,:)
        uu(iay,it,:,:) =   uu(iay,itm1,:,:)                                    &
                                         + (uu(iay,is,:,:) - uu(iay,ism1,:,:))
        uu(iaz,it,:,:) =   uu(iaz,itm1,:,:)                                    &
                                         + (uu(iaz,is,:,:) - uu(iaz,ism1,:,:))
#endif /* CT */
#ifdef CD
        uu(iax,it,:,:) =   uu(iax,is  ,:,:)
        uu(iay,it,:,:) =   uu(iay,is  ,:,:)
        uu(iaz,it,:,:) =   uu(iaz,is  ,:,:)
#endif /* CD */

      end do

!-   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -
!
    case(bnd_current_free)

! "current-free" boundary conditions are the same as "open", but the magnetic
! field is extrapolated to be free of current density
!
      do i = 1, ng

! prepare indices
!
        it   = ie + i
        itm1 = it - 1

! update fluid variables
!
        uu(  :,it,:,:) = uu(  :,itm1,:,:)

! update magnetic variables
!
! calculate additional index
!
        itm2 = it - 2

! exptrapolate magnetic potential
!
        uu(iax,it,:,:) =         uu(iax,itm1,:,:)
        uu(iay,it,:,:) = 2.0d0 * uu(iay,itm1,:,:) - uu(iay,itm2,:,:)
        uu(iaz,it,:,:) = 2.0d0 * uu(iaz,itm1,:,:) - uu(iaz,itm2,:,:)

! calculate By and Bz from nabla(A)=0
!
        do k = 1, km
#ifdef R3D
          km1 = max( 1, k - 1)
          kp1 = min(km, k + 1)
#endif /* R3D */
          do j = 1, jm
            jm1 = max( 1, j - 1)
            jp1 = min(jm, j + 1)

#ifndef R2D
            day = uu(iay,itm1,jp1,k) - 2.0d0 * uu(iay,itm1,j,k)                &
                                                          + uu(iay,itm1,jm1,k)
#ifdef R3D
            daz = uu(iay,itm1,j,kp1) - 2.0d0 * uu(iay,itm1,j,k)                &
                                                          + uu(iay,itm1,j,km1)
#endif /* R3D */

#ifdef R3D
            uu(iay,it,j,k) = 2.0d0 * uu(iay,itm1,j,k) - uu(iay,itm2,j,k)       &
                                                   - day * dxyi2 - daz * dxzi2
#else /* R3D */
            uu(iay,it,j,k) = 2.0d0 * uu(iay,itm1,j,k) - uu(iay,itm2,j,k)       &
                                                   - day * dxyi2
#endif /* R3D */
#endif /* !R2D */

            day = uu(iaz,itm1,jp1,k) - 2.0d0 * uu(iaz,itm1,j,k)                &
                                                          + uu(iaz,itm1,jm1,k)
#ifdef R3D
            daz = uu(iaz,itm1,j,kp1) - 2.0d0 * uu(iaz,itm1,j,k)                &
                                                          + uu(iaz,itm1,j,km1)
#endif /* R3D */

#ifdef R3D
            uu(iaz,it,j,k) = 2.0d0 * uu(iaz,itm1,j,k) - uu(iaz,itm2,j,k)       &
                                                   - day * dxyi2 - daz * dxzi2
#else /* R3D */
            uu(iaz,it,j,k) = 2.0d0 * uu(iaz,itm1,j,k) - uu(iaz,itm2,j,k)       &
                                                   - day * dxyi2
#endif /* R3D */
          end do
        end do
      end do

!-   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -
!
    case default

! "periodic" boundary conditions simply copy one side to another
!
      do i = 1, ng

! prepare source and target indices
!
        is = ibl + i
        it = ie  + i

! copy source layer to target
!
        uu(iax:iaz,it,:,:) = uu(iax:iaz,is,:,:)

      end do

    end select

!-------------------------------------------------------------------------------
!
  end subroutine xboundary_magnetic
!
!===============================================================================
!
! subroutine YBOUNDARY_MAGNETIC:
! -----------------------------
!
!   Subroutine updates the boundaries of vector potential components along
!   the Y direction.
!
!   Arguments:
!
!     h - a time interval for the time varying boundary conditions
!
!===============================================================================
!
  subroutine yboundary_magnetic(h)

! include external subroutines and variables used this subroutine
!
    use mesh      , only : ng, im, jm, km, jb, je, jbl, jbu, jel, jeu
    use mesh      , only : dx, dy, dz, dxi, dyi, dzi
#ifdef MPI
    use mpitools  , only : pcoords, pparity, pneighs, periodic
    use mpitools  , only : send_real_array, receive_real_array
#endif /* MPI */
    use variables , only : nb
    use variables , only : uu, qp_bnd
#ifdef MPI
    use variables , only : inx, iny, inz
#endif /* MPI */
    use variables , only : iax, iay, iaz

! local variables are not implicit by default
!
    implicit none

! subroutine arguments
!
    real, intent(in)             :: h

! local variables
!
    integer                      :: iret
    integer                      :: j, js, jt, jtm1, jtp1
    integer                      :: jsm1, jsm2, jtm2, jtp2
    integer                      :: i, im1, ip1, k, km1, kp1
    real                         :: dax, daz

! variables saved between calls
!
    logical, save                :: first = .true.
#ifdef MPI
    integer, save                :: nn    = 1
#endif /* MPI */
    real   , save                :: dyxi  = 1.0, dyzi  = 1.0
    real   , save                :: dyxi2 = 1.0, dyzi2 = 1.0

#ifdef MPI
! array for the MPI exchange
!
    real, dimension(nb,im,ng,km) :: ub
#endif /* MPI */
!
!-------------------------------------------------------------------------------
!
! prepare some variables in the first execution
!
    if (first) then

! calculate some useful factors
!
      dyxi  = dy   * dxi
      dyzi  = dy   * dzi
      dyxi2 = dyxi * dyxi
      dyzi2 = dyzi * dyzi

#ifdef MPI
! calculate the size of array exchange
!
      nn    = nb * im * ng * km
#endif /* MPI */

! change the flag first
!
      first = .false.

    end if

! === lower boundary ===
!
    select case(bnd_type(2,1))

#ifdef MPI
!-   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -
!
    case(bnd_exchange)

! "exchange" boundary conditions are used only between MPI nodes and use
! different schemes for periodic and other types of conditions
!
      if (periodic(2)) then

! this scheme sends at the same time data from even to odd nodes, and then in
! the second step, from odd to even nodes
!
        if (pparity(2) .eq. 0) then

! copy data to the exchange array
!
          ub(inx:inz,:,:,:) = uu(iax:iaz,:,jel:je,:)

! send data to and then receive new data from the right node
!
          call send_real_array   (nn, pneighs(2,2), 211, ub(:,:,:,:), iret)
          call receive_real_array(nn, pneighs(2,2), 212, ub(:,:,:,:), iret)

! update ghost zones
!
          uu(iax:iaz,:,jeu:jm,:) = ub(inx:inz,:,:,:)

        else

! receive data from the left node
!
          call receive_real_array(nn, pneighs(2,1), 211, ub(:,:,:,:), iret)

! update ghost zones
!
          uu(iax:iaz,:,1:jbl,:) = ub(inx:inz,:,:,:)

! copy data to the exchange array
!
          ub(inx:inz,:,:,:) = uu(iax:iaz,:,jb:jbu,:)

! send data to the left node
!
          call send_real_array   (nn, pneighs(2,1), 212, ub(:,:,:,:), iret)

        end if

      else

! receive data from the left node
!
        call receive_real_array(nn, pneighs(2,1), 221, ub(:,:,:,:), iret)

! update ghost zones
!
        uu(iax:iaz,:,1:jbl,:) = ub(inx:inz,:,:,:)

! copy data to the exchange array
!
        ub(inx:inz,:,:,:) = uu(iax:iaz,:,jb:jbu,:)

! send data to the left node
!
        call send_real_array   (nn, pneighs(2,1), 222, ub(:,:,:,:), iret)

      end if
#endif /* MPI */

!-   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -
!
    case(bnd_open, bnd_outflow)

! "open" boundary conditions set the derivative normal to boundary to zero
! "outflow" boundary conditions are the same conditions as "open" but the normal
! component of velocity is set to zero in inflow locations
!
      do j = 1, ng

! prepare indices
!
        jt   = jb - j
        jtp1 = jt + 1

! extrapolate the Y component of vector potential
!
        uu(iay,:,jt,:) = uu(iay,:,jtp1,:)

#ifndef CD
        if (j .gt. 1) then
#endif /* !CD */

! calculate additional index
!
          jtp2 = jt + 2

! extrapolate X and Z components of vector potential
!
          uu(iax,:,jt,:) = 2.0d0 * uu(iax,:,jtp1,:) - uu(iax,:,jtp2,:)
          uu(iaz,:,jt,:) = 2.0d0 * uu(iaz,:,jtp1,:) - uu(iaz,:,jtp2,:)

#ifndef CD
        end if
#endif /* !CD */

      end do

!-   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -
!
    case(bnd_reflective)

! "reflective" boundary conditions reverse layers inverting the normal component
! of velocity;
!
      do j = 1, ng

! prepare the source and target indices
!
        js   = jbl + j
        jt   = jb  - j
#ifdef CT
        jsm1 = js  - 1
        jsm2 = js  - 2
        jtp1 = jt  + 1
#endif /* CT */

! copy source layer to target
!
#ifdef CT
        uu(iay,:,jt,:) =   uu(iay,:,js  ,:)
        if (j .gt. 1) then
          uu(iax,:,jt,:) = uu(iax,:,jtp1,:)                                    &
                                       + (uu(iax,:,jsm2,:) - uu(iax,:,jsm1,:))
          uu(iaz,:,jt,:) = uu(iaz,:,jtp1,:)                                    &
                                       + (uu(iaz,:,jsm2,:) - uu(iaz,:,jsm1,:))
        end if
#endif /* CT */
#ifdef CD
        uu(iax,:,jt,:) =   uu(iax,:,js  ,:)
        uu(iay,:,jt,:) =   uu(iay,:,js  ,:)
        uu(iaz,:,jt,:) =   uu(iaz,:,js  ,:)
#endif /* CD */

      end do

!-   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -
!
    case(bnd_current_free)

! "current-free" boundary conditions are the same as "open", but the magnetic
! field is extrapolated to be free of current density
!
      do j = 1, ng

! prepare indices
!
        jt   = jb - j
        jtp1 = jt + 1

! update magnetic variables
!
! extrapolate Ay
!
        uu(iay,:,jt,:) = uu(iay,:,jtp1,:)

        if (j .gt. 1) then

! calculate additional index
!
          jtp2 = jt + 2

! extrapolate Ax and Az
!
          uu(iax,:,jt,:) = 2.0d0 * uu(iax,:,jtp1,:) - uu(iax,:,jtp2,:)
          uu(iaz,:,jt,:) = 2.0d0 * uu(iaz,:,jtp1,:) - uu(iaz,:,jtp2,:)

! update Ax and Az from nabla(A) = 0
!
          do k = 1, km
#ifdef R3D
            km1 = max( 1, k - 1)
            kp1 = min(km, k + 1)
#endif /* R3D */
            do i = 1, im
              im1 = max( 1, i - 1)
              ip1 = min(im, i + 1)

#ifndef R2D
              dax = uu(iax,ip1,jtp1,k) - 2.0d0 * uu(iax,i,jtp1,k)              &
                                                          + uu(iax,im1,jtp1,k)
#ifdef R3D
              daz = uu(iax,i,jtp1,kp1) - 2.0d0 * uu(iax,i,jtp1,k)              &
                                                          + uu(iax,i,jtp1,km1)
#endif /* R3D */

#ifdef R3D
              uu(iax,i,jt,k) = 2.0d0 * uu(iax,i,jtp1,k) - uu(iax,i,jtp2,k)     &
                                                   - dax * dyxi2 - daz * dyzi2
#else /* R3D */
              uu(iax,i,jt,k) = 2.0d0 * uu(iax,i,jtp1,k) - uu(iax,i,jtp2,k)     &
                                                   - dax * dyxi2
#endif /* R3D */
#endif /* !R2D */

              dax = uu(iaz,ip1,jtp1,k) - 2.0d0 * uu(iaz,i,jtp1,k)              &
                                                          + uu(iaz,im1,jtp1,k)
#ifdef R3D
              daz = uu(iaz,i,jtp1,kp1) - 2.0d0 * uu(iaz,i,jtp1,k)              &
                                                          + uu(iaz,i,jtp1,km1)
#endif /* R3D */

#ifdef R3D
              uu(iaz,i,jt,k) = 2.0d0 * uu(iaz,i,jtp1,k) - uu(iaz,i,jtp2,k)     &
                                                   - dax * dyxi2 - daz * dyzi2
#else /* R3D */
              uu(iaz,i,jt,k) = 2.0d0 * uu(iaz,i,jtp1,k) - uu(iaz,i,jtp2,k)     &
                                                   - dax * dyxi2
#endif /* R3D */
            end do
          end do

        end if
      end do

!-   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -
!
    case default

! "periodic" boundary conditions simply copy one side to another
!
      do j = 1, ng

! prepare source and target indices
!
        js = jeu - j
        jt = jb  - j

! copy source layer to target
!
        uu(iax:iaz,:,jt,:) = uu(iax:iaz,:,js,:)

      end do

    end select

! === upper boundary ===
!
    select case(bnd_type(2,2))

#ifdef MPI
!-   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -
!
    case(bnd_exchange)

! "exchange" boundary conditions are used only between MPI nodes and use
! different schemes for periodic and other types of conditions
!
      if (periodic(2)) then

! this scheme sends at the same time data from even to odd nodes, and then in
! the second step, from odd to even nodes
!
        if (pparity(2) .eq. 0) then

! copy data to the exchange array
!
          ub(inx:inz,:,:,:) = uu(iax:iaz,:,jb:jbu,:)

! send data to and then receive new data from the left node
!
          call send_real_array   (nn, pneighs(2,1), 213, ub(:,:,:,:), iret)
          call receive_real_array(nn, pneighs(2,1), 214, ub(:,:,:,:), iret)

! update ghost zones
!
          uu(iax:iaz,:,1:jbl,:) = ub(inx:inz,:,:,:)

        else

! receive data from the right node
!
          call receive_real_array(nn, pneighs(2,2), 213, ub(:,:,:,:), iret)

! update ghost zones
!
          uu(iax:iaz,:,jeu:jm,:) = ub(inx:inz,:,:,:)

! copy data to the exchange array
!
          ub(inx:inz,:,:,:) = uu(iax:iaz,:,jel:je,:)

! send data to the right node
!
          call send_real_array   (nn, pneighs(2,2), 214, ub(:,:,:,:), iret)

        end if

      else

! copy data to the exchange array
!
        ub(inx:inz,:,:,:) = uu(iax:iaz,:,jel:je,:)

! send data to the right node
!
        call send_real_array   (nn, pneighs(2,2), 221, ub(:,:,:,:), iret)

! receive data from the right node
!
        call receive_real_array(nn, pneighs(2,2), 222, ub(:,:,:,:), iret)

! update ghost zones
!
        uu(iax:iaz,:,jeu:jm,:) = ub(inx:inz,:,:,:)

      end if
#endif /* MPI */

!-   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -
!
    case(bnd_open, bnd_outflow)

! "open" boundary conditions set the derivative normal to boundary to zero
! "outflow" boundary conditions are the same conditions as "open" but the normal
! component of velocity is set to zero in inflow locations
!
      do j = 1, ng

! prepare indices
!
        jt   = je + j
        jtm1 = jt - 1
        jtm2 = jt - 2

! exptrapolate vector potential components
!
        uu(iay,:,jt,:) =         uu(iay,:,jtm1,:)
        uu(iax,:,jt,:) = 2.0d0 * uu(iax,:,jtm1,:) - uu(iax,:,jtm2,:)
        uu(iaz,:,jt,:) = 2.0d0 * uu(iaz,:,jtm1,:) - uu(iaz,:,jtm2,:)

      end do

!-   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -
!
    case(bnd_reflective)

! "reflective" boundary conditions reverse layers inverting the normal component
! of velocity;
!
      do j = 1, ng

! prepare source and target indices
!
        js   = jeu - j
        jt   = je  + j
#ifdef CT
        jsm1 = js  - 1
        jtm1 = jt  - 1
#endif /* CT */

! copy source layer to target
!
#ifdef CT
        uu(iax,:,jt,:) =   uu(iax,:,jtm1,:)                                    &
                                         + (uu(iax,:,js,:) - uu(iax,:,jsm1,:))
        uu(iay,:,jt,:) =   uu(iay,:,js  ,:)
        uu(iaz,:,jt,:) =   uu(iaz,:,jtm1,:)                                    &
                                         + (uu(iaz,:,js,:) - uu(iaz,:,jsm1,:))
#endif /* CT */
#ifdef CD
        uu(iax,:,jt,:) =   uu(iax,:,js  ,:)
        uu(iay,:,jt,:) =   uu(iay,:,js  ,:)
        uu(iaz,:,jt,:) =   uu(iaz,:,js  ,:)
#endif /* CD */

      end do

!-   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -
!
    case(bnd_current_free)

! "current-free" boundary conditions are the same as "open", but the magnetic
! field is extrapolated to be free of current density
!
      do j = 1, ng

! prepare indices
!
        jt   = je + j
        jtm1 = jt - 1
        jtm2 = jt - 2

! exptrapolate magnetic potential
!
        uu(iay,:,jt,:) =         uu(iay,:,jtm1,:)
        uu(iax,:,jt,:) = 2.0d0 * uu(iax,:,jtm1,:) - uu(iax,:,jtm2,:)
        uu(iaz,:,jt,:) = 2.0d0 * uu(iaz,:,jtm1,:) - uu(iaz,:,jtm2,:)

! update Ax and Az from nabla(A) = 0
!
        do k = 1, km
#ifdef R3D
          km1 = max( 1, k - 1)
          kp1 = min(km, k + 1)
#endif /* R3D */
          do i = 1, im
            im1 = max( 1, i - 1)
            ip1 = min(im, i + 1)

#ifndef R2D
            dax = uu(iax,ip1,jtm1,k) - 2.0d0 * uu(iax,i,jtm1,k)                &
                                                          + uu(iax,im1,jtm1,k)
#ifdef R3D
            daz = uu(iax,i,jtm1,kp1) - 2.0d0 * uu(iax,i,jtm1,k)                &
                                                          + uu(iax,i,jtm1,km1)
#endif /* R3D */

#ifdef R3D
            uu(iax,i,jt,k) = 2.0d0 * uu(iax,i,jtm1,k) - uu(iax,i,jtm2,k)       &
                                                   - dax * dyxi2 - daz * dyzi2
#else /* R3D */
            uu(iax,i,jt,k) = 2.0d0 * uu(iax,i,jtm1,k) - uu(iax,i,jtm2,k)       &
                                                   - dax * dyxi2
#endif /* R3D */
#endif /* !R2D */

            dax = uu(iaz,ip1,jtm1,k) - 2.0d0 * uu(iaz,i,jtm1,k)                &
                                                          + uu(iaz,im1,jtm1,k)
#ifdef R3D
            daz = uu(iaz,i,jtm1,kp1) - 2.0d0 * uu(iaz,i,jtm1,k)                &
                                                          + uu(iaz,i,jtm1,km1)
#endif /* R3D */

#ifdef R3D
            uu(iaz,i,jt,k) = 2.0d0 * uu(iaz,i,jtm1,k) - uu(iaz,i,jtm2,k)       &
                                                   - dax * dyxi2 - daz * dyxi2
#else /* R3D */
            uu(iaz,i,jt,k) = 2.0d0 * uu(iaz,i,jtm1,k) - uu(iaz,i,jtm2,k)       &
                                                   - dax * dyxi2
#endif /* R3D */
          end do
        end do
      end do

!-   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -
!
    case default

! "periodic" boundary conditions simply copy one side to another
!
      do j = 1, ng

! prepare source and target indices
!
        js = jbl + j
        jt = je  + j

! copy source layer to target
!
        uu(iax:iaz,:,jt,:) = uu(iax:iaz,:,js,:)

      end do

    end select

!-------------------------------------------------------------------------------
!
  end subroutine yboundary_magnetic
#ifdef R3D
!
!===============================================================================
!
! subroutine ZBOUNDARY_MAGNETIC:
! -----------------------------
!
!   Subroutine updates the boundaries of vector potential along the Z direction.
!
!   Arguments:
!
!     h - a time interval for the time varying boundary conditions
!
!===============================================================================
!
  subroutine zboundary_magnetic(h)

! include external subroutines and variables used this subroutine
!
    use mesh      , only : ng, im, jm, km, kb, ke, kbl, kbu, kel, keu
    use mesh      , only : dx, dy, dz, dxi, dyi, dzi
#ifdef MPI
    use mpitools  , only : pcoords, pparity, pneighs, periodic
    use mpitools  , only : send_real_array, receive_real_array
#endif /* MPI */
    use variables , only : nb
    use variables , only : uu, qp_bnd
    use variables , only : idn, imx, imy, imz
#ifdef ADI
    use variables , only : ien
#endif /* ADI */
#ifdef MPI
    use variables , only : inx, iny, inz
#endif /* MPI */
    use variables , only : iax, iay, iaz

! local variables are not implicit by default
!
    implicit none

! subroutine arguments
!
    real, intent(in)             :: h

! local variables
!
    integer                      :: iret
    integer                      :: k, ks, kt, ktm1, ktp1
    integer                      :: ksm1, ksm2, ktm2, ktp2
    integer                      :: i, im1, ip1, j, jm1, jp1
    real                         :: dax, day

! variables saved between calls
!
    logical, save                :: first = .true.
#ifdef MPI
    integer, save                :: nn    = 1
#endif /* MPI */
    real   , save                :: dzxi  = 1.0, dzyi  = 1.0
    real   , save                :: dzxi2 = 1.0, dzyi2 = 1.0

#ifdef MPI
! array for the MPI exchange
!
    real, dimension(nb,im,jm,ng) :: ub
#endif /* MPI */
!
!-------------------------------------------------------------------------------
!
! prepare some variables in the first execution
!
    if (first) then

! calculate some useful factors
!
      dzxi  = dz * dxi
      dzyi  = dz * dyi
      dzxi2 = dzxi * dzxi
      dzyi2 = dzyi * dzyi

#ifdef MPI
! calculate the size of array exchange
!
      nn    = nb * im * jm * ng
#endif /* MPI */

! change the flag first
!
      first = .false.

    end if

! === lower boundary ===
!
    select case(bnd_type(3,1))

#ifdef MPI
!-   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -
!
    case(bnd_exchange)

! "exchange" boundary conditions are used only between MPI nodes and use
! different schemes for periodic and other types of conditions
!
      if (periodic(3)) then

! this scheme sends at the same time data from even to odd nodes, and then in
! the second step, from odd to even nodes
!
        if (pparity(3) .eq. 0) then

! copy data to the exchange array
!
          ub(inx:inz,:,:,:) = uu(iax:iaz,:,:,kel:ke)

! send data to and then receive new data from the right node
!
          call send_real_array   (nn, pneighs(3,2), 311, ub(:,:,:,:), iret)
          call receive_real_array(nn, pneighs(3,2), 312, ub(:,:,:,:), iret)

! update ghost zones
!
          uu(iax:iaz,:,:,keu:km) = ub(inx:inz,:,:,:)

        else

! receive data from the left node
!
          call receive_real_array(nn, pneighs(3,1), 311, ub(:,:,:,:), iret)

! update ghost zones
!
          uu(iax:iaz,:,:,1:kbl) = ub(inx:inz,:,:,:)

! copy data to the exchange array
!
          ub(inx:inz,:,:,:) = uu(iax:iaz,:,:,kb:kbu)

! send data to the left node
!
          call send_real_array(nn, pneighs(3,1), 312, ub(:,:,:,:), iret)

        end if

      else

! receive data from the left node
!
        call receive_real_array(nn, pneighs(3,1), 321, ub(:,:,:,:), iret)

! update ghost zones
!
        uu(iax:iaz,:,:,1:kbl) = ub(inx:inz,:,:,:)

! copy data to the exchange array
!
        ub(inx:inz,:,:,:) = uu(iax:iaz,:,:,kb:kbu)

! send data to the left node
!
        call send_real_array   (nn, pneighs(3,1), 322, ub(:,:,:,:), iret)

      end if
#endif /* MPI */

!-   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -
!
    case(bnd_open, bnd_outflow)

! "open" boundary conditions set the derivative normal to boundary to zero
! "outflow" boundary conditions are the same conditions as "open" but the normal
! component of velocity is set to zero in inflow locations
!
      do k = 1, ng

! prepare indices
!
        kt   = ke + k
        ktp1 = kt + 1

! extrapolate the Z component of vector potential
!
        uu(iaz,:,:,kt) = uu(iaz,:,:,ktp1)

#ifndef CD
        if (k .gt. 1) then
#endif /* !CD */

! calculate additional index
!
          ktp2 = kt + 2

! extrapolate X and Y components of vector potential
!
          uu(iax,:,:,kt) = 2.0d0 * uu(iax,:,:,ktp1) - uu(iax,:,:,ktp2)
          uu(iay,:,:,kt) = 2.0d0 * uu(iay,:,:,ktp1) - uu(iay,:,:,ktp2)

#ifndef CD
        end if
#endif /* !CD */

      end do

!-   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -
!
    case(bnd_reflective)

! "reflective" boundary conditions reverse layers inverting the normal component
! of velocity;
!
      do k = 1, ng

! prepare the source and target indices
!
        ks   = kbl + k
        kt   = kb  - k
#ifdef CT
        ksm1 = ks  - 1
        ksm2 = ks  - 2
        ktp1 = kt  + 1
#endif /* CT */

! copy source layer to target
!
#ifdef CT
        if (k .gt. 1) then
          uu(iax,:,:,kt) = uu(iax,:,:,ktp1)                                    &
                                       + (uu(iax,:,:,ksm2) - uu(iax,:,:,ksm1))
          uu(iay,:,:,kt) = uu(iay,:,:,ktp1)                                    &
                                       + (uu(iay,:,:,ksm2) - uu(iay,:,:,ksm1))
        end if
        uu(iaz,:,:,kt) =   uu(iaz,:,:,ks  )
#endif /* CT */
#ifdef CD
        uu(iax,:,:,kt) =   uu(iax,:,:,ks  )
        uu(iay,:,:,kt) =   uu(iay,:,:,ks  )
        uu(iaz,:,:,kt) =   uu(iaz,:,:,ks  )
#endif /* CD */

      end do

!-   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -
!
    case(bnd_current_free)

! "current-free" boundary conditions are the same as "open", but the magnetic
! field is extrapolated to be free of current density
!
      do k = 1, ng

! prepare indices
!
        kt   = kb - k
        ktp1 = kt + 1

! update magnetic variables
!
! extrapolate Az
!
        uu(iaz,:,:,kt) = uu(iaz,:,:,ktp1)

        if (k .gt. 1) then

! calculate additional index
!
          ktp2 = kt + 2

! extrapolate Ax and Ay
!
          uu(iax,:,:,kt) = 2.0d0 * uu(iax,:,:,ktp1) - uu(iax,:,:,ktp2)
          uu(iay,:,:,kt) = 2.0d0 * uu(iay,:,:,ktp1) - uu(iay,:,:,ktp2)

! update Ax and Ay from nabla(A) = 0
!
          do j = 1, jm
            jm1 = max( 1, j - 1)
            jp1 = min(jm, j + 1)
            do i = 1, im
              im1 = max( 1, i - 1)
              ip1 = min(im, i + 1)

              dax = uu(iax,ip1,j,ktp1) - 2.0d0 * uu(iax,i,j,ktp1)              &
                                                          + uu(iax,im1,j,ktp1)
              day = uu(iax,i,jp1,ktp1) - 2.0d0 * uu(iax,i,j,ktp1)              &
                                                          + uu(iax,i,jm1,ktp1)

              uu(iax,i,j,kt) = 2.0d0 * uu(iax,i,j,ktp1) - uu(iax,i,j,ktp2)     &
                                                   - dax * dzxi2 - day * dzyi2

              dax = uu(iay,ip1,j,ktp1) - 2.0d0 * uu(iay,i,j,ktp1)              &
                                                          + uu(iay,im1,j,ktp1)
              day = uu(iay,i,jp1,ktp1) - 2.0d0 * uu(iay,i,j,ktp1)              &
                                                          + uu(iay,i,jm1,ktp1)

              uu(iay,i,j,kt) = 2.0d0 * uu(iay,i,j,ktp1) - uu(iay,i,j,ktp2)     &
                                                   - dax * dzxi2 - day * dzyi2
            end do
          end do

        end if
      end do

!-   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -
!
    case default

! "periodic" boundary conditions simply copy one side to another
!
      do k = 1, ng

! prepare source and target indices
!
        ks = keu - k
        kt = kb  - k

! copy source layer to target
!
        uu(iax:iaz,:,:,kt) = uu(iax:iaz,:,:,ks)

      end do

    end select

! === upper boundary ===
!
    select case(bnd_type(3,2))

#ifdef MPI
!-   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -
!
    case(bnd_exchange)

! "exchange" boundary conditions are used only between MPI nodes and use
! different schemes for periodic and other types of conditions
!
      if (periodic(3)) then

! this scheme sends at the same time data from even to odd nodes, and then in
! the second step, from odd to even nodes
!
        if (pparity(3) .eq. 0) then

! copy data to the exchange array
!
          ub(inx:inz,:,:,:) = uu(iax:iaz,:,:,kb:kbu)

! send data to and then receive new data from the left node
!
          call send_real_array   (nn, pneighs(3,1), 313, ub(:,:,:,:), iret)
          call receive_real_array(nn, pneighs(3,1), 314, ub(:,:,:,:), iret)

! update ghost zones
!
          uu(iax:iaz,:,:,1:kbl) = ub(inx:inz,:,:,:)

        else

! receive data from the right node
!
          call receive_real_array(nn, pneighs(3,2), 313, ub(:,:,:,:), iret)

! update ghost zones
!
          uu(iax:iaz,:,:,keu:km) = ub(inx:inz,:,:,:)

! copy data to the exchange array
!
          ub(inx:inz,:,:,:) = uu(iax:iaz,:,:,kel:ke)

! send data to the right node
!
          call send_real_array(nn, pneighs(3,2), 314, ub(:,:,:,:), iret)

        end if

      else

! copy data to the exchange array
!
        ub(inx:inz,:,:,:) = uu(iax:iaz,:,:,kel:ke)

! send data to the right node
!
        call send_real_array   (nn, pneighs(3,2), 321, ub(:,:,:,:), iret)

! receive data from the right node
!
        call receive_real_array(nn, pneighs(3,2), 322, ub(:,:,:,:), iret)

! update ghost zones
!
        uu(iax:iaz,:,:,keu:km) = ub(inx:inz,:,:,:)

      end if
#endif /* MPI */

!-   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -
!
    case(bnd_open, bnd_outflow)

! "open" boundary conditions set the derivative normal to boundary to zero
! "outflow" boundary conditions are the same conditions as "open" but the normal
! component of velocity is set to zero in inflow locations
!
      do k = 1, ng

! prepare indices
!
        kt   = ke + k
        ktm1 = kt - 1
        ktm2 = kt - 2

! exptrapolate vector potential components
!
        uu(iax,:,:,kt) = 2.0d0 * uu(iax,:,:,ktm1) - uu(iax,:,:,ktm2)
        uu(iay,:,:,kt) = 2.0d0 * uu(iay,:,:,ktm1) - uu(iay,:,:,ktm2)
        uu(iaz,:,:,kt) =         uu(iaz,:,:,ktm1)

      end do

!-   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -
!
    case(bnd_reflective)

! "reflective" boundary conditions reverse layers inverting the normal component
! of velocity;
!
      do k = 1, ng

! prepare source and target indices
!
        ks   = keu - k
        kt   = ke  + k
#ifdef CT
        ksm1 = ks  - 1
        ktm1 = kt  - 1
#endif /* CT */

! copy source layer to target
!
#ifdef CT
        uu(iax,:,:,kt) =   uu(iax,:,:,ktm1)                                    &
                                         + (uu(iax,:,:,ks) - uu(iax,:,:,ksm1))
        uu(iay,:,:,kt) =   uu(iay,:,:,ktm1)                                    &
                                         + (uu(iay,:,:,ks) - uu(iay,:,:,ksm1))
        uu(iaz,:,:,kt) =   uu(iaz,:,:,ks  )
#endif /* CT */
#ifdef CD
        uu(iax,:,:,kt) =   uu(iax,:,:,ks  )
        uu(iay,:,:,kt) =   uu(iay,:,:,ks  )
        uu(iaz,:,:,kt) =   uu(iaz,:,:,ks  )
#endif /* CD */

      end do

!-   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -
!
    case(bnd_current_free)

! "current-free" boundary conditions are the same as "open", but the magnetic
! field is extrapolated to be free of current density
!
      do k = 1, ng

! prepare indices
!
        kt   = ke + k
        ktm1 = kt - 1

! update magnetic variables
!
        ktm2 = kt - 2

! exptrapolate magnetic potential
!
        uu(iax,:,:,kt) = 2.0d0 * uu(iax,:,:,ktm1) - uu(iax,:,:,ktm2)
        uu(iay,:,:,kt) = 2.0d0 * uu(iay,:,:,ktm1) - uu(iay,:,:,ktm2)
        uu(iaz,:,:,kt) =         uu(iaz,:,:,ktm1)

! update Ax and Ay from nabla(A) = 0
!
        do j = 1, jm
          jm1 = max( 1, j - 1)
          jp1 = min(jm, j + 1)
          do i = 1, im
            im1 = max( 1, i - 1)
            ip1 = min(im, i + 1)

            dax = uu(iax,ip1,j,ktm1) - 2.0d0 * uu(iax,i,j,ktm1)                &
                                                          + uu(iax,im1,j,ktm1)
            day = uu(iax,i,jp1,ktm1) - 2.0d0 * uu(iax,i,j,ktm1)                &
                                                          + uu(iax,i,jm1,ktm1)

            uu(iax,i,j,kt) = 2.0d0 * uu(iax,i,j,ktm1) - uu(iax,i,j,ktm2)       &
                                                   - dax * dzyi2 - day * dzyi2

            dax = uu(iay,ip1,j,ktm1) - 2.0d0 * uu(iay,i,j,ktm1)                &
                                                          + uu(iay,im1,j,ktm1)
            day = uu(iay,i,jp1,ktm1) - 2.0d0 * uu(iay,i,j,ktm1)                &
                                                          + uu(iay,i,jm1,ktm1)

            uu(iay,i,j,kt) = 2.0d0 * uu(iay,i,j,ktm1) - uu(iay,i,j,ktm2)       &
                                                   - dax * dzyi2 - day * dzyi2
          end do
        end do
      end do

!-   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -   - -
!
    case default

! "periodic" boundary conditions simply copy one side to another
!
      do k = 1, ng

! prepare source and target indices
!
        ks = kbl + k
        kt = ke  + k

! copy source layer to target
!
        uu(iax:iaz,:,:,kt) = uu(iax:iaz,:,:,ks)

      end do

    end select

!-------------------------------------------------------------------------------
!
  end subroutine zboundary_magnetic
#endif /* R3D */
#endif /* VPOT */

!===============================================================================
!
end module boundaries
