!!******************************************************************************
!!
!!  This file is part of the GODUNOV source code, a program to perform
!!  Newtonian or relativistic magnetohydrodynamical simulations.
!!
!!  Copyright (C) 2007-2013 Grzegorz Kowal <grzegorz@amuncode.org>
!!
!!  This program is free software: you can redistribute it and/or modify
!!  it under the terms of the GNU General Public License as published by
!!  the Free Software Foundation, either version 3 of the License, or
!!  (at your option) any later version.
!!
!!  This program is distributed in the hope that it will be useful,
!!  but WITHOUT ANY WARRANTY; without even the implied warranty of
!!  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!!  GNU General Public License for more details.
!!
!!  You should have received a copy of the GNU General Public License
!!  along with this program.  If not, see <http://www.gnu.org/licenses/>.
!!
!!******************************************************************************
!!
!! module: EQUATIONS
!!
!!  This module handles supported sets of equations, provides subroutines to
!!  convert between conserved and primitive variables, calculate flux and
!!  characteristic speeds.
!!
!!******************************************************************************
!
module equations

! module variables are not implicit by default
!
  implicit none

#ifdef ADI
! adiabatic heat ratio
!
  real, save :: gamma   = 5.0d+00 / 3.0d+00

! additional adiabatic parameters
!
  real, save :: gammam1 = 2.0d+00 / 3.0d+00, gammam1i = 1.5d+00
#endif /* ADI */

#ifdef ISO
! isothermal speed of sound and its second power
!
  real, save :: csnd    = 1.0d+00
  real, save :: csnd2   = 1.0d+00
#endif /* ISO */

#ifdef SHOCK_DETECTION
! strong shock threshold
!
  real, save :: shock   = 4.0d+00
#endif /* SHOCK_DETECTION */

! by default everything is public
!
  public

!- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
!
  contains
!
!===============================================================================
!
! subroutine INITIALIZE_EQUATIONS:
! -------------------------------
!
!   Subroutine sets the default values of module parameters and obtains their
!   values from the PARAMETERS module.
!
!===============================================================================
!
  subroutine initialize_equations()

! include external procedures and variables
!
    use parameters, only : get_parameter_real

! local variables are not implicit by default
!
    implicit none
!
!-------------------------------------------------------------------------------
!
#ifdef ADI
! obtain the adiabatic specific heat ratio
!
    call get_parameter_real("gamma" , gamma )

! calculate additional parameters
!
    gammam1  = gamma - 1.0d+00
    gammam1i = 1.0d+00 / gammam1
#endif /* ADI */

#ifdef ISO
! obtain the isothermal sound speed
!
    call get_parameter_real("csnd"  , csnd  )

! calculate additional parameters
!
    csnd2 = csnd * csnd
#endif /* ISO */

#ifdef SHOCK_DETECTION
! strong shock threshold
!
    call get_parameter_real("shock" , shock )

#ifdef ISO
! normalize shock parameter to correspond to the isothermal pressure
!
    shock = shock / csnd2
#endif /* ISO */
#endif /* SHOCK_DETECTION */

!-------------------------------------------------------------------------------
!
  end subroutine initialize_equations
#ifdef HYDRO
!
!===============================================================================
!
! subroutine CONS2PRIM:
! --------------------
!
!   Subroutine converts the conservative representation of variables to
!   its corresponding primitive representation.
!
!   Arguments:
!
!     n - the length of input and output vectors;
!     u - the input array of conservative variables;
!     q - the output array of primitive variables;
!
!===============================================================================
!
  subroutine cons2prim(n, u, q)

! include external procedures and variables
!
    use variables , only : nt
    use variables , only : idn, imx, imy, imz, ivx, ivy, ivz
#ifdef ADI
    use variables , only : ipr, ien
#endif /* ADI */

! local variables are not implicit by default
!
    implicit none

! input/output arguments
!
    integer              , intent(in)  :: n
    real, dimension(nt,n), intent(in)  :: u
    real, dimension(nt,n), intent(out) :: q

! local variables
!
    integer :: i
#ifdef ADI
    real    :: ek, ei
#endif /* ADI */
!
!-------------------------------------------------------------------------------
!
    do i = 1, n

      q(idn,i) = u(idn,i)
      q(ivx,i) = u(imx,i) / u(idn,i)
      q(ivy,i) = u(imy,i) / u(idn,i)
      q(ivz,i) = u(imz,i) / u(idn,i)
#ifdef ADI
      ek       = 0.5d+00 * sum(u(imx:imz,i) * q(ivx:ivz,i))
      ei       = u(ien,i) - ek
      q(ipr,i) = gammam1 * ei
#endif /* ADI */

    end do

!-------------------------------------------------------------------------------
!
  end subroutine cons2prim
!
!===============================================================================
!
! subroutine PRIM2CONS:
! --------------------
!
!   Subroutine converts the primitive variable representation to its
!   corresponding conservative representation.
!
!   Arguments:
!
!     n - the length of input and output vectors;
!     q - the input array of primitive variables;
!     u - the output array of conservative variables;
!
!===============================================================================
!
  subroutine prim2cons(n, q, u)

! include external procedures and variables
!
    use variables , only : nt
    use variables , only : idn, imx, imy, imz, ivx, ivy, ivz
#ifdef ADI
    use variables , only : ipr, ien
#endif /* ADI */

! local variables are not implicit by default
!
    implicit none

! input/output arguments
!
    integer              , intent(in)  :: n
    real, dimension(nt,n), intent(in)  :: q
    real, dimension(nt,n), intent(out) :: u

! local variables
!
    integer :: i
#ifdef ADI
    real    :: ek, ei
#endif /* ADI */
!
!-------------------------------------------------------------------------------
!
    do i = 1, n

      u(idn,i) = q(idn,i)
      u(imx,i) = q(idn,i) * q(ivx,i)
      u(imy,i) = q(idn,i) * q(ivy,i)
      u(imz,i) = q(idn,i) * q(ivz,i)
#ifdef ADI
      ek       = 0.5d+00 * sum(u(imx:imz,i) * q(ivx:ivz,i))
      ei       = gammam1i * q(ipr,i)
      u(ien,i) = ei + ek
#endif /* ADI */

    end do

!-------------------------------------------------------------------------------
!
  end subroutine prim2cons
!
!===============================================================================
!
! subroutine FLUXSPEED:
! --------------------
!
!   Subroutine calculates fluxes and characteristic speeds from primitive and
!   conservative variables.
!
!   Arguments:
!
!     n - the length of input and output vectors;
!     q - the input array of primitive variables;
!     u - the input array of conservative variables;
!     f - the output vector of fluxes;
!     c - the output vector of characteristic speeds;
!
!===============================================================================
!
  subroutine fluxspeed(n, q, u, f, c)

! include external procedures and variables
!
    use variables , only : nt
    use variables , only : idn, imx, imy, imz, ivx, ivy, ivz
#ifdef ADI
    use variables , only : ipr, ien
#endif /* ADI */

! local variables are not implicit by default
!
    implicit none

! input/output arguments
!
    integer              , intent(in)  :: n
    real, dimension(nt,n), intent(in)  :: q, u
    real, dimension(nt,n), intent(out) :: f
    real, dimension(n)   , intent(out) :: c

! local variables
!
    integer :: i
!
!-------------------------------------------------------------------------------
!
    do i = 1, n

! calculate the hydrodynamic fluxes
!
      f(idn,i) = u(imx,i)
      f(imx,i) = q(ivx,i) * u(imx,i)
      f(imy,i) = q(ivx,i) * u(imy,i)
      f(imz,i) = q(ivx,i) * u(imz,i)
#ifdef ISO
      f(imx,i) = f(imx,i) + csnd2 * q(idn,i)
#endif /* ISO */
#ifdef ADI
      f(imx,i) = f(imx,i) + q(ipr,i)
      f(ien,i) = q(ivx,i) * (u(ien,i) + q(ipr,i))
#endif /* ADI */

! calculate the speed of sound
!
#ifdef ADI
      c(i) = sqrt(gamma * q(ipr,i) / q(idn,i))
#endif /* ADI */
#ifdef ISO
      c(i) = csnd
#endif /* ISO */

    end do

!-------------------------------------------------------------------------------
!
  end subroutine fluxspeed
!
!===============================================================================
!
! function MAXSPEED:
! -----------------
!
!   Function scans the variable array and returns the maximum speed in the
!   system.
!
!===============================================================================
!
  function maxspeed()

! include external procedures and variables
!
    use mesh      , only : ib, ie, jb, je, kb, ke
    use variables , only : qp
    use variables , only : ivx, ivz
#ifdef ADI
    use variables , only : idn, ipr
#endif /* ADI */

! local variables are not implicit by default
!
    implicit none

! return variable
!
    real     :: maxspeed

! local variables
!
    integer  :: i, j, k
    real     :: vv, v, c
!
!-------------------------------------------------------------------------------
!
! reset the maximum speed
!
    maxspeed = 0.0d+00

! iterate over all positions
!
    do k = kb, ke
      do j = jb, je
        do i = ib, ie

! calculate the velocity amplitude
!
          vv = sum(qp(ivx:ivz,i,j,k) * qp(ivx:ivz,i,j,k))
          v  = sqrt(vv)

#ifdef ADI
! calculate the adiabatic speed of sound
!
          c = sqrt(gamma * qp(ipr,i,j,k) / qp(idn,i,j,k))
#endif /* ADI */

! calculate the maximum speed
!
#ifdef ISO
          maxspeed = max(maxspeed, v + csnd)
#endif /* ISO */
#ifdef ADI
          maxspeed = max(maxspeed, v + c)
#endif /* ADI */

        end do
      end do
    end do

! return the value
!
    return

!-------------------------------------------------------------------------------
!
  end function maxspeed
!
!===============================================================================
!
! subroutine UPDATE_PRIMITIVE_VARIABLES:
! -------------------------------------
!
!   Subroutine updates primitive variables from their conservative
!   representation.  This process is done ones after advance of the conserved
!   variables due to their evolution in time.
!
!===============================================================================
!
  subroutine update_primitive_variables()

! include external procedures and variables
!
    use mesh     , only : in, ib, ie, jb, je, kb, ke
    use mesh     , only : xc, yc, zc
#ifdef DEBUG
    use variables, only : idn
#ifdef ADI
    use variables, only : ipr, ien
#endif /* ADI */
#endif /* DEBUG */
    use variables, only : nu, nt
    use variables, only : uu, qp

! local variables are not implicit by default
!
    implicit none

! temporary variables
!
#ifdef DEBUG
    integer                :: i
#endif /* DEBUG */
    integer                :: j, k

! temporary array to store conserved variable vector
!
    real, dimension(nt,in) :: u
!
!-------------------------------------------------------------------------------
!
! update primitive variables
!
    do k = kb, ke
      do j = jb, je

! copy variables to temporary array of conserved variables
!
        u(  1: nu,1:in) = uu(  1: nu,ib:ie,j,k)

! convert conserved variables to primitive ones
!
        call cons2prim(in, u(1:nt,1:in), qp(1:nt,ib:ie,j,k))

#ifdef DEBUG
! check the positivity of density and pressure
!
        do i = ib, ie
          if (qp(idn,i,j,k) <= 0.0d+00) then
            write(*,*)
            write(*,*) 'Unphysical state: ρ ≤ 0', i, j, k
            write(*,"('X = ',3(1pe24.16))") xc(i), yc(j), zc(k)
            write(*,"('Q = ',5(1pe24.16))") qp(1:nt,i,j,k)
            write(*,"('U = ',5(1pe24.16))") uu(1:nt,i,j,k)
            stop
          end if
#ifdef ADI
          if (uu(ien,i,j,k) <= 0.0d+00) then
            write(*,*)
            write(*,*) 'Unphysical state: E ≤ 0', i, j, k
            write(*,"('X = ',3(1pe24.16))") xc(i), yc(j), zc(k)
            write(*,"('Q = ',5(1pe24.16))") qp(1:nt,i,j,k)
            write(*,"('U = ',5(1pe24.16))") uu(1:nt,i,j,k)
            stop
          end if
          if (qp(ipr,i,j,k) <= 0.0d+00) then
            write(*,*)
            write(*,*) 'Unphysical state: p ≤ 0', i, j, k
            write(*,"('X = ',3(1pe24.16))") xc(i), yc(j), zc(k)
            write(*,"('Q = ',5(1pe24.16))") qp(1:nt,i,j,k)
            write(*,"('U = ',5(1pe24.16))") uu(1:nt,i,j,k)
            stop
          end if
#endif /* ADI */
        end do
#endif /* DEBUG */

      end do
    end do

!-------------------------------------------------------------------------------
!
  end subroutine update_primitive_variables
#endif /* HYDRO */
#ifdef MHD
!
!===============================================================================
!
! subroutine CONS2PRIM:
! --------------------
!
!   Subroutine converts the conservative representation of variables to
!   its corresponding primitive representation.
!
!   Arguments:
!
!     n - the length of input and output vectors;
!     u - the input array of conservative variables;
!     q - the output array of primitive variables;
!
!===============================================================================
!
  subroutine cons2prim(n, u, q)

! include external procedures and variables
!
    use variables , only : nt
    use variables , only : idn, imx, imy, imz, ivx, ivy, ivz, ibx, iby, ibz
#ifdef ADI
    use variables , only : ipr, ien
#endif /* ADI */
#ifdef GLM
    use variables , only : iph
#endif /* GLM */

! local variables are not implicit by default
!
    implicit none

! input/output arguments
!
    integer              , intent(in)  :: n
    real, dimension(nt,n), intent(in)  :: u
    real, dimension(nt,n), intent(out) :: q

! local variables
!
    integer :: i
#ifdef ADI
    real    :: ei, ek, em
#endif /* ADI */
!
!-------------------------------------------------------------------------------
!
    do i = 1, n

      q(idn,i) = u(idn,i)
      q(ivx,i) = u(imx,i) / u(idn,i)
      q(ivy,i) = u(imy,i) / u(idn,i)
      q(ivz,i) = u(imz,i) / u(idn,i)
      q(ibx,i) = u(ibx,i)
      q(iby,i) = u(iby,i)
      q(ibz,i) = u(ibz,i)
#ifdef ADI
      ek       = 0.5d+00 * sum(u(imx:imz,i) * q(ivx:ivz,i))
      em       = 0.5d+00 * sum(q(ibx:ibz,i) * q(ibx:ibz,i))
      ei       = u(ien,i) - (ek + em)
      q(ipr,i) = gammam1 * ei
#endif /* ADI */
#ifdef GLM
      q(iph,i) = u(iph,i)
#endif /* GLM */

    end do

!-------------------------------------------------------------------------------
!
  end subroutine cons2prim
!
!===============================================================================
!
! subroutine PRIM2CONS:
! --------------------
!
!   Subroutine converts the primitive variable representation to its
!   corresponding conservative representation.
!
!   Arguments:
!
!     n - the length of input and output vectors;
!     q - the input array of primitive variables;
!     u - the output array of conservative variables;
!
!===============================================================================
!
  subroutine prim2cons(n, q, u)

! include external procedures and variables
!
    use variables , only : nt
    use variables , only : idn, imx, imy, imz, ivx, ivy, ivz, ibx, iby, ibz
#ifdef ADI
    use variables , only : ipr, ien
#endif /* ADI */
#ifdef GLM
    use variables , only : iph
#endif /* GLM */

! local variables are not implicit by default
!
    implicit none

! input/output arguments
!
    integer              , intent(in)  :: n
    real, dimension(nt,n), intent(in)  :: q
    real, dimension(nt,n), intent(out) :: u

! local variables
!
    integer :: i
#ifdef ADI
    real    :: ei, ek, em
#endif /* ADI */
!
!-------------------------------------------------------------------------------
!
    do i = 1, n

      u(idn,i) = q(idn,i)
      u(imx,i) = q(idn,i) * q(ivx,i)
      u(imy,i) = q(idn,i) * q(ivy,i)
      u(imz,i) = q(idn,i) * q(ivz,i)
      u(ibx,i) = q(ibx,i)
      u(iby,i) = q(iby,i)
      u(ibz,i) = q(ibz,i)
#ifdef ADI
      ei       = gammam1i * q(ipr,i)
      ek       = 0.5d+00 * sum(u(imx:imz,i) * q(ivx:ivz,i))
      em       = 0.5d+00 * sum(q(ibx:ibz,i) * q(ibx:ibz,i))
      u(ien,i) = ei + ek + em
#endif /* ADI */
#ifdef GLM
      u(iph,i) = q(iph,i)
#endif /* GLM */

    end do

!-------------------------------------------------------------------------------
!
  end subroutine prim2cons
!
!===============================================================================
!
! subroutine FLUXSPEED:
! --------------------
!
!   Subroutine calculates fluxes and characteristic speeds from primitive and
!   conservative variables.
!
!   Arguments:
!
!     n - the length of input and output vectors;
!     q - the input array of primitive variables;
!     u - the input array of conservative variables;
!     f - the output vector of fluxes;
!     c - the output vector of characteristic speeds;
!
!===============================================================================
!
  subroutine fluxspeed(n, q, u, f, c)

! include external procedures and variables
!
    use variables , only : nt
    use variables , only : idn, imx, imy, imz, ivx, ivy, ivz
    use variables , only : ibx, iby, ibz
#ifdef ADI
    use variables , only : ipr, ien
#endif /* ADI */
#ifdef GLM
    use variables , only : iph
    use variables , only : cmax2
#endif /* GLM */

! local variables are not implicit by default
!
    implicit none

! input/output arguments
!
    integer              , intent(in)  :: n
    real, dimension(nt,n), intent(in)  :: q, u
    real, dimension(nt,n), intent(out) :: f
    real, dimension(n)   , intent(out) :: c

! local variables
!
    integer :: i
    real    :: bb, vb, pm, pt
    real    :: cs2, ca2, cx2, cf2, cl2
!
!-------------------------------------------------------------------------------
!
    do i = 1, n

! prepare pressures and scalar product
!
      bb = sum(q(ibx:ibz,i) * q(ibx:ibz,i))
      pm = 0.5d+00 * bb
#ifdef ADI
      vb = sum(q(ivx:ivz,i) * q(ibx:ibz,i))
      pt = q(ipr,i)
#endif /* ADI */
#ifdef ISO
      pt = csnd2 * q(idn,i)
#endif /* ISO */
      pt = pt + pm

! calculate the magnetohydrodynamic fluxes
!
      f(idn,i) = u(imx,i)
      f(imx,i) = q(ivx,i) * u(imx,i) - q(ibx,i) * q(ibx,i)
      f(imy,i) = q(ivx,i) * u(imy,i) - q(ibx,i) * q(iby,i)
      f(imz,i) = q(ivx,i) * u(imz,i) - q(ibx,i) * q(ibz,i)
      f(imx,i) = f(imx,i) + pt
#ifdef ADI
      f(ien,i) = q(ivx,i) * (u(ien,i) + pt) - q(ibx,i) * vb
#endif /* ADI */
#ifdef GLM
      f(ibx,i) = q(iph,i)
      f(iph,i) = cmax2 * q(ibx,i)
#else /* GLM */
      f(ibx,i) = 0.0d+00
#endif /* GLM */
      f(iby,i) = q(ivx,i) * q(iby,i) - q(ibx,i) * q(ivy,i)
      f(ibz,i) = q(ivx,i) * q(ibz,i) - q(ibx,i) * q(ivz,i)

! calculate the fast magnetosonic speed
!
#ifdef ADI
      cs2  = gamma * q(ipr,i) / q(idn,i)
#endif /* ADI */
#ifdef ISO
      cs2  = csnd2
#endif /* ISO */
      ca2  = sum(q(ibx:ibz,i) * q(ibx:ibz,i)) / q(idn,i)
      cx2  = q(ibx,i) * q(ibx,i) / q(idn,i)
      cf2  = cs2 + ca2
      cl2  = max(0.0d+00, cf2 * cf2 - 4.0d+00 * cs2 * cx2)

      c(i) = sqrt(0.5d+00 * (cf2 + sqrt(cl2)))

    end do

!-------------------------------------------------------------------------------
!
  end subroutine fluxspeed
!
!===============================================================================
!
! function MAXSPEED:
! -----------------
!
!   Function scans the variable array and returns the maximum speed in the
!   system.
!
!===============================================================================
!
  function maxspeed()

! include external procedures and variables
!
    use mesh      , only : ib, ie, jb, je, kb, ke
    use variables , only : qp
    use variables , only : idn, ivx, ivz, ibx, ibz
#ifdef ADI
    use variables , only : ipr
#endif /* ADI */

! local variables are not implicit by default
!
    implicit none

! return variable
!
    real     :: maxspeed

! local variables
!
    integer  :: i, j, k
    real     :: vv, bb, v, c
!
!-------------------------------------------------------------------------------
!
! reset the maximum speed
!
    maxspeed = 0.0d+00

! iterate over all positions
!
    do k = kb, ke
      do j = jb, je
        do i = ib, ie

! calculate the velocity amplitude
!
          vv = sum(qp(ivx:ivz,i,j,k) * qp(ivx:ivz,i,j,k))
          v  = sqrt(vv)
          bb = sum(qp(ibx:ibz,i,j,k) * qp(ibx:ibz,i,j,k))

! calculate the fast magnetosonic speed
!
#ifdef ISO
          c = sqrt(csnd2 + bb / qp(idn,i,j,k))
#endif /* ISO */
#ifdef ADI
          c = sqrt((gamma * qp(ipr,i,j,k) + bb) / qp(idn,i,j,k))
#endif /* ADI */

! calculate the maximum of speed
!
          maxspeed = max(maxspeed, v + c)

        end do
      end do
    end do

! return the value
!
    return

!-------------------------------------------------------------------------------
!
  end function maxspeed
!
!===============================================================================
!
! subroutine UPDATE_PRIMITIVE_VARIABLES:
! -------------------------------------
!
!   Subroutine updates primitive variables from their conservative
!   representation.  This process is done ones after advance of the conserved
!   variables due to their evolution in time.
!
!===============================================================================
!
  subroutine update_primitive_variables()

! include external procedures and variables
!
    use mesh     , only : in, ib, ie, jb, je, kb, ke
    use mesh     , only : xc, yc, zc
#ifdef DEBUG
    use variables, only : idn
#ifdef ISO
    use variables, only : ivz, imz
#endif /* ISO */
#ifdef ADI
    use variables, only : ipr, ien
#endif /* ADI */
#endif /* DEBUG */
    use variables, only : ibx, ibz
#ifdef GLM
    use variables, only : iph
#endif /* GLM */
    use variables, only : nu, nt
    use variables, only : uu, qp

! local variables are not implicit by default
!
    implicit none

! temporary variables
!
#ifdef DEBUG
    integer                :: i
#endif /* DEBUG */
    integer                :: j, k

! temporary array to store conserved variable vector
!
    real, dimension(nt,in) :: u
!
!-------------------------------------------------------------------------------
!
! update primitive variables
!
    do k = kb, ke
      do j = jb, je

#ifdef GLM
! copy variables to temporary array of conserved variables
!
        u(  1: nt,1:in) = uu(  1: nt,ib:ie,j,k)
#else /* GLM */
! copy variables to temporary array of conserved variables
!
        u(  1: nu,1:in) = uu(  1: nu,ib:ie,j,k)

! copy magnetic field
!
        u(ibx:ibz,1:in) = qp(ibx:ibz,ib:ie,j,k)
#endif /* GLM */

! convert conserved variables to primitive ones
!
        call cons2prim(in, u(1:nt,1:in), qp(1:nt,ib:ie,j,k))

#ifdef DEBUG
! check the positivity of density and pressure
!
        do i = ib, ie
          if (qp(idn,i,j,k) <= 0.0d+00) then
            write(*,*)
            write(*,*) 'Unphysical state: ρ ≤ 0', i, j, k
            write(*,"('X = ',3(1pe24.16))") xc(i), yc(j), zc(k)
#ifdef ADI
            write(*,"('Q = ',5(1pe24.16))") qp(  1:ipr,i,j,k)
            write(*,"('    ',4(1pe24.16))") qp(ibx:nt ,i,j,k)
            write(*,"('U = ',5(1pe24.16))") uu(  1:ien,i,j,k)
            write(*,"('    ',4(1pe24.16))") uu(ibx:nt ,i,j,k)
#endif /* ADI */
#ifdef ISO
            write(*,"('Q = ',4(1pe24.16))") qp(  1:ivz,i,j,k)
            write(*,"('    ',4(1pe24.16))") qp(ibx:nt ,i,j,k)
            write(*,"('U = ',4(1pe24.16))") uu(  1:imz,i,j,k)
            write(*,"('    ',4(1pe24.16))") uu(ibx:nt ,i,j,k)
#endif /* ISO */
            stop
          end if
#ifdef ADI
          if (uu(ien,i,j,k) <= 0.0d+00) then
            write(*,*)
            write(*,*) 'Unphysical state: E ≤ 0', i, j, k
            write(*,"('X = ',3(1pe24.16))") xc(i), yc(j), zc(k)
            write(*,"('Q = ',5(1pe24.16))") qp(  1:ipr,i,j,k)
            write(*,"('    ',4(1pe24.16))") qp(ibx:nt ,i,j,k)
            write(*,"('U = ',5(1pe24.16))") uu(  1:ien,i,j,k)
            write(*,"('    ',4(1pe24.16))") uu(ibx:nt ,i,j,k)
            stop
          end if
          if (qp(ipr,i,j,k) <= 0.0d+00) then
            write(*,*)
            write(*,*) 'Unphysical state: p ≤ 0', i, j, k
            write(*,"('X = ',3(1pe24.16))") xc(i), yc(j), zc(k)
            write(*,"('Q = ',5(1pe24.16))") qp(  1:ipr,i,j,k)
            write(*,"('    ',4(1pe24.16))") qp(ibx:nt ,i,j,k)
            write(*,"('U = ',5(1pe24.16))") uu(  1:ien,i,j,k)
            write(*,"('    ',4(1pe24.16))") uu(ibx:nt ,i,j,k)
            stop
          end if
#endif /* ADI */
        end do
#endif /* DEBUG */

      end do
    end do

!-------------------------------------------------------------------------------
!
  end subroutine update_primitive_variables
#endif /* MHD */
#ifdef BFIELD
!
!===============================================================================
!
! subroutine UPDATE_MAGNETIC_VARIABLES:
! ------------------------------------
!
!   Subroutine calculates the face and cell centered magnetic field components
!   from vector potential.
!
!===============================================================================
!
  subroutine update_magnetic_variables()

! include external procedures and variables
!
#ifdef VPOT
    use mesh     , only : im, jm, km
    use operators, only : curl, atob
    use variables, only : iax, iaz
    use variables, only : ibx, iby, ibz
    use variables, only : inx, iny, inz
    use variables, only : buni
    use variables, only : bi
    use variables, only : uu, qp
#endif /* VPOT */

! local variables are not implicit by default
!
    implicit none

!-------------------------------------------------------------------------------
!
#ifdef VPOT
! calculate magnetic field at cell interfaces and centers
!
#ifdef CD
    call curl(im, jm, km, uu(iax:iaz,:,:,:), qp(ibx:ibz,:,:,:), 'c')
    call atob(im, jm, km, uu(iax:iaz,:,:,:), bi(inx:inz,:,:,:), 'c')
#endif /* CD */
#ifdef CT
    call curl(im, jm, km, uu(iax:iaz,:,:,:), bi(inx:inz,:,:,:), 'f')
    call atob(im, jm, km, uu(iax:iaz,:,:,:), qp(ibx:ibz,:,:,:), 'f')
#endif /* CT */

! add uniform component to magnetic field components
!
    if (buni(inx) /= 0.0d+00) then
      qp(ibx,:,:,:) = qp(ibx,:,:,:) + buni(inx)
      bi(inx,:,:,:) = bi(inx,:,:,:) + buni(inx)
    end if
    if (buni(iny) /= 0.0d+00) then
      qp(iby,:,:,:) = qp(iby,:,:,:) + buni(iny)
      bi(iny,:,:,:) = bi(iny,:,:,:) + buni(iny)
    end if
#ifndef R2D
    if (buni(inz) /= 0.0d+00) then
      qp(ibz,:,:,:) = qp(ibz,:,:,:) + buni(inz)
      bi(inz,:,:,:) = bi(inz,:,:,:) + buni(inz)
    end if
#endif /* !R2D */
#endif /* VPOT */

!-------------------------------------------------------------------------------
!
  end subroutine update_magnetic_variables
#endif /* BFIELD */
#ifdef SHOCK_DETECTION
!
!===============================================================================
!
! subroutine DETECT_SHOCK:
! -----------------------
!
!   Subroutine detects strong shocks by checking if the local divergence of
!   velocity is negative and the normalized gradient of pressure or density
!   (in the adiabatic or isothermal case, respectively) is larger then a given
!   threshold value.
!
!===============================================================================
!
  subroutine detect_shock()

! include external procedures and variables
!
    use mesh     , only : im, jm, km
    use mesh     , only : dxih, dyih, dzih
    use variables, only : qp, sh
    use variables, only : ivx, ivy, ivz
#ifdef ADI
    use variables, only : ipr
#endif /* ADI */
#ifdef ISO
    use variables, only : idn
#endif /* ISO */

! local variables are not implicit by default
!
    implicit none

! temporary variables
!
    integer :: i, j, k
    integer :: im1, jm1, km1
    integer :: ip1, jp1, kp1
    real    :: dv, dpx, dpy, dpz
#ifdef ADI
    real    :: dpr
#endif /* ADI */
#ifdef ISO
    real    :: ddn
#endif /* ISO */
!
!-------------------------------------------------------------------------------
!
! find strong shocks from the velocity divergence and pressure (or density)
! gradient
!
#ifdef R3D
    do k = 2, km - 1
      km1 = k - 1
      kp1 = k + 1
      do j = 2, jm - 1
        jm1 = j - 1
        jp1 = j + 1
        do i = 2, im - 1
          im1 = i - 1
          ip1 = i + 1

! reset the shock indicator at the current location
!
          sh(i,j,k) = .false.

! calculate the divergence of velocity
!
          dv = (qp(ivx,ip1,j,k) - qp(ivx,im1,j,k)) * dxih                      &
             + (qp(ivy,i,jp1,k) - qp(ivy,i,jm1,k)) * dyih                      &
             + (qp(ivz,i,j,kp1) - qp(ivz,i,j,km1)) * dzih

! if the divergence of velocity is negative, check the pressure gradient
!
          if (dv < 0.0d+00) then

#ifdef ADI
! calculate the normalized pressure gradient
!
            dpx = abs(qp(ipr,ip1,j,k) - qp(ipr,im1,j,k))                       &
                / min(qp(ipr,ip1,j,k),  qp(ipr,im1,j,k), qp(ipr,i,j,k))
            dpy = abs(qp(ipr,i,jp1,k) - qp(ipr,i,jm1,k))                       &
                / min(qp(ipr,i,jp1,k),  qp(ipr,i,jm1,k), qp(ipr,i,j,k))
            dpz = abs(qp(ipr,i,j,kp1) - qp(ipr,i,j,km1))                       &
                / min(qp(ipr,i,j,kp1),  qp(ipr,i,j,km1), qp(ipr,i,j,k))
            dpr = dpx + dpy + dpz

! check if the pressure gradient is larger then the specified threshold
!
            if (dpr >= shock) sh(i,j,k) = .true.
#endif /* ADI */
#ifdef ISO
! calculate the normalized density gradient
!
            dpx = abs(qp(idn,ip1,j,k) - qp(idn,im1,j,k))                       &
                / min(qp(idn,ip1,j,k),  qp(idn,im1,j,k), qp(idn,i,j,k))
            dpy = abs(qp(idn,i,jp1,k) - qp(idn,i,jm1,k))                       &
                / min(qp(idn,i,jp1,k),  qp(idn,i,jm1,k), qp(idn,i,j,k))
            dpz = abs(qp(idn,i,j,kp1) - qp(idn,i,j,km1))                       &
                / min(qp(idn,i,j,kp1),  qp(idn,i,j,km1), qp(idn,i,j,k))
            ddn = dpx + dpy + dpz

! check if the density gradient is larger then the specified threshold
!
            if (ddn >= shock) sh(i,j,k) = .true.
#endif /* ISO */

          end if

        end do
      end do
    end do
#else /* R3D */
    do k = 1, km
      do j = 2, jm - 1
        jm1 = j - 1
        jp1 = j + 1
        do i = 2, im - 1
          im1 = i - 1
          ip1 = i + 1

! reset the shock indicator at the current location
!
          sh(i,j,k) = .false.

! calculate the divergence of velocity
!
          dv = (qp(ivx,ip1,j,k) - qp(ivx,im1,j,k)) * dxih                      &
             + (qp(ivy,i,jp1,k) - qp(ivy,i,jm1,k)) * dyih

! if the divergence of velocity is negative, check the pressure or density
! gradient
!
          if (dv < 0.0d+00) then

#ifdef ADI
! calculate the normalized pressure gradient
!
            dpx = abs(qp(ipr,ip1,j,k) - qp(ipr,im1,j,k))                       &
                / min(qp(ipr,ip1,j,k),  qp(ipr,im1,j,k), qp(ipr,i,j,k))
            dpy = abs(qp(ipr,i,jp1,k) - qp(ipr,i,jm1,k))                       &
                / min(qp(ipr,i,jp1,k),  qp(ipr,i,jm1,k), qp(ipr,i,j,k))
            dpr = dpx + dpy

! check if the pressure gradient is larger then the specified threshold
!
            if (dpr >= shock) sh(i,j,k) = .true.
#endif /* ADI */
#ifdef ISO
! calculate the normalized density gradient
!
            dpx = abs(qp(idn,ip1,j,k) - qp(idn,im1,j,k))                       &
                / min(qp(idn,ip1,j,k),  qp(idn,im1,j,k), qp(idn,i,j,k))
            dpy = abs(qp(idn,i,jp1,k) - qp(idn,i,jm1,k))                       &
                / min(qp(idn,i,jp1,k),  qp(idn,i,jm1,k), qp(idn,i,j,k))
            ddn = dpx + dpy

! check if the density gradient is larger then the specified threshold
!
            if (ddn >= shock) sh(i,j,k) = .true.
#endif /* ISO */

          end if

        end do
      end do
    end do
#endif /* R3D */

!-------------------------------------------------------------------------------
!
  end subroutine detect_shock
#endif /* SHOCK_DETECTION */

!===============================================================================
!
end module equations
