!!******************************************************************************
!!
!!  This file is part of the GODUNOV source code, a program to perform
!!  Newtonian or relativistic magnetohydrodynamical simulations.
!!
!!  Copyright (C) 2007-2013 Grzegorz Kowal <grzegorz@amuncode.org>
!!
!!  This program is free software: you can redistribute it and/or modify
!!  it under the terms of the GNU General Public License as published by
!!  the Free Software Foundation, either version 3 of the License, or
!!  (at your option) any later version.
!!
!!  This program is distributed in the hope that it will be useful,
!!  but WITHOUT ANY WARRANTY; without even the implied warranty of
!!  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!!  GNU General Public License for more details.
!!
!!  You should have received a copy of the GNU General Public License
!!  along with this program.  If not, see <http://www.gnu.org/licenses/>.
!!
!!******************************************************************************
!!
!! module: FORCING
!!
!!  This module handles turbulence driving and energy injection.
!!
!!******************************************************************************
!
module forcing

#ifdef FORCING
#ifdef PROFILE
! include external subroutines
!
  use timers, only : set_timer, start_timer, stop_timer
#endif /* PROFILE */

! module variables are not implicit by default
!
  implicit none

#ifdef PROFILE
! the forcing module timer indices
!
  integer, save :: ifi, ifc, ifr, iff
#endif /* PROFILE */

! === module runtime parameters ===
!
  real   , save :: pwfor = 1.0d+00 ! the power of forcing
  real   , save :: kl    = 2.0d+00 ! the lower bound for the injection scale
  real   , save :: kf    = 2.5d+00 ! the injection scale of the amplitude peak
  real   , save :: ku    = 3.0d+00 ! the upper bound for the injection scale
  real   , save :: kc    = 0.2d+00 ! the half width of the injection spectrum
  real   , save :: dtfor = 1.0d-06 ! the forcing time step
  real   , save :: tbfor = 0.0d+00 ! these to parameters describe the time for
  real   , save :: tefor = 0.0d+00 ! gradual increase of turbulence
  real   , save :: tsfor = 9.9d+09 ! the time when to turn off turbulence
  real   , save :: drate = 1.0d+03 ! the injection rate for random forcing
  real   , save :: famp  = 1.0d+00 ! the eddy amplitude
  integer, save :: kd    = 1
  integer, save :: ndiv  = 2       ! the index increment for the transforms
#ifdef SUBVOL
  real   , save :: rcut  = 9.9d+09 ! parameters describing the injection region
  real   , save :: ycut  = 9.9d+09
  real   , save :: rdel  = 1.0d+00
  real   , save :: ydel  = 1.0d+00
  real   , save :: ysep  = 0.0d+00
#endif /* SUBVOL */

! module variables
!
  integer     , save                              :: nf
  integer     , save                              :: nlf, nlf3
  real        , save                              :: dtf, sqdtf, pwf
#ifdef RANDOM
  real        , save                              :: eamp, ewth
#else /* RANDOM */
  integer     , dimension(:,:), allocatable, save :: k_table
  real        , dimension(:)  , allocatable, save :: f_table
  real        , dimension(:,:), allocatable, save :: e1, e2
  real        , dimension(:)  , allocatable, save :: sini, cosi, ga, gb
  complex     , dimension(:,:), allocatable, save :: ffor, fvel
#ifdef FAST
  integer     , dimension(:,:), allocatable, save :: kix, kiy, kiz
  real        , dimension(:)  , allocatable, save :: sinq, cosq
#else /* FAST */
  real        , dimension(:,:), allocatable, save :: sinx, siny, sinz          &
                                                   , cosx, cosy, cosz
#endif /* FAST */
#ifdef SUBVOL
  real                                     , save :: fp
  real        , dimension(:,:), allocatable, save :: fr
  real        , dimension(:  ), allocatable, save :: fv
#endif /* SUBVOL */
#endif /* RANDOM */
#ifdef INT_FOR
  real(kind=8)                             , save :: efor, cfor, vfor, sfor    &
                                                   , afrx, afry, afrz
#endif /* INT_FOR */

! module parameters
!
  integer, parameter :: prec = 1000000

! by default everything is private
!
  private

! declare public variables
!
  public :: nf, pwfor, dtfor, kl, kf, ku, kc
#ifdef INT_FOR
  public :: efor, cfor, vfor, sfor, afrx, afry, afrz
#endif /* INT_FOR */

! declare public subroutines
!
  public :: initialize_forcing, finalize_forcing, drive
!
!-------------------------------------------------------------------------------

  contains
!
!===============================================================================
!
! subroutine INITIALIZE_FORCING:
! -----------------------------
!
!   Subroutine initializes forcing by calculating forcing parameters and
!   preparing transformation arrays.
!
!===============================================================================
!
  subroutine initialize_forcing()

! include external procedures and variables
!
    use constants , only : pi, pi2
#ifdef RANDOM
    use constants , only : ehi
#endif /* RANDOM */
    use mesh      , only : in, jn, kn
#ifdef RANDOM
    use mesh      , only : xmin, ymin, zmin, xmax, ymax, zmax
    use mesh      , only : dvolh, dvol
#else /* RANDOM */
    use mesh      , only : dx, dy, dz, dxh, dyh, dzh, xlen, ylen, zlen
#endif /* RANDOM */
    use mesh      , only : im, jm, km, ib, jb, kb, ie, je, ke
    use mesh      , only : xc, yc, zc
    use mpitools  , only : master
#ifdef MPI
    use mpitools  , only : reduce_sum_real
#endif /* MPI */
#ifdef SUBVOL
    use mpitools  , only : pdims
#endif /* SUBVOL */
    use parameters, only : get_parameter_real, get_parameter_integer

! local variables are not implicit by default
!
    implicit none

! local variables
!
    character(len=32)   :: fmt
    integer             :: i, j, k
#ifdef RANDOM
    real                :: fx, fy, fz, fc
    real                :: einj
#else /* RANDOM */
    integer             :: kmx, l, q, jbg, kbg
    real                :: amp, r, fi
    real                :: kk, kr, kx, ky, kz, kxy, kyz, kxz
    real                :: phi
#ifdef FAST
    integer             :: lmx
    integer             :: ix, iy, iz
    real                :: pidx
#endif /* FAST */
#ifdef SUBVOL
    real                :: rr, yy
#endif /* SUBVOL */
#endif /* RANDOM */
#ifdef MPI
    integer             :: iret
#endif /* MPI */

#ifdef RANDOM
! local arrays
!
    real, dimension(im) :: xt, xf
    real, dimension(jm) :: yt, yf
#ifdef R3D
    real, dimension(jm) :: zt, zf
#endif /* R3D */
#endif /* RANDOM */
!
!-------------------------------------------------------------------------------
!
#ifdef PROFILE
! set timer descriptions
!
    call set_timer('forcing initialization' , ifi)
    call set_timer('forcing calculation'    , ifc)
    call set_timer('forcing real to fourier', ifr)
    call set_timer('forcing fourier to real', iff)

! start time accounting for the MPI initialization
!
    call start_timer(ifi)
#endif /* PROFILE */

! read forcing runtime parameters from the parameter file
!
    call get_parameter_real   ("pwfor", pwfor)
    call get_parameter_real   ("kf"   , kf   )
    call get_parameter_real   ("kl"   , kl   )
    call get_parameter_real   ("ku"   , ku   )
    call get_parameter_real   ("kc"   , kc   )
    call get_parameter_real   ("dtfor", dtfor)
    call get_parameter_real   ("tbfor", tbfor)
    call get_parameter_real   ("tefor", tefor)
    call get_parameter_real   ("tsfor", tsfor)
    call get_parameter_real   ("drate", drate)
    call get_parameter_real   ("famp" , famp )
    call get_parameter_integer("kd"   , kd   )
    call get_parameter_integer("ndiv" , ndiv )
#ifdef SUBVOL
    call get_parameter_real   ("rcut" , rcut )
    call get_parameter_real   ("ycut" , ycut )
    call get_parameter_real   ("rdel" , rdel )
    call get_parameter_real   ("ydel" , ydel )
    call get_parameter_real   ("ysep" , ysep )
#endif /* SUBVOL */

! check parameters
!
    pwfor = max(0.0d+00, pwfor)
    kf    = max(0.0d+00, kf   )
    kc    = max(0.0d+00, kc   )
    kl    = max(0.0d+00, kl   )
    kl    = min(kl   , kf - kc)
    ku    = max(ku   , kf + kc)
    kd    = max(      1, kd   )
    dtfor = max(0.0d+00, dtfor)

! initialize global variables
!
    nf    = 0

#ifdef INT_FOR
! reset the integrals of the injected energies
!
    efor  = 0.0d+00
    cfor  = 0.0d+00
    vfor  = 0.0d+00
    sfor  = 0.0d+00
    afrx  = 0.0d+00
    afry  = 0.0d+00
    afrz  = 0.0d+00
#endif /* INT_FOR */

#ifdef RANDOM
! calculate the width of the eddy corresponding to the injection scale
!
    ewth  = 1.0d+00 / (pi2 * kf)

! prepare temporary arrays
!
    xt(:) = xc(:) - 0.5d+00 * (xmin + xmax)
    yt(:) = yc(:) - 0.5d+00 * (ymin + ymax)
#ifdef R3D
    zt(:) = zc(:) - 0.5d+00 * (zmin + zmax)
#endif /* R3D */
    xt(:) = xt(:) / ewth
    yt(:) = yt(:) / ewth
#ifdef R3D
    zt(:) = zt(:) / ewth
#endif /* R3D */
    xf(:) = exp(- xt(:)**2)
    yf(:) = exp(- yt(:)**2)
#ifdef R3D
    zf(:) = exp(- zt(:)**2)
#endif /* R3D */
    xt(:) = (xt(:) / ewth)**2
    yt(:) = (yt(:) / ewth)**2

! calculate energy injected from one eddy
!
    einj = 0.0d+00
    do k = kb, ke
      do j = jb, je
        do i = ib, ie

! calculate factors
!
#ifdef R3D
          fc = xf(i) * yf(j) * zf(k)
#else /* R3D */
          fc = xf(i) * yf(j)
#endif /* R3D */

! calculate component amplitudes
!
          fx = yt(j)
          fy = xt(i)

! update the injection energy from one eddy
!
          einj = einj + fc * (fx + fy)

        end do
      end do
    end do

#ifdef MPI
! sum the eddy energy from all processors
!
    call reduce_sum_real(einj, iret)
#endif /* MPI */

! normalize energy
!
    einj = einj * dvolh

! calculate the amplitude and energy of one eddy
!
    eamp = sqrt(pwfor / drate / einj)
    einj = eamp * eamp * einj

! print forcing parameters on the screen
!
    if (master) then
      fmt = '(3x,a15,1x,":",'
      write (*, '(1x,a)'       )   'FORCING'
#ifdef FVEL
      write (*, fmt // '1a14)'   ) , 'type           ', 'random in V'
#else /* FVEL */
      write (*, fmt // '1a14)'   ) , 'type           ', 'random in A'
#endif /* FVEL */
      write (*, fmt // '1pe14.6)') , 'injection power', pwfor
      write (*, fmt // '1pe14.6)') , 'injection scale', kf
      write (*, fmt // '1pe14.6)') , 'eddy amplitude ', eamp
      write (*, fmt // '1pe14.6)') , 'eddy width     ', ewth
      write (*, fmt // '1pe14.6)') , 'eddy maximum   ', eamp * ehi / ewth
      write (*, fmt // '1pe14.6)') , 'fourier maximum', eamp * ehi * kf
      write (*, fmt // '1pe14.6)') , 'eddy energy    ', einj
      write (*, fmt // '1pe14.6)') , 'total power    ', einj * drate
      write (*,'(a)')       ''
    end if
#endif /* RANDOM */

#if defined ALVELIUS || defined NEW
! initialize global variables
!
    dtf   = dtfor
    sqdtf = sqrt(dtfor)
    pwf   = pwfor
    kmx   = int(ku + 1)

! reset the number of forcing components and normalization coefficient
!
    l  = 0
    fi = 0.0d+00

! iterate over all Fourier nodes in order to find appropriate forcing components
!
#ifdef R3D
    do i = 0, kmx, kd
      if (i == 0) then
        jbg = 0
      else
        jbg = - kmx
      end if
      do j = jbg, kmx, kd
        if (i == 0 .and. j == 0) then
          kbg = 0
        else
          kbg = - kmx
        end if
        do k = kbg, kmx, kd

          kk = sqrt(real(i*i + j*j + k*k))

          if (kk >= kl .and. kk <= ku) then

            l = l + 1

            kr = (kk - kf) / kc
            fi = fi + exp(- kr**2)

          end if

        end do
      end do
    end do
#else /* R3D */
    do i = 0, kmx, kd
      if (i == 0) then
        jbg = 0
      else
        jbg = - kmx
      end if
      do j = jbg, kmx, kd

        kk = sqrt(real(i*i + j*j))

        if (kk >= kl .and. kk <= ku) then

          l = l + 1

          kr = (kk - kf) / kc
          fi = fi + exp(- kr**2)

        end if

      end do
    end do
#endif /* R3D */

#ifdef SUBVOL
! allocate window arrays
!
    allocate(fr(im,km))
    allocate(fv(jm))

! prepare window factors along the XZ-plane
!
    do k = 1, km
      do i = 1, im

! by default the window function is 1.0
!
        fr(i,k) = 1.0d+00

! calculate radius
!
#ifdef R3D
        rr = xc(i)**2 + zc(k)**2
        rr = (sqrt(rr) - rcut) / rdel
#else /* R3D */
        rr = (abs(xc(i)) - rcut) / rdel
#endif /* R3D */

        if (rr > 0.0d+00) fr(i,k) = exp(- rr * rr)

      end do
    end do

! prepare window factors along the Y direction
!
    do j = 1, jm

! by default the window function is 1.0
!
      fv(j) = 1.0d+00

#ifdef SUBVOL_LAYERS
! center both layers at the same position
!
      yy = yc(j) - sign(1.0d+00, yc(j)) * ysep

! bring the edges to the center
!
      rr = (abs(yy) - ycut) / ydel
#else /* SUBVOL_LAYERS */
! bring the edges to the center
!
      rr = (abs(yc(j)) - ycut) / ydel
#endif /* SUBVOL_LAYERS */

! calculate the Gaussian decay of the edge
!
      if (rr > 0.0d+00) fv(j) = exp(- rr * rr)

    end do

! calculate fraction of space where the injection of turbulence has place
!
    rr = sum(fr(ib:ie,kb:ke)**2) * dx * dz / pdims(2)
    yy = sum(fv(jb:je)**2) * dy / (pdims(1) * pdims(3))
#ifdef MPI
    call reduce_sum_real(rr, iret)
    call reduce_sum_real(yy, iret)
#endif /* MPI */

! correct the factor by the subvolume fraction
!
    fp = rr * yy

! correct the factor by the subvolume fraction
!
    fi = fi * fp
#endif /* SUBVOL */

! set the number of components and prepare forcing power and amplitude
!
    nlf  = l
    nlf3 = 3 * l
    if (pwfor > 0.0d+00) then
#ifdef R3D
      amp = sqrt((4.0d+00 / 3.0d+00) * pi * pwf / fi)
#else /* R3D */
      amp = sqrt((8.0d+00 / 3.0d+00) * pi * pwf / fi)
#endif /* R3D */
    else
      amp = famp
#ifdef R3D
      pwf = 0.750d+00 * amp**2 * fi / pi
#else /* R3D */
      pwf = 0.375d+00 * amp**2 * fi / pi
#endif /* R3D */
    end if

! print forcing parameters on the screen
!
    if (master) then
      fmt = '(3x,a15,1x,":",'
      write (*, '(1x,a)'       )   'FORCING'
#ifdef ALVELIUS
      write (*, fmt // '1a14)'   ) , 'type           ', 'Alvelius in V'
#endif /* ALVELIUS */
#ifdef NEW
      write (*, fmt // '1a14)'   ) , 'type           ', 'Fourier in V'
#endif /* NEW */
      write (*, fmt // '1i14)'   ) , 'no. of modes   ', nlf
      write (*, fmt // '1pe14.6)') , 'injection power', pwfor
      write (*, fmt // '1pe14.6)') , 'injection scale', kf
      write (*, fmt // '1pe14.6)') , 'amplitude      ', amp
      write (*, fmt // '1pe14.6)') , 'norm. coeff.   ', fi
      write (*,'(a)')       ''
    end if

! multiply amplitude by the square of forcing time step so we don't have to
! multiply this at each step
!
    if (pwfor > 0.0d+00) then
      amp = sqrt(dtf) * amp
    else
      pwf = pwf / dtf
    end if

! allocate global forcing tables
!
    allocate(f_table(nlf))
    allocate(k_table(nlf,3))
    allocate(e1     (nlf,3))
    allocate(e2     (nlf,3))
    allocate(ffor   (nlf,3))
    allocate(fvel   (nlf,3))

! reset the forcing mode counter
!
    l = 0

! iterate over all forcing modes and prepare helper arrays
!
#ifdef R3D
    do i = 0, kmx, kd
      if (i == 0) then
        jbg = 0
      else
        jbg = - kmx
      end if
      do j = jbg, kmx, kd
        if (i == 0 .and. j == 0) then
          kbg = 0
        else
          kbg = - kmx
        end if
        do k = kbg, kmx, kd

          kk = sqrt(real(i*i + j*j + k*k))

          if (kk >= kl .and. kk <= ku) then

            l = l + 1

! compute table of isotropic k vectors
!
            k_table(l,1) = i
            k_table(l,2) = j
            k_table(l,3) = k

! compute forcing amplitude
!
            kr = (kk - kf) / kc
            f_table(l) = amp * exp(- 0.5d+00 * kr**2)

! compute wave vector components
!
            kx = i
            ky = j
            kz = k

#ifdef FXDIR
! compute components of the unit vectors describing the direction of the two
! force terms
!
            kyz = sqrt(ky**2 + kz**2)

            if (kyz > 0.0d+00) then
              e1(l,1) =   0.0d+00
              e1(l,2) =   kz / kyz
              e1(l,3) = - ky / kyz

              e2(l,1) = - kyz / kk
              e2(l,2) =   kx*ky / (kk*kyz)
              e2(l,3) =   kx*kz / (kk*kyz)
            else
              kxz = sqrt(kx**2 + kz**2)

              e1(l,1) = - kz / kxz
              e1(l,2) =   0.0d+00
              e1(l,3) =   kx / kxz

              e2(l,1) =   kx*ky / (kk*kxz)
              e2(l,2) = - kxz / kk
              e2(l,3) =   kz*ky / (kk*kxz)
            end if
#endif /* FXDIR */
#ifdef FYDIR
! compute components of the unit vectors describing the direction of the two
! force terms
!
            kxz = sqrt(kx**2 + kz**2)

            if (kxz > 0.0d+00) then
              e1(l,1) = - kz / kxz
              e1(l,2) =   0.0d+00
              e1(l,3) =   kx / kxz

              e2(l,1) =   kx*ky / (kk*kxz)
              e2(l,2) = - kxz / kk
              e2(l,3) =   kz*ky / (kk*kxz)
            else
              kxy = sqrt(kx**2 + ky**2)

              e1(l,1) =   ky / kxy
              e1(l,2) = - kx / kxy
              e1(l,3) =   0.0d+00

              e2(l,1) =   kx*kz / (kk*kxy)
              e2(l,2) =   ky*kz / (kk*kxy)
              e2(l,3) = - kxy / kk
            end if
#endif /* FYDIR */
#ifdef FZDIR
! compute components of the unit vectors describing the direction of the two
! force terms
!
            kxy = sqrt(kx**2 + ky**2)

            if (kxy > 0.0d+00) then
              e1(l,1) =   ky / kxy
              e1(l,2) = - kx / kxy
              e1(l,3) =   0.0d+00

              e2(l,1) =   kx*kz / (kk*kxy)
              e2(l,2) =   ky*kz / (kk*kxy)
              e2(l,3) = - kxy / kk
            else
              kyz = sqrt(ky**2 + kz**2)

              e1(l,1) =   0.0d+00
              e1(l,2) =   kz / kyz
              e1(l,3) = - ky / kyz

              e2(l,1) = - kyz / kk
              e2(l,2) =   kx*ky / (kk*kyz)
              e2(l,3) =   kx*kz / (kk*kyz)
            end if
#endif /* FZDIR */

          end if

        end do
      end do
    end do
#else /* R3D */
    do i = 0, kmx, kd
      if (i == 0) then
        jbg = 0
      else
        jbg = - kmx
      end if
      do j = jbg, kmx, kd

        kk = sqrt(real(i*i + j*j))

        if (kk >= kl .and. kk <= ku) then

          l = l + 1

! compute table of isotropic k vectors
!
          k_table(l,1) = i
          k_table(l,2) = j
          k_table(l,3) = 0

! compute forcing amplitude
!
          kr = (kk - kf) / kc
          f_table(l) = amp * exp(- 0.5d+00 * kr**2)

! compute wave vector components
!
          kx = i
          ky = j
          kz = 0

! compute components of the unit vectors describing the direction of the two
! force terms
!
          kxy = sqrt(kx**2 + ky**2)

          e1(l,1) =   ky / kxy
          e1(l,2) = - kx / kxy
          e1(l,3) =   0.0d+00

          e2(l,1) =   0.0d+00
          e2(l,2) =   0.0d+00
          e2(l,3) =   1.0d+00

        end if
      end do
    end do
#endif /* R3D */

! prepare sinus and cosines
!
    allocate(sini(0:prec))
    allocate(cosi(0:prec))
    do i = 0, prec
      sini(i) = sin((pi2 * i) / prec)
      cosi(i) = cos((pi2 * i) / prec)
    end do

! prepare anisotropy functions
!
    allocate(ga(0:prec))
    allocate(gb(0:prec))
#ifdef FANISO
    if (fani >= 1.0d+00) then
      do i = 0, prec
        gb(i) = tanh(fani * pi * (1.0d+00 * i / prec - 0.5d+00))               &
                                                   / tanh(0.5d+00 * pi * fani)
        gb(i) = min(1.0d+00, max(- 1.0d+00, - gb(i)))
        ga(i) = sqrt(max(0.0d+00, 1.0d+00 - gb(i)**2))
      end do
    else
      do i = 0, prec
        ga(i) = tanh(pi * (1.0d+00 * i / prec - 0.5d+00) / fani)               &
                                                   / tanh(0.5d+00 * pi / fani)
        ga(i) = min(1.0d+00, max(- 1.0d+00, - ga(i)))
        gb(i) = sqrt(max(0.0d+00, 1.0d+00 - ga(i)**2))
      end do
    end if
#else
    do i = 0, prec
      ga(i) = sini(i)
      gb(i) = cosi(i)
    end do
#endif /* FANISO */

#ifdef FAST
! check if the conditions for fast forcing are fulfilled
!
  if (dx /= dy) then
    if (master) then
      write (*,'(a)')       ''
      write (*,'(a,a)') 'ERROR   : ', ' Conditions for fast Fourier' //        &
                                      ' transform are not fulfilled!'
      write (*,'(a,a)') 'ERROR   : ', ' It requires dx = dy = dz!'
      write (*,'(a)')       ''
    end if
    stop
  end if
#ifdef R3D
  if (dx /= dz) then
    if (master) then
      write (*,'(a)')       ''
      write (*,'(a,a)') 'ERROR   : ', ' Conditions for fast Fourier' //        &
                                      ' transform are not fulfilled!'
      write (*,'(a,a)') 'ERROR   : ', ' It requires dx = dy = dz!'
      write (*,'(a)')       ''
    end if
    stop
  end if
#endif /* R3D */

! allocate arrays for indices
!
    allocate(kix(nlf,im))
    allocate(kiy(nlf,jm))
    allocate(kiz(nlf,km))

! fill the index arrays
!
    do i = 1, im
      ix = 2 * int((xc(i) - floor(xc(i))) / dx) + 1
      do l = 1, nlf
        kix(l,i) = k_table(l,1) * ix
      end do
    end do
    do j = 1, jm
      iy = 2 * int((yc(j) - floor(yc(j))) / dy) + 1
      do l = 1, nlf
        kiy(l,j) = k_table(l,2) * iy
      end do
    end do
    lmx = maxval(abs(kix)) + maxval(abs(kiy))
#ifdef R3D
    do k = 1, km
      iz = 2 * int((zc(k) - floor(zc(k))) / dz) + 1
      do l = 1, nlf
        kiz(l,k) = k_table(l,3) * iz
      end do
    end do
    lmx = lmx + maxval(abs(kiz))
#endif /* R3D */

! allocate temporary arrays for Fourier transforms
!
    allocate(sinq(-lmx:lmx))
    allocate(cosq(-lmx:lmx))

! prepare temporary arrays for Fourier transforms
!
    pidx = pi * dx
    do i = - lmx, lmx
      phi     = pidx * i
      sinq(i) = sin(phi)
      cosq(i) = cos(phi)
    end do
#else /* FAST */
! allocate temporary arrays for Fourier transforms
!
    allocate(sinx(-kmx:kmx,im))
    allocate(cosx(-kmx:kmx,im))
    allocate(siny(-kmx:kmx,jm))
    allocate(cosy(-kmx:kmx,jm))
    allocate(sinz(-kmx:kmx,km))
    allocate(cosz(-kmx:kmx,km))

! prepare temporary arrays for Fourier transforms
!
    do l = - kmx, kmx
      do i = 1, im
        phi       = pi2 * l * xc(i)
        sinx(l,i) = sin(phi)
        cosx(l,i) = cos(phi)
      end do
      do j = 1, jm
        phi       = pi2 * l * yc(j)
        siny(l,j) = sin(phi)
        cosy(l,j) = cos(phi)
      end do
      do k = 1, km
        phi       = pi2 * l * zc(k)
        sinz(l,k) = sin(phi)
        cosz(l,k) = cos(phi)
      end do
    end do
#endif /* FAST */
#endif /* ALVELIUS | NEW */

#ifdef PROFILE
! stop accounting time for forcing
!
    call stop_timer(ifi)
#endif /* PROFILE */

!-------------------------------------------------------------------------------
!
  end subroutine initialize_forcing
!
!===============================================================================
!
! subroutine FINALIZE_FORCING:
! ---------------------------
!
!   Subroutine deallocates all local forcing arrays.
!
!===============================================================================
!
  subroutine finalize_forcing()

! local variables are not implicit by default
!
    implicit none
!
!-------------------------------------------------------------------------------
!
#ifdef PROFILE
! start accounting time for forcing
!
    call start_timer(ifi)
#endif /* PROFILE */

#ifndef RANDOM
    if (allocated(k_table)) deallocate(k_table)
    if (allocated(f_table)) deallocate(f_table)
    if (allocated(e1))      deallocate(e1)
    if (allocated(e2))      deallocate(e2)
    if (allocated(ffor))    deallocate(ffor)
    if (allocated(fvel))    deallocate(fvel)
    if (allocated(sini))    deallocate(sini)
    if (allocated(cosi))    deallocate(cosi)
    if (allocated(ga))      deallocate(ga)
    if (allocated(gb))      deallocate(gb)
#ifdef FAST
    if (allocated(kix))     deallocate(kix)
    if (allocated(kiy))     deallocate(kiy)
    if (allocated(kiz))     deallocate(kiz)
    if (allocated(sinq))    deallocate(sinq)
    if (allocated(cosq))    deallocate(cosq)
#else /* FAST */
    if (allocated(sinx))    deallocate(sinx)
    if (allocated(cosx))    deallocate(cosx)
    if (allocated(siny))    deallocate(siny)
    if (allocated(cosy))    deallocate(cosy)
    if (allocated(sinz))    deallocate(sinz)
    if (allocated(cosz))    deallocate(cosz)
#endif /* FAST */
#ifdef SUBVOL
    if (allocated(fr))      deallocate(fr)
    if (allocated(fv))      deallocate(fv)
#endif /* SUBVOL */
#endif /* !RANDOM */

#ifdef PROFILE
! stop accounting time for forcing
!
    call stop_timer(ifi)
#endif /* PROFILE */

!-------------------------------------------------------------------------------
!
  end subroutine finalize_forcing
!
!===============================================================================
!
! subroutine DRIVE:
! ----------------
!
!   Subroutine calls the forcing evolution procedure to advance the forcing
!   term in time. The forcing evolution method is selected through
!   a compilation flag FORCING_METHOD.
!
!===============================================================================
!
  subroutine drive()

! include external procedures and variables
!
    use mesh, only : time

! local variables are not implicit by default
!
    implicit none
!
!-------------------------------------------------------------------------------
!
! the injection time is limited by two parameters, tbfor and tsfor. If the time
! is not between those parameters, we don't inject turbulence.
!
    if (time < tbfor) return
    if (time > tsfor) return

#if !defined RANDOM && defined FCORR
! calculate Fourier coefficients of velocity
!
    call real_to_fourier()
#endif /* !RANDOM & FCORR */

#ifdef PROFILE
! start accounting time of forcing evolution
!
    call start_timer(ifc)
#endif /* PROFILE */

! evolve the forcing term in Fourier space
!
#ifdef RANDOM
    call drive_random()
#else /* RANDOM */
#ifdef ALVELIUS
    call drive_alvelius()
#endif /* ALVELIUS */
#ifdef NEW
    call drive_new()
#endif /* NEW */
#endif /* RANDOM */

#ifdef PROFILE
! stop accounting time for the forcing evolution
!
    call stop_timer(ifc)
#endif /* PROFILE */

#ifndef RANDOM
! transform the forcing term from the Fourier to real space and update
! the conservative variables with the forcing term
!
    call fourier_to_real()
#endif /* !RANDOM */

!-------------------------------------------------------------------------------
!
  end subroutine drive
#ifdef RANDOM
!
!===============================================================================
!
! subroutine DRIVE_RANDOM:
! -----------------------
!
!   Subroutine calls the forcing evolution procedure to advance in time the
!   forcing term.  This subroutine adds a random small scale component.
!
!   References:
!
!     [1] Kowal, G., Lazarian, A., Vishniac, E. T., & Otmianowska-Mazur, K.,
!         "Reconnection Studies under Different Types of Turbulence Driving",
!         Nonlinear Processes in Geophysics, 2012, 19, 297-314
!
!===============================================================================
!
  subroutine drive_random()

! include external procedures and variables
!
#ifdef SUBVOL
    use constants, only : pi
#endif /* SUBVOL */
#ifdef INT_FOR
    use mesh     , only : ib, ie, jb, je, kb, ke
    use mesh     , only : dvolh
#endif /* INT_FOR */
    use mesh     , only : im, jm, km
    use mesh     , only : xc, yc, zc, xi, yi, zi, time, dt
#ifndef SUBVOL
    use mesh     , only : xmin, ymin, zmin
#endif /* !SUBVOL */
    use mesh     , only : xlen, ylen, zlen
    use mpitools , only : periodic
    use random   , only : randomn, randomu
    use variables, only : u0
#ifdef FVEL
    use variables, only : du
    use variables, only : idn, imx, imy, imz, ivx, ivy, ivz
#endif /* FVEL */
#ifdef FMAG
#ifdef VPOT
    use variables, only : iax, iay, iaz
#endif /* VPOT */
#endif /* FMAG */
#ifdef ADI
    use variables, only : ien
#endif /* ADI */

! local variables are not implicit by default
!
    implicit none

! local variables
!
    integer :: i, j, k, ni
    real    :: xl, yl, zl, rx, ry, rz
    real    :: fx, fy, fz, fc
    real    :: rate, amp
    real    :: en
#ifdef INT_FOR
    integer :: ith = 0
#endif /* INT_FOR */

#ifdef SUBVOL
! saved variables
!
    logical, save :: first = .true.
    real   , save :: ff    = 1.0d+00
#endif /* SUBVOL */

! local arrays
!
    real        , dimension(im) :: xt, xf
    real        , dimension(jm) :: yt, yf
    real        , dimension(km) :: zt, zf

#ifdef INT_FOR
! variables to store force correlations
!
    real(kind=8)                :: ef, cf, vf, sf
    real(kind=8), dimension(3)  :: af
#endif /* INT_FOR */
!
!-------------------------------------------------------------------------------
!
#ifdef SUBVOL
! prepare save coefficients at the first call
!
    if (first) then

! calculate factor corresponding to the time of injection
!
      ff    = 0.5d+00 * pi / max(dt, tefor - tbfor)

! reset the first call run
!
      first = .false.

    end if

! calculate the injection rate, the time factor, and the eddy amplitude for
! the case of subvolume injection
!
    rate = nf / max(dt, time - tbfor)
    amp  = eamp
    if (time < tefor) then
      fc  = sin(ff * max(0.0d+00, time - tbfor))
      amp = fc * fc * eamp
    end if
#else /* SUBVOL */
! calculate the injection rate and eddy amplitude for the full domain injection
!
    rate = nf / max(dt, time)
    amp  = eamp
#endif /* SUBVOL */

! return, if the injection rate is larger than the desired, otherwise inject
! more eddies
!
    if (rate >= drate) return

! calculate the number of injections per time step
!
    ni = int(drate * dt) + 1

#ifdef FVEL
! reset array dU(:,:,:,:) which temporarily stores the velocity increment
!
#ifdef R2D
    du(ivx:ivy,:,:,:) = 0.0d+00
#else /* R2D */
    du(ivx:ivz,:,:,:) = 0.0d+00
#endif /* R2D */
#endif /* FVEL */

! inject ni-eddies
!
    do while (ni > 0)

! obtain the random directional vector of the injected eddy
!
#ifdef R2D
      fc = 0.0d+00
      do while (fc <= 0.0d+00)
        fz = randomn()
        fc = abs(fz)
      end do

      fz = amp * fz / fc
#else /* R2D */
      fc = 0.0d+00
      do while (fc <= 0.0d+00)
        fx = randomn()
        fy = randomn()
        fz = randomn()
        fc = sqrt(fx * fx + fy * fy + fz * fz)
      end do

      fx = amp * fx / fc
      fy = amp * fy / fc
      fz = amp * fz / fc
#ifdef R2D5
      fz = sqrt(2.0d+00) * fz
#endif /* R2D5 */
#endif /* R2D */

! obtain the random positions of the injected eddy withing the injection region
!
#ifdef SUBVOL
      xl = rcut * randomn()
      yl = ycut * randomn()
#ifdef R3D
      zl = rcut * randomn()
#endif /* R3D */
      yl = yl + sign(ysep, yl)
#else /* SUBVOL */
      xl = xlen * randomu() + xmin
      yl = ylen * randomu() + ymin
#ifdef R3D
      zl = zlen * randomu() + zmin
#endif /* R3D */
#endif /* SUBVOL */

#ifdef FVEL
! prepare the renormalized coordinates (xt, yt, zt) and the injection factors
! (xf, yf, zf) along each direction
!
      xt(:) = xc(:) - xl
      if (periodic(1)) then
        do i = 1, im
          rx    = xt(i) / xlen   + 0.5d+00
          rx    = rx - floor(rx) - 0.5d+00
          xt(i) = xlen * rx
        end do
      end if
      xt(:) = xt(:) / ewth
      xf(:) = exp(- 0.5d+00 * xt(:)**2)
      xt(:) = xt(:) / ewth

      yt(:) = yc(:) - yl
      if (periodic(2)) then
        do j = 1, jm
          ry    = yt(j) / ylen   + 0.5d+00
          ry    = ry - floor(ry) - 0.5d+00
          yt(j) = ylen * ry
        end do
      end if
      yt(:) = yt(:) / ewth
      yf(:) = exp(- 0.5d+00 * yt(:)**2)
      yt(:) = yt(:) / ewth

#ifdef R3D
      zt(:) = zc(:) - zl
      if (periodic(3)) then
        do k = 1, km
          rz    = zt(k) / zlen   + 0.5d+00
          rz    = rz - floor(rz) - 0.5d+00
          zt(k) = zlen * rz
        end do
      end if
      zt(:) = zt(:) / ewth
      zf(:) = exp(- 0.5d+00 * zt(:)**2)
      zt(:) = zt(:) / ewth
#endif /* R3D */

! iterate over all position in the domain
!
      do k = 1, km
        do j = 1, jm
          do i = 1, im

! calculate the amplitude factor from the directional factors
!
#ifdef R3D
            fc = xf(i) * yf(j) * zf(k)
#else /* R3D */
            fc = xf(i) * yf(j)
#endif /* R3D */

! calculate the vector components using the curl of injection vector potential
! in order to get the divergence free forcing
!
#ifdef R3D
            rx = fc * (fy * zt(k) - fz * yt(j))
            ry = fc * (fz * xt(i) - fx * zt(k))
            rz = fc * (fx * yt(j) - fy * xt(i))
#else /* R3D */
            rx = fc * (           - fz * yt(j))
            ry = fc * (fz * xt(i)             )
#ifndef R2D
            rz = fc * (fx * yt(j) - fy * xt(i))
#endif /* !R2D */
#endif /* R3D */

! update the velocity components at the current position
!
#ifdef R2D
            du(ivx,i,j,k) = du(ivx,i,j,k) + rx
            du(ivy,i,j,k) = du(ivy,i,j,k) + ry
#else /* R2D */
            du(ivx,i,j,k) = du(ivx,i,j,k) + rx
            du(ivy,i,j,k) = du(ivy,i,j,k) + ry
            du(ivz,i,j,k) = du(ivz,i,j,k) + rz
#endif /* R2D */
          end do
        end do
      end do
#endif /* FVEL */

#if defined VPOT && defined FMAG
! prepare the renormalized coordinates (xt, yt, zt) and the injection factors
! (xf, yf, zf) along each direction
!
      xt(:) = xc(:) - xl
      xf(:) = xi(:) - xl
      if (periodic(1)) then
        do i = 1, im
          rx    = xt(i) / xlen   + 0.5d+00
          rx    = rx - floor(rx) - 0.5d+00
          xt(i) = xlen * rx

          rx    = xf(i) / xlen   + 0.5d+00
          rx    = rx - floor(rx) - 0.5d+00
          xf(i) = xlen * rx
        end do
      end if
      xt(:) = exp(- 0.5d+00 * (xt(:) / ewth)**2)
      xf(:) = exp(- 0.5d+00 * (xf(:) / ewth)**2)

      yt(:) = yc(:) - yl
      yf(:) = yi(:) - yl
      if (periodic(2)) then
        do j = 1, jm
          ry    = yt(j) / ylen   + 0.5d+00
          ry    = ry - floor(ry) - 0.5d+00
          yt(j) = ylen * ry

          ry    = yf(j) / ylen   + 0.5d+00
          ry    = ry - floor(ry) - 0.5d+00
          yf(j) = ylen * ry
        end do
      end if
      yt(:) = exp(- 0.5d+00 * (yt(:) / ewth)**2)
      yf(:) = exp(- 0.5d+00 * (yf(:) / ewth)**2)
#ifdef R3D
      zt(:) = zc(:) - zl
      zf(:) = zi(:) - zl
      if (periodic(3)) then
        do k = 1, km
          rz    = zt(k) / zlen   + 0.5d+00
          rz    = rz - floor(rz) - 0.5d+00
          zt(k) = zlen * rz

          rz    = zf(k) / zlen   + 0.5d+00
          rz    = rz - floor(rz) - 0.5d+00
          zf(k) = zlen * rz
        end do
      end if
      zt(:) = exp(- 0.5d+00 * (zt(:) / ewth)**2)
      zf(:) = exp(- 0.5d+00 * (zf(:) / ewth)**2)
#endif /* R3D */

! iterate over all position in the domain
!
      do k = 1, km
        do j = 1, jm
          do i = 1, im
#ifdef R2D
! calculate the vector potential factors
!
            rz = xf(i) * yf(j)

! update the vector potential components
!
            u0(iaz,i,j,k) = u0(iaz,i,j,k) + fz * rz
#else /* R2D */
! calculate the vector potential factors
!
#ifdef R3D
            rx = xt(i) * yf(j) * zf(k)
            ry = xf(i) * yt(j) * zf(k)
            rz = xf(i) * yf(j) * zt(k)
#else /* R3D */
            rx = xt(i) * yf(j)
            ry = xf(i) * yt(j)
            rz = xf(i) * yf(j)
#endif /* R3D */

! update the vector potential components
!
            u0(iax,i,j,k) = u0(iax,i,j,k) + fx * rx
            u0(iay,i,j,k) = u0(iay,i,j,k) + fy * ry
            u0(iaz,i,j,k) = u0(iaz,i,j,k) + fz * rz
#endif /* R2D */

          end do
        end do
      end do
#endif /* VPOT & FMAG */

! increase the total number of injected eddies
!
      nf = nf + 1

! decrease the number of injected eddies in this loop
!
      ni = ni - 1

    end do

#ifdef INT_FOR
! reset the forcing correlation integrals
!
    ef    = 0.0d+00
    cf    = 0.0d+00
    vf    = 0.0d+00
    sf    = 0.0d+00
    af(:) = 0.0d+00
#endif /* INT_FOR */

#ifdef FVEL
! iterate over all position in the domain
!
    do k = 1, km
      do j = 1, jm
        do i = 1, im

#if defined INT_FOR || defined ADI
! calculate the kinetic energy before injection
!
          en   = sum(u0(imx:imz,i,j,k)**2) / u0(idn,i,j,k)
#endif /* INT_FOR | ADI */

! get the forcing components at current position
!
          fx = du(ivx,i,j,k)
          fy = du(ivy,i,j,k)
#ifndef R2D
          fz = du(ivz,i,j,k)
#endif /* !R2D */

! update the momentum components
!
          u0(imx,i,j,k) = u0(imx,i,j,k) + u0(idn,i,j,k) * fx
          u0(imy,i,j,k) = u0(imy,i,j,k) + u0(idn,i,j,k) * fy
#ifndef R2D
          u0(imz,i,j,k) = u0(imz,i,j,k) + u0(idn,i,j,k) * fz
#endif /* !R2D */

#if defined INT_FOR || defined ADI
! calculate the injected kinetic energy
!
          en = sum(u0(imx:imz,i,j,k)**2) / u0(idn,i,j,k) - en

#ifdef ADI
! add the injected kinetic energy do the total energy
!
          u0(ien,i,j,k) = u0(ien,i,j,k) + 0.5d+00 * en
#endif /* ADI */
#endif /* INT_FOR | ADI */

#ifdef INT_FOR
! calculate the injected energies
!
          du(idn,i,j,k) = en
          du(ivx,i,j,k) = fx * fx
          du(ivy,i,j,k) = fy * fy
#ifndef R2D
          du(ivz,i,j,k) = fz * fz
#endif /* !R2D */
#endif /* INT_FOR */

        end do
      end do
    end do

#ifdef INT_FOR
! iterate over all position in the domain
!
    do k = kb, ke
      do j = jb, je
        do i = ib, ie

! calculate the injected energy due to the force-force correlation
!
          en = u0(idn,i,j,k) * (du(ivx,i,j,k) + du(ivy,i,j,k) + du(ivz,i,j,k))

! sum up the correlation integrals
!
          sf    = sf    +  du(idn,i,j,k) / u0(idn,i,j,k)
          ef    = ef    +  du(idn,i,j,k)
          cf    = cf    +  en
          vf    = vf    + (du(idn,i,j,k) - en)
          af(1) = af(1) +  du(ivx,i,j,k) * u0(idn,i,j,k)
          af(2) = af(2) +  du(ivy,i,j,k) * u0(idn,i,j,k)
          af(3) = af(3) +  du(ivz,i,j,k) * u0(idn,i,j,k)
        end do
      end do
    end do
#endif /* INT_FOR */
#endif /* FVEL */

#ifdef INT_FOR
! sum up all injected energies normalized by the cell volume
!
    efor = efor + dvolh * ef
    cfor = cfor + dvolh * cf
    vfor = vfor + dvolh * vf
    sfor = sfor + dvolh * sf
    afrx = afrx + dvolh * af(1)
    afry = afry + dvolh * af(2)
    afrz = afrz + dvolh * af(3)
#endif /* INT_FOR */

!-------------------------------------------------------------------------------
!
  end subroutine drive_random
#else /* RANDOM */
#ifdef ALVELIUS
!
!===============================================================================
!
! subroutine DRIVE_ALVELIUS:
! -------------------------
!
!   Subroutine calls the forcing evolution procedure to advance in time the
!   forcing term.  This subroutine uses a forcing evolution method by
!   Alvelius (1999).
!
!   References:
!
!     [1] Alvelius, K.,
!         "Random forcing of three-dimensional homogeneous turbulence",
!         Physics of Fluids, 1999, 11, 7, 1880-1889
!
!===============================================================================
!
  subroutine drive_alvelius()

! include external procedures and variables
!
    use constants , only : pi2
    use mesh      , only : time
    use random    , only : randomu

! local variables are not implicit by default
!
    implicit none

! local variables
!
    integer            :: i, l
    integer            :: nnf, nif
    integer            :: ith1, ith2, iphi, ipsi
    real               :: psi, th1, th2, phi
    complex            :: aran, bran
#ifdef FCORR
    real               :: xi1re, xi1im, xi2re, xi2im, tanth1, div

! local arrays
!
    real, dimension(3) :: fre, fim

! local parameters
!
    real, parameter    :: eps = epsilon(div)
#endif /* FCORR */
!
!-------------------------------------------------------------------------------
!
! compute the total number of forcing iterations and number of iterations
! to perform in this step
!
#ifdef SUBVOL
    nnf = nint((time - tbfor) / dtf)
#else /* SUBVOL */
    nnf = nint(time / dtf)
#endif /* SUBVOL */
    nif = max(0, nnf - nf)
    nf  = nnf

! iterate over all forcing components
!
    do l = 1, nlf

#ifdef FCORR
! separate real and imaginary parts of the velocity Fourier coefficients into
! different arrays
!
      fre(1:3) = real (fvel(l,1:3))
      fim(1:3) = aimag(fvel(l,1:3))

! project the velocity components on the e₁ and e₂ directions
!
      xi1re =   sum(fre(1:3) * e1(l,1:3))
      xi1im = - sum(fim(1:3) * e1(l,1:3))
      xi2re =   sum(fre(1:3) * e2(l,1:3))
      xi2im = - sum(fim(1:3) * e2(l,1:3))
#endif /* FCORR */

! reset the forcing component amplitudes
!
      aran = cmplx(0.0d+00, 0.0d+00)
      bran = cmplx(0.0d+00, 0.0d+00)

! evolve the forcing for the interva of the current time step
!
      do i = 1, nif

! obtain the random phase φ
!
        phi  = randomu()
        iphi = nint(prec * phi)

#ifdef FCORR
! obtain the random phase ψ
!
        psi  = randomu()
        ipsi = nint(prec * psi)

! calculate tan(θ₁) (eq. 15 in [1])
!
        div  = - ga(iphi) * xi1im                                              &
                        + gb(iphi) * (sini(ipsi) * xi2re - cosi(ipsi) * xi2im)
        if (abs(div) .gt. eps) then
          tanth1 = ga(iphi) * xi1re                                            &
                        + gb(iphi) * (sini(ipsi) * xi2im + cosi(ipsi) * xi2re)
          tanth1 = tanth1 / div
        else
          tanth1 = 0.0d+00
        end if

! calculate angles θ₁ and θ₂
!
        th1  = atan(tanth1) / pi2
        th1  = th1 - floor(th1)
        th2  = psi + th1
        th2  = th2 - floor(th2)

        ith1 = nint(prec * th1)
        ith2 = nint(prec * th2)
#else /* FCORR */
! obtain random phases θ₁ and θ₂
!
        th1  = randomu()
        th2  = randomu()

        ith1 = nint(prec * th1)
        ith2 = nint(prec * th2)
#endif /* FCORR */

! update the forcing component amplitudes (eq. 13 in [1])
!
        aran = aran + ga(iphi) * cmplx(cosi(ith1), sini(ith1))
        bran = bran + gb(iphi) * cmplx(cosi(ith2), sini(ith2))

      end do

! calculate the Fourier components of forcing for this step
!
      ffor(l,1) = f_table(l) * (aran * e1(l,1) + bran * e2(l,1))
      ffor(l,2) = f_table(l) * (aran * e1(l,2) + bran * e2(l,2))
      ffor(l,3) = f_table(l) * (aran * e1(l,3) + bran * e2(l,3))

    end do

!-------------------------------------------------------------------------------
!
  end subroutine drive_alvelius
#endif /* ALVELIUS */
#ifdef NEW
!
!===============================================================================
!
! subroutine DRIVE_NEW:
! --------------------
!
!   Subroutine calls the forcing evolution procedure to advance in time the
!   forcing term.  This subroutine uses an improved (faster) version of
!   the forcing evolution method by Alvelius (1999).
!
!   References:
!
!     [1] Alvelius, K.,
!         "Random forcing of three-dimensional homogeneous turbulence",
!         Physics of Fluids, 1999, 11, 7, 1880-1889
!
!===============================================================================
!
  subroutine drive_new()

! include external procedures and variables
!
    use constants , only : pi2
    use mesh      , only : time, dt
    use random    , only : randomu

! local variables are not implicit by default
!
    implicit none

! local variables
!
    integer            :: l
    integer            :: ith1, ith2, iphi, ipsi
    real               :: psi, th1, th2, phi
    real               :: fpw, fac
    complex            :: aran, bran
#ifdef FCORR
    real               :: xi1re, xi1im, xi2re, xi2im, tanth1, div

! local arrays
!
    real, dimension(3) :: fre, fim

! local parameters
!
    real, parameter    :: eps = epsilon(div)
#endif /* FCORR */
!
!-------------------------------------------------------------------------------
!
! iterate over all forcing components
!
    do l = 1, nlf

#ifdef FCORR
! separate real and imaginary parts of the velocity Fourier coefficients into
! different arrays
!
      fre(1:3) = real (fvel(l,1:3))
      fim(1:3) = aimag(fvel(l,1:3))

! project the velocity components on the e₁ and e₂ directions
!
      xi1re =   sum(fre(1:3) * e1(l,1:3))
      xi1im = - sum(fim(1:3) * e1(l,1:3))
      xi2re =   sum(fre(1:3) * e2(l,1:3))
      xi2im = - sum(fim(1:3) * e2(l,1:3))
#endif /* FCORR */

! obtain the random phase φ
!
      phi  = randomu()
      iphi = nint(prec * phi)

#ifdef FCORR
! obtain the random phase ψ
!
      psi  = randomu()
      ipsi = nint(prec * psi)

! calculate tan(θ₁) (eq. 15 in [1])
!
      div  = - ga(iphi) * xi1im                                                &
                        + gb(iphi) * (sini(ipsi) * xi2re - cosi(ipsi) * xi2im)
      if (abs(div) .gt. eps) then
        tanth1 = ga(iphi) * xi1re                                              &
                        + gb(iphi) * (sini(ipsi) * xi2im + cosi(ipsi) * xi2re)
        tanth1 = tanth1 / div
      else
        tanth1 = 0.0d+00
      end if

! calculate angles θ₁ and θ₂
!
      th1  = atan(tanth1) / pi2
      th1  = th1 - floor(th1)
      th2  = psi + th1
      th2  = th2 - floor(th2)

      ith1 = nint(prec * th1)
      ith2 = nint(prec * th2)
#else /* FCORR */
! obtain random phases θ₁ and θ₂
!
      th1  = randomu()
      th2  = randomu()

      ith1 = nint(prec * th1)
      ith2 = nint(prec * th2)
#endif /* FCORR */

! update the forcing component amplitudes (eq. 13 in [1])
!
      aran = ga(iphi) * cmplx(cosi(ith1), sini(ith1))
      bran = gb(iphi) * cmplx(cosi(ith2), sini(ith2))

! calculate the Fourier components of forcing for this step
!
      ffor(l,1) = f_table(l) * (aran * e1(l,1) + bran * e2(l,1))
      ffor(l,2) = f_table(l) * (aran * e1(l,2) + bran * e2(l,2))
      ffor(l,3) = f_table(l) * (aran * e1(l,3) + bran * e2(l,3))

    end do

! normalize amplitudes to the correct input power
!
    fpw       = abs(sum(ffor(:,:) * conjg(ffor(:,:))))
#ifdef SUBVOL
    fpw       = fpw * fp
#endif /* SUBVOL */
#ifdef R3D
    fac       = sqrt(4.0d+00 * dt * pwfor / fpw)
#else /* R3D */
    fac       = sqrt(8.0d+00 * dt * pwfor / fpw)
#endif /* R3D */
    ffor(:,:) = fac * ffor(:,:)

!-------------------------------------------------------------------------------
!
  end subroutine drive_new
#endif /* NEW */
!
!===============================================================================
!
! subroutine REAL_TO_FOURIER:
! --------------------------
!
!   Subroutine calculates the fourier coefficients of velocity corresponding
!   to the forcing modes. These coefficients are necessary to reduce
!   the velocity-force correlation.
!
!===============================================================================
!
  subroutine real_to_fourier()

! include external procedures and variables
!
    use mesh     , only : ib, jb, kb, ie, je, ke
    use mesh     , only : dvol
    use variables, only : qp
    use variables, only : ivx, ivy, ivz
#ifdef MPI
    use mpitools , only : reduce_sum_complex_array
#endif /* MPI */

! local variables are not implicit by default
!
    implicit none

! local variables
!
#ifdef MPI
    integer :: iret
#endif /* MPI */
    integer :: i, j, k, l
    real    :: sn, cs
#ifdef FAST
    integer :: q
#else /* FAST */
    integer :: li, lj, lk
    real    :: snx, sny, snz, snp, csx, csy, csz, csp
#endif /* FAST */
    complex :: ph
!
!-------------------------------------------------------------------------------
!
#ifdef PROFILE
! start accounting time for the real to fourier transform
!
    call start_timer(ifr)
#endif /* PROFILE */

! reset the fourier components of velocity
!
    fvel(:,:) = cmplx(0.0d+00, 0.0d+00)

! iterate over all positions
!
#ifdef R3D
    do k = kb, ke, ndiv
#else /* R3D */
    do k = kb, ke
#endif /* R3D */
      do j = jb, je, ndiv
        do i = ib, ie, ndiv

! iterate over all forcing components
!
          do l = 1, nlf

#ifdef FAST
! calculate the fast fourier array index
!
#ifdef R3D
            q  = kix(l, i) + kiy(l, j) + kiz(l, k)
#else /* R3D */
            q  = kix(l, i) + kiy(l, j)
#endif /* R3D */

! get the fourier transform coefficients
!
            sn = sinq(q)
            cs = cosq(q)
#else /* FAST */
! get the directional wave numbers
!
            li = k_table(l,1)
            lj = k_table(l,2)
#ifdef R3D
            lk = k_table(l,3)
#endif /* R3D */

! get the directional trigonometric functions
!
            snx = sinx(li, i)
            csx = cosx(li, i)
            sny = siny(lj, j)
            csy = cosy(lj, j)
#ifdef R3D
            snz = sinz(lk, k)
            csz = cosz(lk, k)
#endif /* R3D */

! calculate the functions for sum of phases
!
#ifdef R3D
            snp = snx * csy + csx * sny
            csp = csx * csy - snx * sny

            sn  = snp * csz + csp * snz
            cs  = csp * csz - snp * snz
#else /* R3D */
            sn  = snx * csy + csx * sny
            cs  = csx * csy - snx * sny
#endif /* R3D */
#endif /* FAST */

! prepare the complex factor
!
            ph = cmplx(cs, - sn)

! update the fourier coefficients for velocity
!
            fvel(l,1) = fvel(l,1) + ph * qp(ivx,i,j,k)
            fvel(l,2) = fvel(l,2) + ph * qp(ivy,i,j,k)
#ifdef R3D
            fvel(l,3) = fvel(l,3) + ph * qp(ivz,i,j,k)
#endif /* R3D */

          end do

        end do
      end do
    end do

! normalize the fourier velocity coefficients
!
    fvel(:,:) = fvel(:,:) * dvol

#ifdef MPI
! update the fourier components from all processors

    call reduce_sum_complex_array(nlf3, fvel(:,:), iret)
#endif /* MPI */

#ifdef PROFILE
! stop accounting time for the real to fourier transform
!
    call stop_timer(ifr)
#endif /* PROFILE */

!-------------------------------------------------------------------------------
!
  end subroutine real_to_fourier
!
!===============================================================================
!
! subroutine FOURIER_TO_REAL:
! --------------------------
!
!   Subroutine transforms forcing components from the fourier to real space and
!   updates variables with the forcing term.
!
!===============================================================================
!
  subroutine fourier_to_real()

! obtain required variables from other modules
!
#ifdef SUBVOL
    use constants, only : pi
#endif /* SUBVOL */
    use mesh     , only : im, jm, km, ib, jb, kb, ie, je, ke
#ifdef SUBVOL
    use mesh     , only : time, dt
#endif /* SUBVOL */
#ifdef INT_FOR
    use mesh     , only : dvolh
#endif /* INT_FOR */
    use variables, only : u0
    use variables, only : idn, imx, imy, imz, ivx, ivy, ivz
#ifdef ADI
    use variables, only : ien
#endif /* ADI */
#ifdef INT_FOR
    use variables, only : du
#endif /* INT_FOR */

! declare all variables as implicit
!
    implicit none

! local variables
!
    integer :: l, i, j, k
    real    :: sn, cs
    real    :: fx, fy, fz, ek
#ifdef FAST
    integer :: q
#else /* FAST */
    integer :: li, lj, lk
    real    :: snx, sny, snz, snp, csx, csy, csz, csp
#endif /* !FAST */
    integer :: ith = 0
#ifdef SUBVOL
    real       :: fc
    real, save :: ft
#endif /* SUBVOL */

#ifdef SUBVOL
! saved variables
!
    logical, save :: first = .true.
    real   , save :: ff    = 1.0d+00
#endif /* SUBVOL */

! local arrays
!
    real        , dimension(  nlf,3) :: fre, fim
#ifdef INT_FOR
    real(kind=8)                     :: ef, cf, vf, sf
    real(kind=8), dimension(3)       :: af
#endif /* INT_FOR */
!
!-------------------------------------------------------------------------------
!
#ifdef PROFILE
! start accounting time for the fourier to real transform
!
    call start_timer(iff)
#endif /* PROFILE */

#ifdef SUBVOL
! prepare save coefficients at the first call
!
    if (first) then

! calculate factor corresponding to the time of injection
!
      ff    = 0.5d+00 * pi / max(dt, tefor - tbfor)

! reset the first call run
!
      first = .false.

    end if

! calculate the time factor
!
    ft = 1.0d+00
    if (time < tefor) then
      fc = sin(ff * max(0.0d+00, time - tbfor))
      ft = fc * fc
    end if
#endif /* SUBVOL */

! separate real and imaginary parts of the forcing Fourier coefficients into
! different arrays
!
    fre(:,1:3) = real (ffor(:,1:3))
    fim(:,1:3) = aimag(ffor(:,1:3))

! iterate over all domain positions
!
    do k = 1, km
      do j = 1, jm
        do i = 1, im

! reset the local amplitudes
!
          fx = 0.0d+00
          fy = 0.0d+00
          fz = 0.0d+00

! iterate over all forcing components
!
          do l = 1, nlf

#ifdef FAST
! calculate the fast fourier array index
!
#ifdef R3D
            q  = kix(l, i) + kiy(l, j) + kiz(l, k)
#else /* R3D */
            q  = kix(l, i) + kiy(l, j)
#endif /* R3D */

! get the fourier transform coefficients
!
            sn = sinq(q)
            cs = cosq(q)
#else /* FAST */
! get the directional wave numbers
!
            li = k_table(l,1)
            lj = k_table(l,2)
#ifdef R3D
            lk = k_table(l,3)
#endif /* R3D */

! get the directional trigonometric functions
!
            snx = sinx(li, i)
            csx = cosx(li, i)
            sny = siny(lj, j)
            csy = cosy(lj, j)
#ifdef R3D
            snz = sinz(lk, k)
            csz = cosz(lk, k)
#endif /* R3D */

! calculate the functions for sum of phases
!
#ifdef R3D
            snp = snx * csy + csx * sny
            csp = csx * csy - snx * sny

            sn  = snp * csz + csp * snz
            cs  = csp * csz - snp * snz
#else /* R3D */
            sn  = snx * csy + csx * sny
            cs  = csx * csy - snx * sny
#endif /* R3D */
#endif /* FAST */

! update the local amplitude by the current forcing component contribution
!
            fx  = fx + fre(l,1) * cs + fim(l,1) * sn
            fy  = fy + fre(l,2) * cs + fim(l,2) * sn
#ifdef R3D
            fz  = fz + fre(l,3) * cs + fim(l,3) * sn
#endif /* R3D */

          end do

#ifdef SUBVOL
! calculate the factor for the injection region
!
          fc = ft * fv(j) * fr(i,k)

! limit the forcing amplitudes by the injection region factor
!
          fx = fc * fx
          fy = fc * fy
#ifndef R2D
          fz = fc * fz
#endif /* R2D */
#endif /* SUBVOL */

! update the conservative variables
!
#if defined INT_FOR || defined ADI
          ek   = sum(u0(imx:imz,i,j,k)**2) / u0(idn,i,j,k)
#endif /* INT_FOR | ADI */

          u0(imx,i,j,k) = u0(imx,i,j,k) + u0(idn,i,j,k) * fx
          u0(imy,i,j,k) = u0(imy,i,j,k) + u0(idn,i,j,k) * fy
#ifdef R3D
          u0(imz,i,j,k) = u0(imz,i,j,k) + u0(idn,i,j,k) * fz
#endif /* R3D */

#if defined INT_FOR || defined ADI
! calculate forcing correlations
!
          ek   = sum(u0(imx:imz,i,j,k)**2)/u0(idn,i,j,k) - ek
#endif /* INT_FOR | ADI */
#ifdef ADI
          u0(ien,i,j,k) = u0(ien,i,j,k) + 0.5d+00 * ek
#endif /* ADI */
#ifdef INT_FOR
          du(1,i,j,k) = ek
          du(2,i,j,k) = fx * fx
          du(3,i,j,k) = fy * fy
#ifdef R3D
          du(4,i,j,k) = fz * fz
#endif /* R3D */
#endif /* INT_FOR */

        end do
      end do
    end do

#ifdef INT_FOR
! reset coefficients
!
    ef    = 0.0d+00
    cf    = 0.0d+00
    vf    = 0.0d+00
    sf    = 0.0d+00
    af(:) = 0.0d+00

! update correlations
!
    do k = kb, ke
      do j = jb, je
        do i = ib, ie
          ek = u0(idn,i,j,k) * (du(2,i,j,k) + du(3,i,j,k) + du(4,i,j,k))

          sf    = sf    +  du(1,i,j,k) / u0(1,i,j,k)
          ef    = ef    +  du(1,i,j,k)
          cf    = cf    +  ek
          vf    = vf    + (du(1,i,j,k) - ek)
          af(1) = af(1) + du(2,i,j,k) * u0(idn,i,j,k)
          af(2) = af(2) + du(3,i,j,k) * u0(idn,i,j,k)
          af(3) = af(3) + du(4,i,j,k) * u0(idn,i,j,k)
        end do
      end do
    end do

    efor = efor + dvolh * ef
    cfor = cfor + dvolh * cf
    vfor = vfor + dvolh * vf
    sfor = sfor + dvolh * sf
    afrx = afrx + dvolh * af(1)
    afry = afry + dvolh * af(2)
    afrz = afrz + dvolh * af(3)
#endif /* INT_FOR */

#ifdef PROFILE
! stop accounting time for the fourier to real transform
!
    call stop_timer(iff)
#endif /* PROFILE */

!-------------------------------------------------------------------------------
!
  end subroutine fourier_to_real
#endif /* RANDOM */
#endif /* FORCING */

!===============================================================================
!
end module forcing
