!!******************************************************************************
!!
!!  This file is part of the GODUNOV source code, a program to perform
!!  Newtonian or relativistic magnetohydrodynamical simulations.
!!
!!  Copyright (C) 2007-2013 Grzegorz Kowal <grzegorz@amuncode.org>
!!
!!  This program is free software: you can redistribute it and/or modify
!!  it under the terms of the GNU General Public License as published by
!!  the Free Software Foundation, either version 3 of the License, or
!!  (at your option) any later version.
!!
!!  This program is distributed in the hope that it will be useful,
!!  but WITHOUT ANY WARRANTY; without even the implied warranty of
!!  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!!  GNU General Public License for more details.
!!
!!  You should have received a copy of the GNU General Public License
!!  along with this program.  If not, see <http://www.gnu.org/licenses/>.
!!
!!******************************************************************************
!!
!! module: MESH
!!
!!  This module provides subroutines and variables related to mesh generation
!!  and coordinate system.
!!
!!******************************************************************************
!
module mesh

#ifdef PROFILE
! include external procedures
!
  use timers, only : set_timer, start_timer, stop_timer
#endif /* PROFILE */

! module variables are not implicit by default
!
  implicit none

#ifdef PROFILE
! timer indices
!
  integer, save :: imi
#endif /* PROFILE */

!=== MODULE VARIABLES ===
!
! time variables:
!
!   step - the integration step number
!   time - the evolution time in the code units
!   dt   - the current time step
!   dti  - the initial time step
!   dtt  - the fractional time step (used in the Runge-Kutta substeps)
!   dtn  - the new time step estimated from the CFL condition
!   dth  - the new time step estimated from the hydrodynamic terms
!   dtd  - the new time step estimated from the diffusive terms
!
  integer, save :: step = 0
  real   , save :: time = 0.0d+00
  real   , save :: dt   = 1.0d+00
  real   , save :: dti  = 1.0d-08
  real   , save :: dtt  = 1.0d+00
  real   , save :: dtn  = 1.0d+00
  real   , save :: dth  = 1.0d+00
  real   , save :: dtd  = 1.0d+00

! domain dimensions:
!
!   nn - the uniform resolution, used to initial resolution in other directions
!   in - the resolution in the X direction
!   jn - the resolution in the Y direction
!   kn - the resolution in the Z direction
!   ng - the number of ghost cells
!
  integer, save :: nn   = 1
  integer, save :: in   = 1
  integer, save :: jn   = 1
  integer, save :: kn   = 1
  integer, save :: ng   = 1

! domain bounds:
!
!   xmin, ymin, zmin - the lower coordinate bounds in the X, Y, and Z directions
!   xmax, ymax, zmax - the upper coordinate bounds in the X, Y, and Z directions
!   xlen, ylen, zlen - the domain sizes in the X, Y, and Z directions
!
  real   , save :: xmin  = 0.0d+00
  real   , save :: ymin  = 0.0d+00
  real   , save :: zmin  = 0.0d+00
  real   , save :: xmax  = 1.0d+00
  real   , save :: ymax  = 1.0d+00
  real   , save :: zmax  = 1.0d+00
  real   , save :: xlen  = 1.0d+00
  real   , save :: ylen  = 1.0d+00
  real   , save :: zlen  = 1.0d+00

! mesh dimensions:
!
!   im, jm, km - the subdomain dimensions including the ghost zones
!
  integer, save :: im    = 1, jm    = 1, km    = 1

! domain indices:
!
!   ib , jb , kb  - the first indices of domain (after the ghost zones)
!   ie , je , ke  - the last indices of domain (before the ghost zones)
!   ibl, jbl, kbl - the indices used mostly in the boundary update
!   iel, jel, kel
!   ibu, jbu, kbu
!   ieu, jeu, keu
!
  integer, save :: ib    = 1, jb    = 1, kb    = 1
  integer, save :: ie    = 1, je    = 1, ke    = 1
  integer, save :: ibl   = 1, jbl   = 1, kbl   = 1
  integer, save :: iel   = 1, jel   = 1, kel   = 1
  integer, save :: ibu   = 1, jbu   = 1, kbu   = 1
  integer, save :: ieu   = 1, jeu   = 1, keu   = 1

! coordinates increments and area and volume elements
!
  real   , save :: dx    = 1.0d+00, dy    = 1.0d+00, dz    = 1.0d+00
  real   , save :: dxh   = 1.0d+00, dyh   = 1.0d+00, dzh   = 1.0d+00
  real   , save :: dxi   = 1.0d+00, dyi   = 1.0d+00, dzi   = 1.0d+00
  real   , save :: dxih  = 1.0d+00, dyih  = 1.0d+00, dzih  = 1.0d+00
  real   , save :: vol   = 1.0d+00, dvol  = 1.0d+00, dvolh = 1.0d+00
  real   , save :: daxy  = 1.0d+00, daxz  = 1.0d+00, dayz  = 1.0d+00
  real   , save :: dxmin = 1.0d+00

! cell and face centered coordinate vectors
!
  real   , save, dimension(:), allocatable :: xc, yc, zc
  real   , save, dimension(:), allocatable :: xi, yi, zi

! by default everything is public
!
  public

!- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
!
  contains
!
!===============================================================================
!
! subroutine INITIALIZE_MESH:
! --------------------------
!
!   Subroutine initializes mesh coordinates and other mesh parameters.
!
!   Arguments:
!
!    pdims(:)   - the domain division for multiple processors;
!    pcoords(:) - the current processor coordinate;
!
!===============================================================================
!
  subroutine initialize_mesh(pdims, pcoords)

! include external procedures and variables
!
    use constants , only : pi
    use parameters, only : get_parameter_integer, get_parameter_real           &
                         , get_parameter_string

! local variables are not implicit by default
!
    implicit none

! subroutine arguments
!
    integer, dimension(3), intent(in) :: pdims, pcoords

! local variables
!
    integer :: i, j, k
!
!-------------------------------------------------------------------------------
!
#ifdef PROFILE
! set timer descriptions
!
    call set_timer('mesh initialization', imi)

! start accounting time for the coordinate initialization
!
    call start_timer(imi)
#endif /* PROFILE */

! first obtain all parameters from the parameter module
!
    call get_parameter_real   ("dti" , dti  )
    call get_parameter_integer("nn"  , nn   )
    in = nn
    jn = nn
#ifdef R3D
    kn = nn
#endif /* R3D */
    call get_parameter_integer("in"  , in   )
    call get_parameter_integer("jn"  , jn   )
    call get_parameter_integer("kn"  , kn   )
    call get_parameter_integer("ng"  , ng   )
    call get_parameter_real   ("xmin", xmin )
    call get_parameter_real   ("xmax", xmax )
    call get_parameter_real   ("ymin", ymin )
    call get_parameter_real   ("ymax", ymax )
    call get_parameter_real   ("zmin", zmin )
    call get_parameter_real   ("zmax", zmax )

! divide the resolution among the pieces according to the MPI domain division
!
    if (mod(in, pdims(1)) == 0) then
      in = in / pdims(1)
    else
      write(*,*) 'ERROR: Division impossible since mod(in,ip) is not zero!'
    end if
    if (mod(jn, pdims(2)) == 0) then
      jn = jn / pdims(2)
    else
      write(*,*) 'ERROR: Division impossible since mod(jn,jp) is not zero!'
    end if
#ifdef R3D
    if (mod(kn, pdims(3)) == 0) then
      kn = kn / pdims(3)
    else
      write(*,*) 'ERROR: Division impossible since mod(kn,kp) is not zero!'
    end if
#endif /* R3D */

! calculate dimensions and indices along the X direction
!
    if (in > 1) then

      im  = in + 2 * ng

      ib  =  1 + ng
      ie  = in + ng

      ibl = ib - 1
      ibu = ib + ng - 1
      iel = ie - ng + 1
      ieu = ie + 1

    end if

! calculate dimensions and indices along the Y direction
!
    if (jn > 1) then

      jm  = jn + 2 * ng

      jb  =  1 + ng
      je  = jn + ng

      jbl = jb - 1
      jbu = jb + ng - 1
      jel = je - ng + 1
      jeu = je + 1

    end if

! calculate dimensions and indices along the Z direction
!
    if (kn > 1) then

      km  = kn + 2 * ng

      kb  =  1 + ng
      ke  = kn + ng

      kbl = kb - 1
      kbu = kb + ng - 1
      kel = ke - ng + 1
      keu = ke + 1

    end if

! calculate the domain physical lengths
!
    xlen  = xmax - xmin
    ylen  = ymax - ymin
#ifdef R3D
    zlen  = zmax - zmin
#endif /* R3D */

! calculate the domain volume
!
    vol   = xlen * ylen * zlen

! prepare the coordinate increments
!
    dx    = xlen / (in * pdims(1))
    dy    = ylen / (jn * pdims(2))
#ifdef R3D
    dz    = zlen / (kn * pdims(3))
#endif /* R3D */

! prepare the half coordinate intervals
!
    dxh   = 0.5d+00 * dx
    dyh   = 0.5d+00 * dy
#ifdef R3D
    dzh   = 0.5d+00 * dz
#endif /* R3D */

! calculate the inversed elements
!
    dxi   =  (in * pdims(1)) / xlen
    dyi   =  (jn * pdims(2)) / ylen
#ifdef R3D
    dzi   =  (kn * pdims(3)) / zlen
#endif /* R3D */

    dxih  = 0.5d+00 * dxi
    dyih  = 0.5d+00 * dyi
#ifdef R3D
    dzih  = 0.5d+00 * dzi
#endif /* R3D */

! calculate the area and volume elements
!
    daxy  = dx * dy
    daxz  = dx * dz
    dayz  = dy * dz
    dvol  = dx * dy * dz
    dvolh = 0.5d+00 * dvol

! find the minimum coordinate interval
!
#ifdef R3D
    dxmin = min(dx, dy, dz)
#else /* R3D */
    dxmin = min(dx, dy)
#endif /* R3D */

! allocate the variables for coordinates
!
    allocate(xc(im))
    allocate(yc(jm))
    allocate(zc(km))
    allocate(xi(im))
    allocate(yi(jm))
    allocate(zi(km))

! generate the coordinates
!
    if (in > 1) then
      xc(:) = ((/(i, i = 1, im)/) - ng - 0.5d+00 + in * pcoords(1)) * dx + xmin
      xi(:) = xc(:) + dxh
    else
      xc(:) = 0.0d+00
      xi(:) = 0.0d+00
    end if

    if (jn > 1) then
      yc(:) = ((/(j, j = 1, jm)/) - ng - 0.5d+00 + jn * pcoords(2)) * dy + ymin
      yi(:) = yc(:) + dyh
    else
      yc(:) = 0.0d+00
      yi(:) = 0.0d+00
    end if

    if (kn > 1) then
      zc(:) = ((/(k, k = 1, km)/) - ng - 0.5d+00 + kn * pcoords(3)) * dz + zmin
      zi(:) = zc(:) + dzh
    else
      zc(:) = 0.0d+00
      zi(:) = 0.0d+00
    end if

! initialize time variables
!
    step = 0
    time = 0.0d+00
    dt   = 0.5d+00 * dti
    dtt  = 0.5d+00 * dti
    dtn  = 0.5d+00 * dti
    dth  = 0.5d+00 * dti
    dtd  = 1.0d+16 * dxmin**2

#ifdef PROFILE
! stop accounting time for the coordinate allocation
!
    call stop_timer(imi)
#endif /* PROFILE */

!-------------------------------------------------------------------------------
!
  end subroutine initialize_mesh
!
!===============================================================================
!
! subroutine FINALIZE_MESH:
! ------------------------
!
!   subroutine finalizes allocated mesh coordinates by deallocating them;
!
!===============================================================================
!
  subroutine finalize_mesh()

! local variables are not implicit by default
!
    implicit none
!
!-------------------------------------------------------------------------------
!
#ifdef PROFILE
! start accounting time for the coordinate deallocation
!
    call start_timer(imi)
#endif /* PROFILE */

! deallocate coordinates
!
    if (allocated(xc)) deallocate(xc)
    if (allocated(yc)) deallocate(yc)
    if (allocated(zc)) deallocate(zc)
    if (allocated(xi)) deallocate(xi)
    if (allocated(yi)) deallocate(yi)
    if (allocated(zi)) deallocate(zi)

#ifdef PROFILE
! stop accounting time for the coordinate deallocation
!
    call stop_timer(imi)
#endif /* PROFILE */

!-------------------------------------------------------------------------------
!
  end subroutine finalize_mesh

!===============================================================================
!
end module mesh
