!!******************************************************************************
!!
!!  This file is part of the GODUNOV source code, a program to perform
!!  Newtonian or relativistic magnetohydrodynamical simulations.
!!
!!  Copyright (C) 2007-2013 Grzegorz Kowal <grzegorz@amuncode.org>
!!
!!  This program is free software: you can redistribute it and/or modify
!!  it under the terms of the GNU General Public License as published by
!!  the Free Software Foundation, either version 3 of the License, or
!!  (at your option) any later version.
!!
!!  This program is distributed in the hope that it will be useful,
!!  but WITHOUT ANY WARRANTY; without even the implied warranty of
!!  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!!  GNU General Public License for more details.
!!
!!  You should have received a copy of the GNU General Public License
!!  along with this program.  If not, see <http://www.gnu.org/licenses/>.
!!
!!******************************************************************************
!!
!! module: SCHEMES
!!
!!  The module handles numerical schemes, subroutines to calculate variable
!!  increment and one dimensional Riemann solvers.
!!
!!******************************************************************************
!
module schemes

! module variables are not implicit by default
!
  implicit none

! by default everything is private
!
  private

! declare public subroutines
!
  public :: update

!- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
!
  contains
#ifdef HYDRO
!
!===============================================================================
!
! subroutine UPDATE:
! =================
!
!   Subroutine solves the Riemann problem along each direction and calculates
!   the numerical fluxes, which are used to calculate the conserved variable
!   increment.
!
!===============================================================================
!
  subroutine update()

! include external procedures and variables
!
#ifdef SHOCK_DETECTION
    use equations, only : detect_shock
#endif /* SHOCK_DETECTION */
    use mesh     , only : im, jm, km
    use mesh     , only : dx, dy, dz, dxi, dyi, dzi
    use variables, only : du, qp, qx, qy, qz, fx, fy, fz, sx, sy, sz
    use variables, only : idn, imx, imy, imz, ivx, ivy, ivz
#ifdef ADI
    use variables, only : ien, ipr
#endif /* ADI */
#ifdef SHOCK_DETECTION
    use variables, only : sh
#endif /* SHOCK_DETECTION */

! local variables are not implicit by default
!
    implicit none

! local variables
!
    integer       :: i, j, k

! variables kept between calls
!
    logical, save :: first = .true.
    integer, save :: imm1 = 1, jmm1 = 1, kmm1 = 1
!
!-------------------------------------------------------------------------------
!
! prepare some variables in the first call
!
    if (first) then

! prepare the dimension bounds
!
      imm1 = im - 1
      jmm1 = jm - 1
#ifdef R3D
      kmm1 = km - 1
#endif /* R3D */

! reset the first call flag
!
      first = .false.

    end if

! reset the array of conservative variable increments
!
    du(:,:,:,:) = 0.0d+00

#ifdef SHOCK_DETECTION
! find shock locations
!
    call detect_shock()
#endif /* SHOCK_DETECTION */

! estimate the increment contribution from the X direction fluxes
!
    do k = 1, km
      do j = 1, jm

! copy directional variable vectors to pass to the one dimensional solver
!
        qx(idn,:) = qp(idn,:,j,k)
        qx(ivx,:) = qp(ivx,:,j,k)
        qx(ivy,:) = qp(ivy,:,j,k)
        qx(ivz,:) = qp(ivz,:,j,k)
#ifdef ADI
        qx(ipr,:) = qp(ipr,:,j,k)
#endif /* ADI */
#ifdef SHOCK_DETECTION
        sx(    :) = sh(    :,j,k)
#endif /* SHOCK_DETECTION */

! call one dimensional Riemann solver in order to obtain numerical fluxes
!
        call riemann(im, dx, qx(:,:), fx(:,:), sx(:))

! update the array of increments using numerical fluxes
!
        du(idn,2:im,j,k) = du(idn,2:im,j,k)                                    &
                                       + dxi * (fx(idn,1:imm1) - fx(idn,2:im))
        du(idn,1   ,j,k) = du(idn,1   ,j,k) - dxi * fx(idn,1   )
        du(imx,2:im,j,k) = du(imx,2:im,j,k)                                    &
                                       + dxi * (fx(imx,1:imm1) - fx(imx,2:im))
        du(imx,1   ,j,k) = du(imx,1   ,j,k) - dxi * fx(imx,1   )
        du(imy,2:im,j,k) = du(imy,2:im,j,k)                                    &
                                       + dxi * (fx(imy,1:imm1) - fx(imy,2:im))
        du(imy,1   ,j,k) = du(imy,1   ,j,k) - dxi * fx(imy,1   )
        du(imz,2:im,j,k) = du(imz,2:im,j,k)                                    &
                                       + dxi * (fx(imz,1:imm1) - fx(imz,2:im))
        du(imz,1   ,j,k) = du(imz,1   ,j,k) - dxi * fx(imz,1   )
#ifdef ADI
        du(ien,2:im,j,k) = du(ien,2:im,j,k)                                    &
                                       + dxi * (fx(ien,1:imm1) - fx(ien,2:im))
        du(ien,1   ,j,k) = du(ien,1   ,j,k) - dxi * fx(ien,1   )
#endif /* ADI */

      end do
    end do

! estimate the increment contribution from the Y direction fluxes
!
    do k = 1, km
      do i = 1, im

! copy directional variable vectors to pass to the one dimensional solver
!
        qy(idn,:) = qp(idn,i,:,k)
        qy(ivx,:) = qp(ivy,i,:,k)
        qy(ivy,:) = qp(ivz,i,:,k)
        qy(ivz,:) = qp(ivx,i,:,k)
#ifdef ADI
        qy(ipr,:) = qp(ipr,i,:,k)
#endif /* ADI */
#ifdef SHOCK_DETECTION
        sy(    :) = sh(    i,:,k)
#endif /* SHOCK_DETECTION */

! call one dimensional Riemann solver in order to obtain numerical fluxes
!
        call riemann(jm, dy, qy(:,:), fy(:,:), sy(:))

! update the array of increments using numerical fluxes
!
        du(idn,i,2:jm,k) = du(idn,i,2:jm,k)                                    &
                                       + dyi * (fy(idn,1:jmm1) - fy(idn,2:jm))
        du(idn,i,1   ,k) = du(idn,i,1   ,k) - dyi * fy(idn,1   )
        du(imx,i,2:jm,k) = du(imx,i,2:jm,k)                                    &
                                       + dyi * (fy(imz,1:jmm1) - fy(imz,2:jm))
        du(imx,i,1   ,k) = du(imx,i,1   ,k) - dyi * fy(imz,1   )
        du(imy,i,2:jm,k) = du(imy,i,2:jm,k)                                    &
                                       + dyi * (fy(imx,1:jmm1) - fy(imx,2:jm))
        du(imy,i,1   ,k) = du(imy,i,1   ,k) - dyi * fy(imx,1    )
        du(imz,i,2:jm,k) = du(imz,i,2:jm,k)                                    &
                                       + dyi * (fy(imy,1:jmm1) - fy(imy,2:jm))
        du(imz,i,1   ,k) = du(imz,i,1   ,k) - dyi * fy(imy,1    )
#ifdef ADI
        du(ien,i,2:jm,k) = du(ien,i,2:jm,k)                                    &
                                       + dyi * (fy(ien,1:jmm1) - fy(ien,2:jm))
        du(ien,i,1   ,k) = du(ien,i,1   ,k) - dyi * fy(ien,1    )
#endif /* ADI */

      end do
    end do

#ifdef R3D
! estimate the increment contribution from the Z direction fluxes
!
    do j = 1, jm
      do i = 1, im

! copy directional variable vectors to pass to the one dimensional solver
!
        qz(idn,:) = qp(idn,i,j,:)
        qz(ivx,:) = qp(ivz,i,j,:)
        qz(ivy,:) = qp(ivx,i,j,:)
        qz(ivz,:) = qp(ivy,i,j,:)
#ifdef ADI
        qz(ipr,:) = qp(ipr,i,j,:)
#endif /* ADI */
#ifdef SHOCK_DETECTION
        sz(    :) = sh(    i,j,:)
#endif /* SHOCK_DETECTION */

! call one dimensional Riemann solver in order to obtain numerical fluxes
!
        call riemann(km, dz, qz(:,:), fz(:,:), sz(:))

! update the array of increments using numerical fluxes
!
        du(idn,i,j,2:km) = du(idn,i,j,2:km)                                    &
                                       + dzi * (fz(idn,1:kmm1) - fz(idn,2:km))
        du(idn,i,j,1   ) = du(idn,i,j,1   ) - dzi * fz(idn,1   )
        du(imx,i,j,2:km) = du(imx,i,j,2:km)                                    &
                                       + dzi * (fz(imy,1:kmm1) - fz(imy,2:km))
        du(imx,i,j,1   ) = du(imx,i,j,1   ) - dzi * fz(imy,1   )
        du(imy,i,j,2:km) = du(imy,i,j,2:km)                                    &
                                       + dzi * (fz(imz,1:kmm1) - fz(imz,2:km))
        du(imy,i,j,1   ) = du(imy,i,j,1   ) - dzi * fz(imz,1   )
        du(imz,i,j,2:km) = du(imz,i,j,2:km)                                    &
                                       + dzi * (fz(imx,1:kmm1) - fz(imx,2:km))
        du(imz,i,j,1   ) = du(imz,i,j,1   ) - dzi * fz(imx,1   )
#ifdef ADI
        du(ien,i,j,2:km) = du(ien,i,j,2:km)                                    &
                                       + dzi * (fz(ien,1:kmm1) - fz(ien,2:km))
        du(ien,i,j,1   ) = du(ien,i,j,1   ) - dzi * fz(ien,1   )
#endif /* ADI */

      end do
    end do
#endif /* R3D */

!-------------------------------------------------------------------------------
!
  end subroutine update
#ifdef HLL
!
!===============================================================================
!
! subroutine RIEMANN:
! ==================
!
!   Subroutine solves one dimensional Riemann problem using the HLL method.
!
!   Arguments:
!
!     n - the length of input vectors;
!     h - the spatial step;
!     q - the input array of primitive variables;
!     f - the output array of fluxes;
!     s - the input array of shock indicators;
!
!   References:
!
!     [1] Harten, A., Lax, P. D. & Van Leer, B.
!         "On Upstream Differencing and Godunov-Type Schemes for Hyperbolic
!          Conservation Laws",
!         SIAM Review, 1983, Volume 25, Number 1, pp. 35-61
!
!===============================================================================
!
  subroutine riemann(n, h, q, f, s)

! include external procedures and variables
!
    use equations     , only : prim2cons, fluxspeed
    use interpolations, only : reconstruct, fix_positivity
    use variables     , only : nt
    use variables     , only : idn, ivx
#ifdef ADI
    use variables     , only : ipr
#endif /* ADI */

! local variables are not implicit by default
!
    implicit none

! input / output arguments
!
    integer                 , intent(in)  :: n
    real                    , intent(in)  :: h
    real   , dimension(nt,n), intent(in)  :: q
    real   , dimension(nt,n), intent(out) :: f
    logical, dimension(n)   , intent(in)  :: s

! local variables
!
    integer :: p, i
    real    :: sl, sr, sp, srml

! local arrays to store the states
!
    real, dimension(nt,n) :: ul, ur, ql, qr, fl, fr
    real, dimension(n)    :: cl, cr
!
!-------------------------------------------------------------------------------
!
! reconstruct the left and right states of primitive variables
!
    do p = 1, nt
      call reconstruct(n, h, q(p,:), ql(p,:), qr(p,:), s(:))
    end do

#ifdef FIX_POSITIVITY
! check if the reconstruction doesn't give the negative density or pressure,
! if so, correct the states
!
    call fix_positivity(n, q(idn,:), ql(idn,:), qr(idn,:))
#ifdef ADI
    call fix_positivity(n, q(ipr,:), ql(ipr,:), qr(ipr,:))
#endif /* ADI */
#endif /* FIX_POSITIVITY */

! calculate corresponding conserved variables of the left and right states
!
    call prim2cons(n, ql(:,:), ul(:,:))
    call prim2cons(n, qr(:,:), ur(:,:))

! calculate the physical fluxes and speeds at the states
!
    call fluxspeed(n, ql(:,:), ul(:,:), fl(:,:), cl(:))
    call fluxspeed(n, qr(:,:), ur(:,:), fr(:,:), cr(:))

! iterate over all points
!
    do i = 1, n

! estimate the minimum and maximum speeds
!
      sl = min(ql(ivx,i) - cl(i), qr(ivx,i) - cr(i))
      sr = max(ql(ivx,i) + cl(i), qr(ivx,i) + cr(i))

! calculate the HLL flux
!
      if (sl >= 0.0d0) then

        f(:,i) = fl(:,i)

      else if (sr <= 0.0d0) then

        f(:,i) = fr(:,i)

      else

! prepare coefficients for the intermediate state calculation
!
        sp   = sr * sl
        srml = sr - sl

! calculate the fluxes for the intermediate state
!
        f(:,i) = ((sr * fl(:,i) - sl * fr(:,i))                                &
                                            + sp * (ur(:,i) - ul(:,i))) / srml

      end if

    end do

!-------------------------------------------------------------------------------
!
  end subroutine riemann
#endif /* HLL */
#if defined HLLC || defined HLLCC
!
!===============================================================================
!
! subroutine RIEMANN:
! ==================
!
!   Subroutine solves one dimensional Riemann problem using the HLLC method,
!   by Toro.  In the HLLC method the tangential components of the velocity are
!   discontinuous, which in the HLLCC method they are continuous and calculated
!   from the HLL average.
!
!   Arguments:
!
!     n - the length of input vectors;
!     h - the spatial step;
!     q - the input array of primitive variables;
!     f - the output array of fluxes;
!     s - the input array of shock indicators;
!
!   References:
!
!     [1] Toro, E. F., Spruce, M., & Speares, W.
!         "Restoration of the contact surface in the HLL-Riemann solver",
!         Shock Waves, 1994, Volume 4, Issue 1, pp. 25-34
!
!===============================================================================
!
  subroutine riemann(n, h, q, f, s)

! include external procedures and variables
!
    use equations     , only : prim2cons, fluxspeed
    use interpolations, only : reconstruct, fix_positivity
    use variables     , only : nt
    use variables     , only : idn, imx, imy, imz, ivx, ivy, ivz, ien, ipr

! local variables are not implicit by default
!
    implicit none

! input / output arguments
!
    integer              , intent(in)  :: n
    real                 , intent(in)  :: h
    real, dimension(nt,n), intent(in)  :: q
    real, dimension(nt,n), intent(out) :: f
    logical, dimension(n), intent(in)  :: s

! local variables
!
    integer :: p, i
#ifdef SHOCK_DETECTION
    integer :: ip1
#endif /* SHOCK_DETECTION */
    real    :: dn, mx, pr
#ifdef HLLCC
    real    :: my, mz, vy, vz
#endif /* HLLCC */
    real    :: sl, sr, sm
    real    :: srml, slmv, srmv, slmm, srmm, smvl, smvr

! local arrays to store the states
!
    real, dimension(nt,n) :: ul, ur, ql, qr, fl, fr
    real, dimension(n)    :: cl, cr
    real, dimension(nt)   :: q1l, q1r, u1l, u1r
!
!-------------------------------------------------------------------------------
!
! reconstruct the left and right states of primitive variables
!
    do p = 1, nt
      call reconstruct(n, h, q(p,:), ql(p,:), qr(p,:), s(:))
    end do

#ifdef FIX_POSITIVITY
! check if the reconstruction doesn't give the negative density or pressure,
! if so, correct the states
!
    call fix_positivity(n, q(idn,:), ql(idn,:), qr(idn,:))
    call fix_positivity(n, q(ipr,:), ql(ipr,:), qr(ipr,:))
#endif /* FIX_POSITIVITY */

! calculate corresponding conserved variables of the left and right states
!
    call prim2cons(n, ql(:,:), ul(:,:))
    call prim2cons(n, qr(:,:), ur(:,:))

! calculate the physical fluxes and speeds at the states
!
    call fluxspeed(n, ql(:,:), ul(:,:), fl(:,:), cl(:))
    call fluxspeed(n, qr(:,:), ur(:,:), fr(:,:), cr(:))

! iterate over all points
!
    do i = 1, n

! estimate the minimum and maximum speeds
!
      sl = min(ql(ivx,i) - cl(i), qr(ivx,i) - cr(i))
      sr = max(ql(ivx,i) + cl(i), qr(ivx,i) + cr(i))

! calculate the HLLC flux
!
      if (sl >= 0.0d0) then

        f(:,i) = fl(:,i)

      else if (sr <= 0.0d0) then

        f(:,i) = fr(:,i)

      else ! sl < 0 < sr

#ifdef SHOCK_DETECTION
! calculate the next index
!
        ip1 = min(n, i + 1)

! check if we are in the vicinity of a shock
!
        if (s(i) .or. s(ip1)) then

! prepare coefficients for the intermediate state calculation
!
          sm   = sr * sl
          srml = sr - sl

! calculate the fluxes for the intermediate state
!
          f(:,i) = ((sr * fl(:,i) - sl * fr(:,i))                              &
                                            + sm * (ur(:,i) - ul(:,i))) / srml

        else
#endif /* SHOCK_DETECTION */

! the speed of contact discontinuity (eq. 34 [1], 14 [2])
!
          dn = (sr * ur(idn,i) - sl * ul(idn,i)) - (fr(idn,i) - fl(idn,i))
          mx = (sr * ur(imx,i) - sl * ul(imx,i)) - (fr(imx,i) - fl(imx,i))
          sm = mx / dn

#ifdef HLLCC
! the tangential components of velocity are calculated from the averaged HLL
! momenta
!
          my = (sr * ur(imy,i) - sl * ul(imy,i)) - (fr(imy,i) - fl(imy,i))
          mz = (sr * ur(imz,i) - sl * ul(imz,i)) - (fr(imz,i) - fl(imz,i))
          vy = my / dn
          vz = mz / dn
#endif /* HLLCC */

! compute speed differences
!
          slmv = sl - ql(ivx,i)
          srmv = sr - qr(ivx,i)
          slmm = sl - sm
          srmm = sr - sm
          smvl = sm - ql(ivx,i)
          smvr = sm - qr(ivx,i)

! pressure of intermediate states (eq. 36 [1], 16 [2])
!
          pr = 0.5d0 * ((ql(ipr,i) + ql(idn,i) * slmv * smvl)                  &
                      + (qr(ipr,i) + qr(idn,i) * srmv * smvr))

! intermediate discontinuities
!
          if (sm > 0.0d0) then

! calculate the left intermediate state
!
            u1l(idn) = ql(idn,i) * slmv / slmm
            u1l(imx) = u1l(idn) * sm
#ifdef HLLC
            u1l(imy) = u1l(idn) * ql(ivy,i)
            u1l(imz) = u1l(idn) * ql(ivz,i)
#endif /* HLLC */
#ifdef HLLCC
            u1l(imy) = u1l(idn) * vy
            u1l(imz) = u1l(idn) * vz
#endif /* HLLCC */
            u1l(ien) = (slmv * ul(ien,i) - ql(ipr,i) * ql(ivx,i)               &
                                                             + pr * sm) / slmm

! calculate the left intermediate flux
!
            f(:,i) = fl(:,i) + sl * (u1l(:) - ul(:,i))

          else if (sm < 0.0d0) then

! calculate the right intermediate state
!
            u1r(idn) = qr(idn,i) * srmv / srmm
            u1r(imx) = u1r(idn) * sm
#ifdef HLLC
            u1r(imy) = u1r(idn) * qr(ivy,i)
            u1r(imz) = u1r(idn) * qr(ivz,i)
#endif /* HLLC */
#ifdef HLLCC
            u1r(imy) = u1r(idn) * vy
            u1r(imz) = u1r(idn) * vz
#endif /* HLLCC */
            u1r(ien) = (srmv * ur(ien,i) - qr(ipr,i) * qr(ivx,i)               &
                                                             + pr * sm) / srmm

! calculate right intermediate flux
!
            f(:,i) = fr(:,i) + sr * (u1r(:) - ur(:,i))

          else ! sm = 0

! calculate the left intermediate state
!
            u1l(idn) = ql(idn,i) * slmv / sl
            u1l(imx) = 0.0d0
#ifdef HLLC
            u1l(imy) = u1l(idn) * ql(ivy,i)
            u1l(imz) = u1l(idn) * ql(ivz,i)
#endif /* HLLC */
#ifdef HLLCC
            u1l(imy) = u1l(idn) * vy
            u1l(imz) = u1l(idn) * vz
#endif /* HLLCC */
            u1l(ien) = (slmv * ul(ien,i) - ql(ipr,i) * ql(ivx,i)) / sl

! calculate the left intermediate flux
!
            f(:,i) = fl(:,i) + sl * (u1l(:) - ul(:,i))

! calculate the right intermediate state
!
            u1r(idn) = qr(idn,i) * srmv / sr
            u1r(imx) = 0.0d0
#ifdef HLLC
            u1r(imy) = u1r(idn) * qr(ivy,i)
            u1r(imz) = u1r(idn) * qr(ivz,i)
#endif /* HLLC */
#ifdef HLLCC
            u1r(imy) = u1r(idn) * vy
            u1r(imz) = u1r(idn) * vz
#endif /* HLLCC */
            u1r(ien) = (srmv * ur(ien,i) - qr(ipr,i) * qr(ivx,i)) / sr

! calculate right intermediate flux
!
            f(:,i) = 0.5d0 * (f(:,i) + (fr(:,i) + sr * (u1r(:) - ur(:,i))))

          end if ! sm = 0

#ifdef SHOCK_DETECTION
        end if ! shock vicinity
#endif /* SHOCK_DETECTION */

      end if ! sl < 0.0 & sr > 0.0

    end do

!-------------------------------------------------------------------------------
!
  end subroutine riemann
#endif /* HLLC | HLLCC */
#endif /* HYDRO */
#ifdef MHD
!
!===============================================================================
!
! subroutine UPDATE:
! =================
!
!   Subroutine solves the Riemann problem along each direction and calculates
!   the numerical fluxes, which are used to calculate the conserved variable
!   increment.
!
!===============================================================================
!
  subroutine update()

! include external procedures and variables
!
#ifdef SHOCK_DETECTION
    use equations, only : detect_shock
#endif /* SHOCK_DETECTION */
    use mesh     , only : im, jm, km
    use mesh     , only : dx, dy, dz, dxi, dyi, dzi
    use variables, only : du, qp, qx, qy, qz, fx, fy, fz, sx, sy, sz
#ifdef VPOT
    use variables, only : bi
#endif /* VPOT */
    use variables, only : idn, imx, imy, imz, ivx, ivy, ivz
    use variables, only : inx, iny, inz
    use variables, only : ibx, iby, ibz
#ifdef VPOT
    use variables, only : iax, iay, iaz
    use variables, only : iex, iey, iez
#endif /* VPOT */
#ifdef ADI
    use variables, only : ien, ipr
#endif /* ADI */
#ifdef GLM
    use variables, only : iph
#endif /* GLM */
#ifdef SHOCK_DETECTION
    use variables, only : sh
#endif /* SHOCK_DETECTION */

! local variables are not implicit by default
!
    implicit none

! local variables
!
    integer       :: i, j, k

! variables kept between calls
!
    logical, save :: first = .true.
    integer, save :: imm1 = 1, jmm1 = 1, kmm1 = 1

#ifdef CT
! local arrays to store temporarily electric field for averaging
!
    real, dimension(im,jm,km) :: gy, gz
#endif /* CT */
    real, dimension(im)       :: bx
    real, dimension(jm)       :: by
    real, dimension(km)       :: bz
!
!-------------------------------------------------------------------------------
!
! prepare some variables in the first call
!
    if (first) then

! prepare the dimension bounds
!
      imm1 = im - 1
      jmm1 = jm - 1
#ifdef R3D
      kmm1 = km - 1
#endif /* R3D */

! reset the first call flag
!
      first = .false.

    end if

! reset the array of conservative variable increments
!
    du(:,:,:,:) = 0.0d+00

#ifdef SHOCK_DETECTION
! find shock locations
!
    call detect_shock()
#endif /* SHOCK_DETECTION */

! estimate the increment contribution from the X direction fluxes
!
    do k = 1, km
      do j = 1, jm

! copy directional variable vectors to pass to the one dimensional solver
!
        qx(idn,:) = qp(idn,:,j,k)
        qx(ivx,:) = qp(ivx,:,j,k)
        qx(ivy,:) = qp(ivy,:,j,k)
        qx(ivz,:) = qp(ivz,:,j,k)
        qx(ibx,:) = qp(ibx,:,j,k)
        qx(iby,:) = qp(iby,:,j,k)
        qx(ibz,:) = qp(ibz,:,j,k)
#ifdef ADI
        qx(ipr,:) = qp(ipr,:,j,k)
#endif /* ADI */
#ifdef GLM
        qx(iph,:) = qp(iph,:,j,k)
#endif /* GLM */
#ifdef VPOT
        bx(    :) = bi(inx,:,j,k)
#endif /* VPOT */
#ifdef SHOCK_DETECTION
        sx(    :) = sh(    :,j,k)
#endif /* SHOCK_DETECTION */

! call one dimensional Riemann solver in order to obtain numerical fluxes
!
        call riemann(im, dx, qx(:,:), bx(:), fx(:,:), sx(:))

! update the array of increments using numerical fluxes
!
        du(idn,2:im,j,k) = du(idn,2:im,j,k)                                    &
                                       + dxi * (fx(idn,1:imm1) - fx(idn,2:im))
        du(idn,1   ,j,k) = du(idn,1   ,j,k) - dxi * fx(idn,1   )
        du(imx,2:im,j,k) = du(imx,2:im,j,k)                                    &
                                       + dxi * (fx(imx,1:imm1) - fx(imx,2:im))
        du(imx,1   ,j,k) = du(imx,1   ,j,k) - dxi * fx(imx,1   )
        du(imy,2:im,j,k) = du(imy,2:im,j,k)                                    &
                                       + dxi * (fx(imy,1:imm1) - fx(imy,2:im))
        du(imy,1   ,j,k) = du(imy,1   ,j,k) - dxi * fx(imy,1   )
        du(imz,2:im,j,k) = du(imz,2:im,j,k)                                    &
                                       + dxi * (fx(imz,1:imm1) - fx(imz,2:im))
        du(imz,1   ,j,k) = du(imz,1   ,j,k) - dxi * fx(imz,1   )
#ifdef ADI
        du(ien,2:im,j,k) = du(ien,2:im,j,k)                                    &
                                       + dxi * (fx(ien,1:imm1) - fx(ien,2:im))
        du(ien,1   ,j,k) = du(ien,1   ,j,k) - dxi * fx(ien,1   )
#endif /* ADI */
#ifdef GLM
        du(ibx,2:im,j,k) = du(ibx,2:im,j,k)                                    &
                                       + dxi * (fx(ibx,1:imm1) - fx(ibx,2:im))
        du(ibx,1   ,j,k) = du(ibx,1   ,j,k) - dxi * fx(ibx,1   )
        du(iby,2:im,j,k) = du(iby,2:im,j,k)                                    &
                                       + dxi * (fx(iby,1:imm1) - fx(iby,2:im))
        du(iby,1   ,j,k) = du(iby,1   ,j,k) - dxi * fx(iby,1   )
        du(ibz,2:im,j,k) = du(ibz,2:im,j,k)                                    &
                                       + dxi * (fx(ibz,1:imm1) - fx(ibz,2:im))
        du(ibz,1   ,j,k) = du(ibz,1   ,j,k) - dxi * fx(ibz,1   )
        du(iph,2:im,j,k) = du(iph,2:im,j,k)                                    &
                                       + dxi * (fx(iph,1:imm1) - fx(iph,2:im))
        du(iph,1   ,j,k) = du(iph,1   ,j,k) - dxi * fx(iph,1   )
#endif /* GLM */

! update electric field
!
#ifdef CD
#ifndef FIELDCD
        du(iez,2:im,j,k) = du(iez,2:im,j,k) + (fx(iby,1:imm1) + fx(iby,2:im  ))
        du(iez,1   ,j,k) = du(iez,1   ,j,k) +         2.0d+00 * fx(iby,1     )
#ifndef R2D
        du(iey,2:im,j,k) = du(iey,2:im,j,k) - (fx(ibz,1:imm1) + fx(ibz,2:im  ))
        du(iey,1   ,j,k) = du(iey,1   ,j,k) -         2.0d+00 * fx(ibz,1     )
#endif /* !R2D */
#endif /* !FIELDCD */
#endif /* CD */
#ifdef CT
        gy(1:im,j,k) = fx(iby,1:im)
#ifndef R2D
        gz(1:im,j,k) = fx(ibz,1:im)
#endif /* R2D */
#endif /* CT */

      end do
    end do

#ifdef CT
! average electric field
!
    du(iez,:,1:jmm1,:) = du(iez,:,1:jmm1,:) + (gy(:,1:jmm1,:) + gy(:,2:jm  ,:))
    du(iez,:,  jm  ,:) = du(iez,:,  jm  ,:) +         2.0d+00 * gy(:,  jm  ,:)
#ifdef R2D5
    du(iey,:,:,1:km  ) = du(iey,:,:,1:km  ) -  gz(:,:,1:km  )
#endif /* R2D5 */
#ifdef R3D
    du(iey,:,:,1:kmm1) = du(iey,:,:,1:kmm1) - (gz(:,:,1:kmm1) + gz(:,:,2:km  ))
    du(iey,:,:,  km  ) = du(iey,:,:,  km  ) -         2.0d+00 * gz(:,:,  km  )
#endif /* R3D */
#endif /* CT */

! estimate the increment contribution from the Y direction fluxes
!
    do k = 1, km
      do i = 1, im

! copy directional variable vectors to pass to the one dimensional solver
!
        qy(idn,:) = qp(idn,i,:,k)
        qy(ivx,:) = qp(ivy,i,:,k)
        qy(ivy,:) = qp(ivz,i,:,k)
        qy(ivz,:) = qp(ivx,i,:,k)
        qy(ibx,:) = qp(iby,i,:,k)
        qy(iby,:) = qp(ibz,i,:,k)
        qy(ibz,:) = qp(ibx,i,:,k)
#ifdef ADI
        qy(ipr,:) = qp(ipr,i,:,k)
#endif /* ADI */
#ifdef GLM
        qy(iph,:) = qp(iph,i,:,k)
#endif /* GLM */
#ifdef VPOT
        by(    :) = bi(iny,i,:,k)
#endif /* VPOT */
#ifdef SHOCK_DETECTION
        sy(    :) = sh(    i,:,k)
#endif /* SHOCK_DETECTION */

! call one dimensional Riemann solver in order to obtain numerical fluxes
!
        call riemann(jm, dy, qy(:,:), by(:), fy(:,:), sy(:))

! update the array of increments using numerical fluxes
!
        du(idn,i,2:jm,k) = du(idn,i,2:jm,k)                                    &
                                       + dyi * (fy(idn,1:jmm1) - fy(idn,2:jm))
        du(idn,i,1   ,k) = du(idn,i,1   ,k) - dyi * fy(idn,1   )
        du(imx,i,2:jm,k) = du(imx,i,2:jm,k)                                    &
                                       + dyi * (fy(imz,1:jmm1) - fy(imz,2:jm))
        du(imx,i,1   ,k) = du(imx,i,1   ,k) - dyi * fy(imz,1   )
        du(imy,i,2:jm,k) = du(imy,i,2:jm,k)                                    &
                                       + dyi * (fy(imx,1:jmm1) - fy(imx,2:jm))
        du(imy,i,1   ,k) = du(imy,i,1   ,k) - dyi * fy(imx,1    )
        du(imz,i,2:jm,k) = du(imz,i,2:jm,k)                                    &
                                       + dyi * (fy(imy,1:jmm1) - fy(imy,2:jm))
        du(imz,i,1   ,k) = du(imz,i,1   ,k) - dyi * fy(imy,1    )
#ifdef ADI
        du(ien,i,2:jm,k) = du(ien,i,2:jm,k)                                    &
                                       + dyi * (fy(ien,1:jmm1) - fy(ien,2:jm))
        du(ien,i,1   ,k) = du(ien,i,1   ,k) - dyi * fy(ien,1    )
#endif /* ADI */
#ifdef GLM
        du(ibx,i,2:jm,k) = du(ibx,i,2:jm,k)                                    &
                                       + dyi * (fy(ibz,1:jmm1) - fy(ibz,2:jm))
        du(ibx,i,1   ,k) = du(ibx,i,1   ,k) - dyi * fy(ibz,1   )
        du(iby,i,2:jm,k) = du(iby,i,2:jm,k)                                    &
                                       + dyi * (fy(ibx,1:jmm1) - fy(ibx,2:jm))
        du(iby,i,1   ,k) = du(iby,i,1   ,k) - dyi * fy(ibx,1    )
        du(ibz,i,2:jm,k) = du(ibz,i,2:jm,k)                                    &
                                       + dyi * (fy(iby,1:jmm1) - fy(iby,2:jm))
        du(ibz,i,1   ,k) = du(ibz,i,1   ,k) - dyi * fy(iby,1    )
        du(iph,i,2:jm,k) = du(iph,i,2:jm,k)                                    &
                                       + dyi * (fy(iph,1:jmm1) - fy(iph,2:jm))
        du(iph,i,1   ,k) = du(iph,i,1   ,k) - dyi * fy(iph,1    )
#endif /* GLM */

! update electric field
!
#ifdef CD
#ifndef FIELDCD
#ifndef R2D
        du(iex,i,2:jm,k) = du(iex,i,2:jm,k) + (fy(iby,1:jmm1) + fy(iby,2:jm  ))
        du(iex,i,1   ,k) = du(iex,i,1   ,k) +         2.0d+00 * fy(iby,1     )
#endif /* !R2D */
        du(iez,i,2:jm,k) = du(iez,i,2:jm,k) - (fy(ibz,1:jmm1) + fy(ibz,2:jm  ))
        du(iez,i,1   ,k) = du(iez,i,1   ,k) -         2.0d+00 * fy(ibz,1     )
#endif /* !FIELDCD */
#endif /* CD */
#ifdef CT
#ifndef R2D
        gy(i,1:jm,k) = fy(iby,1:jm)
#endif /* R2D */
        gz(i,1:jm,k) = fy(ibz,1:jm)
#endif /* CT */

      end do
    end do

#ifdef CT
! average electric field
!
#ifdef R2D5
    du(iex,:,:,1:km  ) = du(iex,:,:,1:km  ) +  gy(:,:,1:km  )
#endif /* R2D5 */
#ifdef R3D
    du(iex,:,:,1:kmm1) = du(iex,:,:,1:kmm1) + (gy(:,:,1:kmm1) + gy(:,:,2:km  ))
    du(iex,:,:,  km  ) = du(iex,:,:,  km  ) +         2.0d+00 * gy(:,:,  km  )
#endif /* R3D */
    du(iez,1:imm1,:,:) = du(iez,1:imm1,:,:) - (gz(1:imm1,:,:) + gz(2:im  ,:,:))
    du(iez,  im  ,:,:) = du(iez,  im  ,:,:) -         2.0d+00 * gz(  im  ,:,:)
#endif /* CT */

#ifdef R3D
! estimate the increment contribution from the Z direction fluxes
!
    do j = 1, jm
      do i = 1, im

! copy directional variable vectors to pass to the one dimensional solver
!
        qz(idn,:) = qp(idn,i,j,:)
        qz(ivx,:) = qp(ivz,i,j,:)
        qz(ivy,:) = qp(ivx,i,j,:)
        qz(ivz,:) = qp(ivy,i,j,:)
        qz(ibx,:) = qp(ibz,i,j,:)
        qz(iby,:) = qp(ibx,i,j,:)
        qz(ibz,:) = qp(iby,i,j,:)
#ifdef ADI
        qz(ipr,:) = qp(ipr,i,j,:)
#endif /* ADI */
#ifdef GLM
        qz(iph,:) = qp(iph,i,j,:)
#endif /* GLM */
#ifdef VPOT
        bz(    :) = bi(inz,i,j,:)
#endif /* VPOT */
#ifdef SHOCK_DETECTION
        sz(    :) = sh(    i,j,:)
#endif /* SHOCK_DETECTION */

! call one dimensional Riemann solver in order to obtain numerical fluxes
!
        call riemann(km, dz, qz(:,:), bz(:), fz(:,:), sz(:))

! update the array of increments using numerical fluxes
!
        du(idn,i,j,2:km) = du(idn,i,j,2:km)                                    &
                                       + dzi * (fz(idn,1:kmm1) - fz(idn,2:km))
        du(idn,i,j,1   ) = du(idn,i,j,1   ) - dzi * fz(idn,1   )
        du(imx,i,j,2:km) = du(imx,i,j,2:km)                                    &
                                       + dzi * (fz(imy,1:kmm1) - fz(imy,2:km))
        du(imx,i,j,1   ) = du(imx,i,j,1   ) - dzi * fz(imy,1   )
        du(imy,i,j,2:km) = du(imy,i,j,2:km)                                    &
                                       + dzi * (fz(imz,1:kmm1) - fz(imz,2:km))
        du(imy,i,j,1   ) = du(imy,i,j,1   ) - dzi * fz(imz,1   )
        du(imz,i,j,2:km) = du(imz,i,j,2:km)                                    &
                                       + dzi * (fz(imx,1:kmm1) - fz(imx,2:km))
        du(imz,i,j,1   ) = du(imz,i,j,1   ) - dzi * fz(imx,1   )
#ifdef ADI
        du(ien,i,j,2:km) = du(ien,i,j,2:km)                                    &
                                       + dzi * (fz(ien,1:kmm1) - fz(ien,2:km))
        du(ien,i,j,1   ) = du(ien,i,j,1   ) - dzi * fz(ien,1   )
#endif /* ADI */
#ifdef GLM
        du(ibx,i,j,2:km) = du(ibx,i,j,2:km)                                    &
                                       + dzi * (fz(iby,1:kmm1) - fz(iby,2:km))
        du(ibx,i,j,1   ) = du(ibx,i,j,1   ) - dzi * fz(iby,1   )
        du(iby,i,j,2:km) = du(iby,i,j,2:km)                                    &
                                       + dzi * (fz(ibz,1:kmm1) - fz(ibz,2:km))
        du(iby,i,j,1   ) = du(iby,i,j,1   ) - dzi * fz(ibz,1   )
        du(ibz,i,j,2:km) = du(ibz,i,j,2:km)                                    &
                                       + dzi * (fz(ibx,1:kmm1) - fz(ibx,2:km))
        du(ibz,i,j,1   ) = du(ibz,i,j,1   ) - dzi * fz(ibx,1   )
        du(iph,i,j,2:km) = du(iph,i,j,2:km)                                    &
                                       + dzi * (fz(iph,1:kmm1) - fz(iph,2:km))
        du(iph,i,j,1   ) = du(iph,i,j,1   ) - dzi * fz(iph,1   )
#endif /* GLM */

! update electric field
!
#ifdef CD
#ifndef FIELDCD
        du(iey,i,j,2:km) = du(iey,i,j,2:km) + (fz(iby,1:kmm1) + fz(iby,2:km  ))
        du(iey,i,j,1   ) = du(iey,i,j,1   ) +         2.0d+00 * fz(iby,1     )
        du(iex,i,j,2:km) = du(iex,i,j,2:km) - (fz(ibz,1:kmm1) + fz(ibz,2:km  ))
        du(iex,i,j,1   ) = du(iex,i,j,1   ) -         2.0d+00 * fz(ibz,1     )
#endif /* !FIELDCD */
#endif /* CD */
#ifdef CT
        gy(i,j,1:km) = fz(iby,1:km)
        gz(i,j,1:km) = fz(ibz,1:km)
#endif /* CT */

      end do
    end do

#ifdef CT
! average electric field
!
    du(iey,1:imm1,:,:) = du(iey,1:imm1,:,:) + (gy(1:imm1,:,:) + gy(2:im  ,:,:))
    du(iey,  im  ,:,:) = du(iey,  im  ,:,:) +         2.0d+00 * gy(  im  ,:,:)
    du(iex,:,1:jmm1,:) = du(iex,:,1:jmm1,:) - (gz(:,1:jmm1,:) + gz(:,2:jm  ,:))
    du(iex,:,  jm  ,:) = du(iex,:,  jm  ,:) -         2.0d+00 * gz(:,  jm  ,:)
#endif /* CT */
#endif /* R3D */

#ifdef CD
! average electric field
!
#ifdef FIELDCD
! the electric field is calculated from it's cell centered average values
!
#ifndef R2D
    du(iax,:,:,:) = qp(ivy,:,:,:) * qp(ibz,:,:,:)                              &
                  - qp(ivz,:,:,:) * qp(iby,:,:,:)
    du(iay,:,:,:) = qp(ivz,:,:,:) * qp(ibx,:,:,:)                              &
                  - qp(ivx,:,:,:) * qp(ibz,:,:,:)
#endif /* !R2D */
    du(iaz,:,:,:) = qp(ivx,:,:,:) * qp(iby,:,:,:)                              &
                  - qp(ivy,:,:,:) * qp(ibx,:,:,:)
#endif /* FIELDCD */
#ifdef FLUXCD
! the electric field is averaged from the interface fluxes returned by Riemann
! solver
!
#ifdef R2D5
    du(iax,:,:,:) = du(iex,:,:,:) / 2.0d+00
    du(iay,:,:,:) = du(iey,:,:,:) / 2.0d+00
#endif /* R2D5 */
#ifdef R3D
    du(iax,:,:,:) = du(iex,:,:,:) / 4.0d+00
    du(iay,:,:,:) = du(iey,:,:,:) / 4.0d+00
#endif /* R3D */
    du(iaz,:,:,:) = du(iez,:,:,:) / 4.0d+00
#endif /* FLUXCD */
#ifdef AVGCD
! the electric field is averaged from both above schemes
!
#ifdef R2D5
    du(iax,:,:,:) = (du(iex,:,:,:)                                             &
                   + 2.0d+00 * (qp(ivy,:,:,:) * qp(ibz,:,:,:)                  &
                              - qp(ivz,:,:,:) * qp(iby,:,:,:))) / 4.0d+00
    du(iay,:,:,:) = (du(iey,:,:,:)                                             &
                   + 2.0d+00 * (qp(ivz,:,:,:) * qp(ibx,:,:,:)                  &
                              - qp(ivx,:,:,:) * qp(ibz,:,:,:))) / 4.0d+00
#endif /* R2D5 */
#ifdef R3D
    du(iax,:,:,:) = (du(iex,:,:,:)                                             &
                   + 4.0d+00 * (qp(ivy,:,:,:) * qp(ibz,:,:,:)                  &
                              - qp(ivz,:,:,:) * qp(iby,:,:,:))) / 8.0d+00
    du(iay,:,:,:) = (du(iey,:,:,:)                                             &
                   + 4.0d+00 * (qp(ivz,:,:,:) * qp(ibx,:,:,:)                  &
                              - qp(ivx,:,:,:) * qp(ibz,:,:,:))) / 8.0d+00
#endif /* R3D */
    du(iaz,:,:,:) = (du(iez,:,:,:)                                             &
                   + 4.0d+00 * (qp(ivx,:,:,:) * qp(iby,:,:,:)                  &
                              - qp(ivy,:,:,:) * qp(ibx,:,:,:))) / 8.0d+00
#endif /* AVGCD */
#endif /* CD */
#ifdef CT
#ifdef R3D
    du(iax,:,:,:) = du(iex,:,:,:) / 4.0d+00
    du(iay,:,:,:) = du(iey,:,:,:) / 4.0d+00
#endif /* R3D */
    du(iaz,:,:,:) = du(iez,:,:,:) / 4.0d+00
#endif /* CT */

!-------------------------------------------------------------------------------
!
  end subroutine update
#ifdef HLL
!
!===============================================================================
!
! subroutine RIEMANN:
! ==================
!
!   Subroutine solves one dimensional Riemann problem using the HLL method.
!
!   Arguments:
!
!     n - the length of input vectors;
!     h - the spatial step;
!     q - the input array of primitive variables;
!     b - the input vector of the normal magnetic field component;
!     f - the output array of fluxes;
!     s - the input array of shock indicators;
!
!   References:
!
!     [1] Harten, A., Lax, P. D. & Van Leer, B.
!         "On Upstream Differencing and Godunov-Type Schemes for Hyperbolic
!          Conservation Laws",
!         SIAM Review, 1983, Volume 25, Number 1, pp. 35-61
!
!===============================================================================
!
  subroutine riemann(n, h, q, b, f, s)

! include external procedures and variables
!
    use equations     , only : prim2cons, fluxspeed
    use interpolations, only : reconstruct, fix_positivity
    use variables     , only : nt
    use variables     , only : idn, ivx, ibx
#ifdef ADI
    use variables     , only : ipr
#endif /* ADI */
#ifdef GLM
    use variables     , only : iph
    use variables     , only : cmax
#endif /* GLM */

! local variables are not implicit by default
!
    implicit none

! input / output arguments
!
    integer                 , intent(in)  :: n
    real                    , intent(in)  :: h
    real   , dimension(nt,n), intent(in)  :: q
    real   , dimension(   n), intent(in)  :: b
    real   , dimension(nt,n), intent(out) :: f
    logical, dimension(n)   , intent(in)  :: s

! local variables
!
    integer :: p, i
    real    :: sl, sr, srl, srml

! local arrays to store the states
!
    real, dimension(nt,n) :: ul, ur, ql, qr, fl, fr
    real, dimension(n)    :: cl, cr
!
!-------------------------------------------------------------------------------
!
#ifdef GLM
! reconstruct the left and right states from primitive variables
!
    do p = 1, nt
      call reconstruct(n, h, q(p,:), ql(p,:), qr(p,:), s(:))
    end do

! obtain the state values for Bx and Psi for the GLM-MHD equations
!
    cl(:) = 0.5d+00 * ((qr(ibx,:) + ql(ibx,:)) - (qr(iph,:) - ql(iph,:)) / cmax)
    cr(:) = 0.5d+00 * ((qr(iph,:) + ql(iph,:)) - (qr(ibx,:) - ql(ibx,:)) * cmax)
    ql(ibx,:) = cl(:)
    qr(ibx,:) = cl(:)
    ql(iph,:) = cr(:)
    qr(iph,:) = cr(:)
#else /* GLM */
! reconstruct the left and right states from primitive variables
!
    do p = 1, ibx - 1
      call reconstruct(n, h, q(p,:), ql(p,:), qr(p,:), s(:))
    end do
    do p = ibx + 1, nt
      call reconstruct(n, h, q(p,:), ql(p,:), qr(p,:), s(:))
    end do

! copy the normal component of magnetic field to the left and right states
!
    ql(ibx,:) = b(:)
    qr(ibx,:) = b(:)
#endif /* GLM */

#ifdef FIX_POSITIVITY
! check if the reconstruction doesn't give the negative density or pressure,
! if so, correct the states
!
    call fix_positivity(n, q(idn,:), ql(idn,:), qr(idn,:))
#ifdef ADI
    call fix_positivity(n, q(ipr,:), ql(ipr,:), qr(ipr,:))
#endif /* ADI */
#endif /* FIX_POSITIVITY */

! calculate corresponding conserved variables of the left and right states
!
    call prim2cons(n, ql(:,:), ul(:,:))
    call prim2cons(n, qr(:,:), ur(:,:))

! calculate the physical fluxes and speeds at the states
!
    call fluxspeed(n, ql(:,:), ul(:,:), fl(:,:), cl(:))
    call fluxspeed(n, qr(:,:), ur(:,:), fr(:,:), cr(:))

! iterate over all points
!
    do i = 1, n

! estimate the minimum and maximum speeds
!
      sl = min(ql(ivx,i) - cl(i), qr(ivx,i) - cr(i))
      sr = max(ql(ivx,i) + cl(i), qr(ivx,i) + cr(i))

! calculate the HLL flux
!
      if (sl >= 0.0d0) then

        f(:,i) = fl(:,i)

      else if (sr <= 0.0d0) then

        f(:,i) = fr(:,i)

      else

! prepare coefficients for the intermediate state calculation
!
        srl  = sr * sl
        srml = sr - sl

! calculate the fluxes for the intermediate state
!
        f(:,i) = ((sr * fl(:,i) - sl * fr(:,i))                                &
                                           + srl * (ur(:,i) - ul(:,i))) / srml

      end if

    end do

!-------------------------------------------------------------------------------
!
  end subroutine riemann
#endif /* HLL */
#if defined HLLC || defined HLLCC || defined HLLCL
!
!===============================================================================
!
! subroutine RIEMANN:
! ==================
!
!   Subroutine solves one dimensional Riemann problem using the HLLC method
!   by Gurski or Li.  The HLLC, HLLCC, and HLLC-L differs by the definitions of
!   the tangential components of the velocity and magnetic field.
!
!   Arguments:
!
!     n - the length of input vectors;
!     h - the spatial step;
!     q - the input array of primitive variables;
!     b - the input vector of the normal magnetic field component;
!     f - the output array of fluxes;
!     s - the input array of shock indicators;
!
!   References:
!
!     [1] Gurski, K. F.,
!         "An HLLC-Type Approximate Riemann Solver for Ideal
!          Magnetohydrodynamics",
!         SIAM Journal on Scientific Computing, 2004, Volume 25, Issue 6,
!         pp. 2165–2187
!     [2] Li, S.,
!         "An HLLC Riemann solver for magneto-hydrodynamics",
!         Journal of Computational Physics, 2005, Volume 203, Issue 1,
!         pp. 344-357
!
!===============================================================================
!
  subroutine riemann(n, h, q, b, f, s)

! include external procedures and variables
!
    use equations     , only : prim2cons, fluxspeed
    use interpolations, only : reconstruct, fix_positivity
    use variables     , only : nt
    use variables     , only : idn, imx, imy, imz, ien
    use variables     , only : ivx, ivy, ivz, ipr
    use variables     , only : ibx, iby, ibz
#ifdef GLM
    use variables     , only : iph
    use variables     , only : cmax
#endif /* GLM */

! local variables are not implicit by default
!
    implicit none

! input / output arguments
!
    integer                 , intent(in)  :: n
    real                    , intent(in)  :: h
    real   , dimension(nt,n), intent(in)  :: q
    real   , dimension(   n), intent(in)  :: b
    real   , dimension(nt,n), intent(out) :: f
    logical, dimension(n)   , intent(in)  :: s

! local variables
!
    integer :: p, i
#ifdef SHOCK_DETECTION
    integer :: ip1
#endif /* SHOCK_DETECTION */
    real    :: dn, mx, vy, vz, by, bz, pt, pm
    real    :: sl, sr, srl, sm
    real    :: slmv, srmv, slmm, srmm, smvl, smvr, srml
    real    :: pml, pmr, ptl, ptr
    real    :: fc

! local arrays to store the states
!
    real, dimension(nt,n) :: ul, ur, ql, qr, fl, fr
    real, dimension(n)    :: cl, cr
    real, dimension(nt)   :: q1l, q1r, u1l, u1r
!
!-------------------------------------------------------------------------------
!
#ifdef GLM
! reconstruct the left and right states from primitive variables
!
    do p = 1, nt
      call reconstruct(n, h, q(p,:), ql(p,:), qr(p,:), s(:))
    end do

! obtain the state values for Bx and Psi for the GLM-MHD equations
!
    cl(:) = 0.5d+00 * ((qr(ibx,:) + ql(ibx,:)) - (qr(iph,:) - ql(iph,:)) / cmax)
    cr(:) = 0.5d+00 * ((qr(iph,:) + ql(iph,:)) - (qr(ibx,:) - ql(ibx,:)) * cmax)
    ql(ibx,:) = cl(:)
    qr(ibx,:) = cl(:)
    ql(iph,:) = cr(:)
    qr(iph,:) = cr(:)
#else /* GLM */
! reconstruct the left and right states from primitive variables
!
    do p = 1, ibx - 1
      call reconstruct(n, h, q(p,:), ql(p,:), qr(p,:), s(:))
    end do
    do p = ibx + 1, nt
      call reconstruct(n, h, q(p,:), ql(p,:), qr(p,:), s(:))
    end do

! copy the normal component of magnetic field to the left and right states
!
    ql(ibx,:) = b(:)
    qr(ibx,:) = b(:)
#endif /* GLM */

#ifdef FIX_POSITIVITY
! check if the reconstruction doesn't give the negative density or pressure,
! if so, correct the states
!
    call fix_positivity(n, q(idn,:), ql(idn,:), qr(idn,:))
    call fix_positivity(n, q(ipr,:), ql(ipr,:), qr(ipr,:))
#endif /* FIX_POSITIVITY */

! calculate corresponding conserved variables of the left and right states
!
    call prim2cons(n, ql(:,:), ul(:,:))
    call prim2cons(n, qr(:,:), ur(:,:))

! calculate the physical fluxes and speeds at the states
!
    call fluxspeed(n, ql(:,:), ul(:,:), fl(:,:), cl(:))
    call fluxspeed(n, qr(:,:), ur(:,:), fr(:,:), cr(:))

! iterate over all points
!
    do i = 1, n

! estimate the minimum and maximum speeds
!
      sl = min(ql(ivx,i) - cl(i), qr(ivx,i) - cr(i))
      sr = max(ql(ivx,i) + cl(i), qr(ivx,i) + cr(i))

! calculate the HLLC flux
!
      if (sl >= 0.0d0) then

        f(:,i) = fl(:,i)

      else if (sr <= 0.0d0) then

        f(:,i) = fr(:,i)

      else ! sl < 0 < sr

#ifdef SHOCK_DETECTION
! calculate the next index
!
        ip1 = min(n, i + 1)

! check if we are in the vicinity of a shock
!
        if (s(i) .or. s(ip1)) then

! prepare coefficients for the intermediate state calculation
!
          srl  = sr * sl
          srml = sr - sl

! calculate the fluxes for the intermediate state
!
          f(:,i) = ((sr * fl(:,i) - sl * fr(:,i))                              &
                                           + srl * (ur(:,i) - ul(:,i))) / srml

        else
#endif /* SHOCK_DETECTION */

! the speed of contact discontinuity (eq. 34 [1], 14 [2])
!
          dn = (sr * ur(idn,i) - sl * ul(idn,i)) - (fr(idn,i) - fl(idn,i))
          mx = (sr * ur(imx,i) - sl * ul(imx,i)) - (fr(imx,i) - fl(imx,i))
          sm = mx / dn

! compute speed differences
!
          slmv = sl - ql(ivx,i)
          srmv = sr - qr(ivx,i)
          slmm = sl - sm
          srmm = sr - sm
          smvl = sm - ql(ivx,i)
          smvr = sm - qr(ivx,i)

! calculate magnetic pressure
!
          pml = 0.5d0 * sum(ql(ibx:ibz,i) * ql(ibx:ibz,i))
          pmr = 0.5d0 * sum(qr(ibx:ibz,i) * qr(ibx:ibz,i))

! calculate total pressures
!
          ptl = ql(ipr,i) + pml
          ptr = qr(ipr,i) + pmr

! pressure of intermediate states (eq. 36 [1], 16 [2])
!
          pt = 0.5d0 * ((ptl + ql(idn,i) * slmv * smvl)                        &
                      + (ptr + qr(idn,i) * srmv * smvr))

#if defined HLLC || defined HLLCL
! separate cases when Bx = 0 and Bx ≠ 0
!
          if (ql(ibx,i) == 0.0d0) then

! intermediate discontinuities
!
            if (sm > 0.0d0) then

! calculate the ratio of speed differences
!
              fc = slmv / slmm

! calculate density
!
              dn = ql(idn,i) * fc

! calculate the left intermediate state
!
              u1l(idn) = dn
              u1l(imx) = dn * sm
              u1l(imy) = dn * ql(ivy,i)
              u1l(imz) = dn * ql(ivz,i)
              u1l(ibx) = ql(ibx,i)
              u1l(iby) = ql(iby,i) * fc
              u1l(ibz) = ql(ibz,i) * fc
#ifdef GLM
              u1l(iph) = ql(iph,i)
#endif /* GLM */
              u1l(ien) = (slmv * ul(ien,i) - ptl * ql(ivx,i) + pt * sm) / slmm

! calculate the left intermediate flux
!
              f(:,i) = fl(:,i) + sl * (u1l(:) - ul(:,i))

            else if (sm < 0.0d0) then

! calculate the ratio of speed differences
!
              fc = srmv / srmm

! calculate density
!
              dn = qr(idn,i) * fc

! calculate the right intermediate state
!
              u1r(idn) = dn
              u1r(imx) = dn * sm
              u1r(imy) = dn * qr(ivy,i)
              u1r(imz) = dn * qr(ivz,i)
              u1r(ibx) = qr(ibx,i)
              u1r(iby) = qr(iby,i) * fc
              u1r(ibz) = qr(ibz,i) * fc
#ifdef GLM
              u1r(iph) = qr(iph,i)
#endif /* GLM */
              u1r(ien) = (srmv * ur(ien,i) - ptr * qr(ivx,i) + pt * sm) / srmm

! calculate right intermediate flux
!
              f(:,i) = fr(:,i) + sr * (u1r(:) - ur(:,i))

            else ! sm = 0

! calculate the ratio of speed differences
!
              fc = slmv / sl

! calculate density
!
              dn = ql(idn,i) * fc

! calculate the left intermediate state
!
              u1l(idn) = dn
              u1l(imx) = 0.0d0
              u1l(imy) = dn * ql(ivy,i)
              u1l(imz) = dn * ql(ivz,i)
              u1l(ibx) = ql(ibx,i)
              u1l(iby) = ql(iby,i) * fc
              u1l(ibz) = ql(ibz,i) * fc
              u1l(ien) = (slmv * ul(ien,i) - ptl * ql(ivx,i)) / sl
#ifdef GLM
              u1l(iph) = ql(iph,i)
#endif /* GLM */

! calculate the left intermediate flux
!
              f(:,i) = fl(:,i) + sl * (u1l(:) - ul(:,i))

! calculate the ratio of speed differences
!
              fc = srmv / sr

! calculate density
!
              dn = qr(idn,i) * fc

! calculate the right intermediate state
!
              u1r(idn) = dn
              u1r(imx) = 0.0d0
              u1r(imy) = dn * qr(ivy,i)
              u1r(imz) = dn * qr(ivz,i)
              u1r(ibx) = qr(ibx,i)
              u1r(iby) = qr(iby,i) * fc
              u1r(ibz) = qr(ibz,i) * fc
              u1r(ien) = (srmv * ur(ien,i) - ptr * qr(ivx,i)) / sr
#ifdef GLM
              u1r(iph) = qr(iph,i)
#endif /* GLM */

! calculate right intermediate flux
!
              f(:,i) = 0.5d0 * (f(:,i) + (fr(:,i) + sr * (u1r(:) - ur(:,i))))

            end if ! sm = 0

          else ! Bx ≠ 0

! prepare the difference between the fastests speeds
!
            srml = sr - sl

! when Bx ≠ 0 the tangential components of velocity and magnetic field are
! calculated from the HLL state
!
#ifdef HLLC
            vy = ((sr * ur(imy,i) - sl * ul(imy,i))                            &
                                             - (fr(imy,i) - fl(imy,i))) / dn
            vz = ((sr * ur(imz,i) - sl * ul(imz,i))                            &
                                             - (fr(imz,i) - fl(imz,i))) / dn
#endif /* HLLC */
            by = ((sr * ur(iby,i) - sl * ul(iby,i))                            &
                                             - (fr(iby,i) - fl(iby,i))) / srml
            bz = ((sr * ur(ibz,i) - sl * ul(ibz,i))                            &
                                             - (fr(ibz,i) - fl(ibz,i))) / srml

! intermediate discontinuities
!
            if (sm > 0.0d0) then

! calculate the ratio of speed differences
!
              fc = slmv / slmm

! calculate density
!
              dn = ql(idn,i) * fc

! calculate the left intermediate state
!
              u1l(idn) = dn
              u1l(imx) = dn * sm
#ifdef HLLC
              u1l(imy) = dn * vy
              u1l(imz) = dn * vz
#endif /* HLLC */
#ifdef HLLCL
              fc = ql(ibx,i) / (ql(idn,i) * slmv)
              u1l(imy) = dn * (ql(ivy,i) + fc * (ql(iby,i) - by))
              u1l(imz) = dn * (ql(ivz,i) + fc * (ql(ibz,i) - bz))
#endif /* HLLCL */
              u1l(ibx) = ql(ibx,i)
              u1l(iby) = by
              u1l(ibz) = bz
#ifdef GLM
              u1l(iph) = ql(iph,i)
#endif /* GLM */
              u1l(ien) = (slmv * ul(ien,i) - ptl * ql(ivx,i) + pt * sm) / slmm

! calculate the left intermediate flux
!
              f(:,i) = fl(:,i) + sl * (u1l(:) - ul(:,i))

            else if (sm < 0.0d0) then

! calculate the ratio of speed differences
!
              fc = srmv / srmm

! calculate density
!
              dn = qr(idn,i) * fc

! calculate the right intermediate state
!
              u1r(idn) = dn
              u1r(imx) = dn * sm
#ifdef HLLC
              u1r(imy) = dn * vy
              u1r(imz) = dn * vz
#endif /* HLLC */
#ifdef HLLCL
              fc = qr(ibx,i) / (qr(idn,i) * srmv)
              u1r(imy) = dn * (qr(ivy,i) + fc * (qr(iby,i) - by))
              u1r(imz) = dn * (qr(ivz,i) + fc * (qr(ibz,i) - bz))
#endif /* HLLCL */
              u1r(ibx) = qr(ibx,i)
              u1r(iby) = by
              u1r(ibz) = bz
#ifdef GLM
              u1r(iph) = qr(iph,i)
#endif /* GLM */
              u1r(ien) = (srmv * ur(ien,i) - ptr * qr(ivx,i) + pt * sm) / srmm

! calculate right intermediate flux
!
              f(:,i) = fr(:,i) + sr * (u1r(:) - ur(:,i))

            else ! sm = 0

! calculate the ratio of speed differences
!
              fc = slmv / sl

! calculate density
!
              dn = ql(idn,i) * fc

! calculate the left intermediate state
!
              u1l(idn) = dn
              u1l(imx) = 0.0d0
#ifdef HLLC
              u1l(imy) = dn * vy
              u1l(imz) = dn * vz
#endif /* HLLC */
#ifdef HLLCL
              fc = ql(ibx,i) / (ql(idn,i) * slmv)
              u1l(imy) = dn * (ql(ivy,i) + fc * (ql(iby,i) - by))
              u1l(imz) = dn * (ql(ivz,i) + fc * (ql(ibz,i) - bz))
#endif /* HLLCL */
              u1l(ibx) = ql(ibx,i)
              u1l(iby) = by
              u1l(ibz) = bz
#ifdef GLM
              u1l(iph) = ql(iph,i)
#endif /* GLM */
              u1l(ien) = (slmv * ul(ien,i) - ptl * ql(ivx,i)) / sl

! calculate the left intermediate flux
!
              f(:,i) = fl(:,i) + sl * (u1l(:) - ul(:,i))

! calculate the ratio of speed differences
!
              fc = srmv / sr

! calculate density
!
              dn = qr(idn,i) * fc

! calculate the right intermediate state
!
              u1r(idn) = dn
              u1r(imx) = 0.0d0
#ifdef HLLC
              u1r(imy) = dn * vy
              u1r(imz) = dn * vz
#endif /* HLLC */
#ifdef HLLCL
              fc = qr(ibx,i) / (qr(idn,i) * srmv)
              u1r(imy) = dn * (qr(ivy,i) + fc * (qr(iby,i) - by))
              u1r(imz) = dn * (qr(ivz,i) + fc * (qr(ibz,i) - bz))
#endif /* HLLCL */
              u1r(ibx) = qr(ibx,i)
              u1r(iby) = by
              u1r(ibz) = bz
#ifdef GLM
              u1r(iph) = qr(iph,i)
#endif /* GLM */
              u1r(ien) = (srmv * ur(ien,i) - ptr * qr(ivx,i)) / sr

! calculate right intermediate flux
!
              f(:,i) = 0.5d0 * (f(:,i) + (fr(:,i) + sr * (u1r(:) - ur(:,i))))

            end if ! sm = 0

          end if ! Bx ≠ 0
#endif /* HLLC | HLLCL */
#ifdef HLLCC
! prepare the difference between the fastests speeds
!
          srml = sr - sl

! when Bx ≠ 0 the tangential components of velocity and magnetic field are
! calculated from the HLL state
!
          vy = ((sr * ur(imy,i) - sl * ul(imy,i))                              &
                                             - (fr(imy,i) - fl(imy,i))) / dn
          vz = ((sr * ur(imz,i) - sl * ul(imz,i))                              &
                                             - (fr(imz,i) - fl(imz,i))) / dn
          by = ((sr * ur(iby,i) - sl * ul(iby,i))                              &
                                             - (fr(iby,i) - fl(iby,i))) / srml
          bz = ((sr * ur(ibz,i) - sl * ul(ibz,i))                              &
                                             - (fr(ibz,i) - fl(ibz,i))) / srml

! intermediate discontinuities
!
          if (sm > 0.0d0) then

! calculate the ratio of speed differences
!
            fc = slmv / slmm

! calculate density
!
            dn = ql(idn,i) * fc

! calculate the left intermediate state
!
            u1l(idn) = dn
            u1l(imx) = dn * sm
            u1l(imy) = dn * vy
            u1l(imz) = dn * vz
            u1l(ibx) = ql(ibx,i)
            u1l(iby) = by
            u1l(ibz) = bz
#ifdef GLM
            u1l(iph) = ql(iph,i)
#endif /* GLM */
            u1l(ien) = (slmv * ul(ien,i) - ptl * ql(ivx,i) + pt * sm) / slmm

! calculate the left intermediate flux
!
            f(:,i) = fl(:,i) + sl * (u1l(:) - ul(:,i))

          else if (sm < 0.0d0) then

! calculate the ratio of speed differences
!
            fc = srmv / srmm

! calculate density
!
            dn = qr(idn,i) * fc

! calculate the right intermediate state
!
            u1r(idn) = dn
            u1r(imx) = dn * sm
            u1r(imy) = dn * vy
            u1r(imz) = dn * vz
            u1r(ibx) = qr(ibx,i)
            u1r(iby) = by
            u1r(ibz) = bz
#ifdef GLM
            u1r(iph) = qr(iph,i)
#endif /* GLM */
            u1r(ien) = (srmv * ur(ien,i) - ptr * qr(ivx,i) + pt * sm) / srmm

! calculate right intermediate flux
!
            f(:,i) = fr(:,i) + sr * (u1r(:) - ur(:,i))

          else ! sm = 0

! calculate the ratio of speed differences
!
            fc = slmv / sl

! calculate density
!
            dn = ql(idn,i) * fc

! calculate the left intermediate state
!
            u1l(idn) = dn
            u1l(imx) = 0.0d0
            u1l(imy) = dn * vy
            u1l(imz) = dn * vz
            u1l(ibx) = ql(ibx,i)
            u1l(iby) = by
            u1l(ibz) = bz
#ifdef GLM
            u1l(iph) = ql(iph,i)
#endif /* GLM */
            u1l(ien) = (slmv * ul(ien,i) - ptl * ql(ivx,i)) / sl

! calculate the left intermediate flux
!
            f(:,i) = fl(:,i) + sl * (u1l(:) - ul(:,i))

! calculate the ratio of speed differences
!
            fc = srmv / sr

! calculate density
!
            dn = qr(idn,i) * fc

! calculate the right intermediate state
!
            u1r(idn) = dn
            u1r(imx) = 0.0d0
            u1r(imy) = dn * vy
            u1r(imz) = dn * vz
            u1r(ibx) = qr(ibx,i)
            u1r(iby) = by
            u1r(ibz) = bz
#ifdef GLM
            u1r(iph) = qr(iph,i)
#endif /* GLM */
            u1r(ien) = (srmv * ur(ien,i) - ptr * qr(ivx,i)) / sr

! calculate right intermediate flux
!
            f(:,i) = 0.5d0 * (f(:,i) + (fr(:,i) + sr * (u1r(:) - ur(:,i))))

          end if ! sm = 0
#endif /* HLLCC */

#ifdef SHOCK_DETECTION
        end if ! shock vicinity
#endif /* SHOCK_DETECTION */

      end if ! sl < 0.0 & sr > 0.0

    end do

!-------------------------------------------------------------------------------
!
  end subroutine riemann
#endif /* HLLC | HLLCC | HLLCL */
#ifdef HLLD
#ifdef ADI
!
!===============================================================================
!
! subroutine RIEMANN:
! ------------------
!
!   Subroutine solves one dimensional Riemann problem using the adiabatic HLLD
!   method described by Miyoshi & Kusano.
!
!   Arguments:
!
!     n - the length of input vectors;
!     h - the spatial step;
!     q - the input array of primitive variables;
!     b - the input vector of the normal magnetic field component;
!     f - the output array of fluxes;
!     s - the input array of shock indicators;
!
!   References:
!
!     [1] Miyoshi, T. & Kusano, K.,
!         "A multi-state HLL approximate Riemann solver for ideal
!          magnetohydrodynamics",
!         Journal of Computational Physics, 2005, 208, pp. 315-344
!
!===============================================================================
!
  subroutine riemann(n, h, q, b, f, s)

! include external procedures and variables
!
    use equations     , only : gamma
    use equations     , only : prim2cons, fluxspeed
    use interpolations, only : reconstruct, fix_positivity
    use variables     , only : nt
    use variables     , only : idn, imx, imy, imz, ien
    use variables     , only : ivx, ivy, ivz, ipr
    use variables     , only : ibx, iby, ibz
#ifdef GLM
    use variables     , only : iph
    use variables     , only : cmax
#endif /* GLM */

! local variables are not implicit by default
!
    implicit none

! input / output arguments
!
    integer                 , intent(in)  :: n
    real                    , intent(in)  :: h
    real   , dimension(nt,n), intent(in)  :: q
    real   , dimension(   n), intent(in)  :: b
    real   , dimension(nt,n), intent(out) :: f
    logical, dimension(n)   , intent(in)  :: s

! local variables
!
    integer :: p, i
#ifdef SHOCK_DETECTION
    integer :: ip1
#endif /* SHOCK_DETECTION */
    real    :: sl, sr, sm, sml, smr
    real    :: srl, srml, slmv, srmv, slmm, srmm, smvl, smvr
    real    :: dn, dnl, dnr, dlsq, drsq
    real    :: mx, dv, fc, b2, bs, ds, vbl, vbr, vb1l, vb1r, vb2
    real    :: pml, pmr, ptl, ptr, pt, pm

! local arrays to store the states
!
    real, dimension(nt,n) :: ul, ur, ql, qr, fl, fr
    real, dimension(n)    :: cl, cr
    real, dimension(nt)   :: q1l, q1r, u1l, u1r, q2, u2
!
!-------------------------------------------------------------------------------
!
#ifdef GLM
! reconstruct the left and right states from primitive variables
!
    do p = 1, nt
      call reconstruct(n, h, q(p,:), ql(p,:), qr(p,:), s(:))
    end do

! obtain the state values for Bx and Psi for the GLM-MHD equations
!
    cl(:) = 0.5d+00 * ((qr(ibx,:) + ql(ibx,:)) - (qr(iph,:) - ql(iph,:)) / cmax)
    cr(:) = 0.5d+00 * ((qr(iph,:) + ql(iph,:)) - (qr(ibx,:) - ql(ibx,:)) * cmax)
    ql(ibx,:) = cl(:)
    qr(ibx,:) = cl(:)
    ql(iph,:) = cr(:)
    qr(iph,:) = cr(:)
#else /* GLM */
! reconstruct the left and right states from primitive variables
!
    do p = 1, ibx - 1
      call reconstruct(n, h, q(p,:), ql(p,:), qr(p,:), s(:))
    end do
    do p = ibx + 1, nt
      call reconstruct(n, h, q(p,:), ql(p,:), qr(p,:), s(:))
    end do

! copy the normal component of magnetic field to the left and right states
!
    ql(ibx,:) = b(:)
    qr(ibx,:) = b(:)
#endif /* GLM */

#ifdef FIX_POSITIVITY
! check if the reconstruction doesn't give the negative density or pressure,
! if so, correct the states
!
    call fix_positivity(n, q(idn,:), ql(idn,:), qr(idn,:))
    call fix_positivity(n, q(ipr,:), ql(ipr,:), qr(ipr,:))
#endif /* FIX_POSITIVITY */

! calculate corresponding conserved variables of the left and right states
!
    call prim2cons(n, ql(:,:), ul(:,:))
    call prim2cons(n, qr(:,:), ur(:,:))

! calculate the physical fluxes and speeds at the states
!
    call fluxspeed(n, ql(:,:), ul(:,:), fl(:,:), cl(:))
    call fluxspeed(n, qr(:,:), ur(:,:), fr(:,:), cr(:))

! iterate over all points
!
    do i = 1, n

! estimate the minimum and maximum speeds
!
      sl = min(ql(ivx,i) - cl(i), qr(ivx,i) - cr(i))
      sr = max(ql(ivx,i) + cl(i), qr(ivx,i) + cr(i))

! calculate the HLLD flux
!
      if (sl >= 0.0d0) then

        f(:,i) = fl(:,i)

      else if (sr <= 0.0d0) then

        f(:,i) = fr(:,i)

      else ! sl < 0 < sr

#ifdef SHOCK_DETECTION
! calculate the next index
!
        ip1 = min(n, i + 1)

! check if we are in the vicinity of a shock
!
        if (s(i) .or. s(ip1)) then

! prepare coefficients for the intermediate state calculation
!
          srl  = sr * sl
          srml = sr - sl

! calculate the fluxes for the intermediate state
!
          f(:,i) = ((sr * fl(:,i) - sl * fr(:,i))                              &
                                           + srl * (ur(:,i) - ul(:,i))) / srml

        else
#endif /* SHOCK_DETECTION */

! the speed of contact discontinuity (eq. 38)
!
          dn = (sr * ur(idn,i) - sl * ul(idn,i)) - (fr(idn,i) - fl(idn,i))
          mx = (sr * ur(imx,i) - sl * ul(imx,i)) - (fr(imx,i) - fl(imx,i))
          sm = mx / dn

! prepare speed differences
!
          slmv = sl - ql(ivx,i)
          srmv = sr - qr(ivx,i)
          slmm = sl - sm
          srmm = sr - sm
          smvl = sm - ql(ivx,i)
          smvr = sm - qr(ivx,i)
          srml = sr - sl

! prepare some coefficients
!
          dnl  = ql(idn,i) * slmv
          dnr  = qr(idn,i) * srmv
          b2   = ql(ibx,i) * qr(ibx,i)
          bs   = b2 / gamma

! calculate magnetic pressure
!
          pml = 0.5d0 * sum(ql(ibx:ibz,i) * ql(ibx:ibz,i))
          pmr = 0.5d0 * sum(qr(ibx:ibz,i) * qr(ibx:ibz,i))

! calculate total pressures
!
          ptl = ql(ipr,i) + pml
          ptr = qr(ipr,i) + pmr

! the pressure of the intermediate state (eq. 23)
!
          pt = 0.5d0 * ((ptl + dnl * smvl) + (ptr + dnr * smvr))

! calculate densities for the left and right intermediate states (eq. 43)
!
          q1l(idn) = dnl / slmm
          q1r(idn) = dnr / srmm

! substitute the normal components of velocity and magnetic field
!
          q1l(ivx) = sm
          q1r(ivx) = sm
          q1l(ibx) = ql(ibx,i)
          q1r(ibx) = qr(ibx,i)

! calculate transversal components of  magnetic field for the left state
! (eq. 45 & 47); take into account the B² > γp degeneracy
!
          dv = dnl * slmm - b2
          if (dv /= 0.0d0 .and. bs <= ql(ipr,i)) then

            fc       = (dnl * slmv - b2) / dv

            q1l(iby) = ql(iby,i) * fc
            q1l(ibz) = ql(ibz,i) * fc

          else

! in the case of degeneracy, calculate transversal components of magnetic field
! from the HLL states
!
            q1l(iby) = ((sr * ur(iby,i) - sl * ul(iby,i))                      &
                                             - (fr(iby,i) - fl(iby,i))) / srml
            q1l(ibz) = ((sr * ur(ibz,i) - sl * ul(ibz,i))                      &
                                             - (fr(ibz,i) - fl(ibz,i))) / srml
          end if

! calculate transversal components of velocity (eq. 42)
!
          dv = ql(ibx,i) / dnl
          q1l(ivy) = ql(ivy,i) + dv * (ql(iby,i) - q1l(iby))
          q1l(ivz) = ql(ivz,i) + dv * (ql(ibz,i) - q1l(ibz))

! calculate transversal components of  magnetic field for the left state
! (eq. 45 & 47); take into account the B² > γp degeneracy
!
          dv = dnr * srmm - b2
          if (dv /= 0.0d0 .and. bs <= qr(ipr,i)) then

            fc       = (dnr * srmv - b2) / dv

            q1r(iby) = qr(iby,i) * fc
            q1r(ibz) = qr(ibz,i) * fc

          else

! in the case of degeneracy, calculate transversal components of magnetic field
! from the HLL states
!
            q1r(iby) = ((sr * ur(iby,i) - sl * ul(iby,i))                      &
                                             - (fr(iby,i) - fl(iby,i))) / srml
            q1r(ibz) = ((sr * ur(ibz,i) - sl * ul(ibz,i))                      &
                                             - (fr(ibz,i) - fl(ibz,i))) / srml

          end if

! calculate transversal components of velocity (eq. 42)
!
          dv = qr(ibx,i) / dnr
          q1r(ivy) = qr(ivy,i) + dv * (qr(iby,i) - q1r(iby))
          q1r(ivz) = qr(ivz,i) + dv * (qr(ibz,i) - q1r(ibz))

! calculate scalar products of the velocity and magnetic field
!
          vbl  = sum(ql (ivx:ivz,i) * ql (ibx:ibz,i))
          vbr  = sum(qr (ivx:ivz,i) * qr (ibx:ibz,i))
          vb1l = sum(q1l(ivx:ivz)   * q1l(ibx:ibz))
          vb1r = sum(q1r(ivx:ivz)   * q1r(ibx:ibz))

! convert the left intermediate state to conservative form
!
          u1l(idn) = q1l(idn)
          u1l(imx) = q1l(idn) * q1l(ivx)
          u1l(imy) = q1l(idn) * q1l(ivy)
          u1l(imz) = q1l(idn) * q1l(ivz)
          u1l(ibx) = q1l(ibx)
          u1l(iby) = q1l(iby)
          u1l(ibz) = q1l(ibz)
#ifdef GLM
          u1l(iph) = ql(iph,i)
#endif /* GLM */

! convert the right intermediate state to conservative form
!
          u1r(idn) = q1r(idn)
          u1r(imx) = q1r(idn) * q1r(ivx)
          u1r(imy) = q1r(idn) * q1r(ivy)
          u1r(imz) = q1r(idn) * q1r(ivz)
          u1r(ibx) = q1r(ibx)
          u1r(iby) = q1r(iby)
          u1r(ibz) = q1r(ibz)
#ifdef GLM
          u1r(iph) = qr(iph,i)
#endif /* GLM */

! calculate the total energy of the left intermediate state (eq. 48)
!
          u1l(ien) = (slmv * ul(ien,i) - ptl * ql(ivx,i) + pt * sm             &
                                            + ql(ibx,i) * (vbl - vb1l)) / slmm

! calculate the total energy of the right intermediate state (eq. 48)
!
          u1r(ien) = (srmv * ur(ien,i) - ptr * qr(ivx,i) + pt * sm             &
                                            + qr(ibx,i) * (vbr - vb1r)) / srmm

! separate cases Bx ≠ 0 and Bx = 0
!
          if (abs(ql(ibx,i)) > 0.0d0) then

! calculate the square root of the left and right intermediate densities
!
            dlsq = sqrt(q1l(idn))
            drsq = sqrt(q1r(idn))

! calculate the left and right going sound plus Alfvén speeds (eq. 51)
!
            sml = sm - abs(ql(ibx,i)) / dlsq
            smr = sm + abs(qr(ibx,i)) / drsq

! intermediate discontinuities
!
            if (sml > 0.0d0) then

! calculate the left intermediate flux, eq. (64)
!
              f(:,i) = fl(:,i) + sl * (u1l(:) - ul(:,i))

            else if (smr < 0.0d0) then

! calculate the right intermediate flux, eq. (64)
!
              f(:,i) = fr(:,i) + sr * (u1r(:) - ur(:,i))

            else ! sml ≤ 0 ≤ smr

! prepare the sign of the normal component of magnetic field
!
              if (ql(ibx,i) >= 0.0d0) then
                bs =   1.0d0
              else
                bs = - 1.0d0
              end if

! calculate the sum and product of density root squares
!
              dv   = dlsq + drsq
              ds   = dlsq * drsq

! calculate velocity components, (eq. 39, and 59, 60)
!
              q2(ivx) = sm
              q2(ivy) = ((dlsq * q1l(ivy) + drsq * q1r(ivy))                   &
                                            + bs * (q1r(iby) - q1l(iby))) / dv
              q2(ivz) = ((dlsq * q1l(ivz) + drsq * q1r(ivz))                   &
                                            + bs * (q1r(ibz) - q1l(ibz))) / dv

! calculate magnetic field components, (eq. 61, 62)
!
              q2(ibx) = ql(ibx,i)
              q2(iby) = ((dlsq * q1r(iby) + drsq * q1l(iby))                   &
                                       + bs * ds * (q1r(ivy) - q1l(ivy))) / dv
              q2(ibz) = ((dlsq * q1r(ibz) + drsq * q1l(ibz))                   &
                                       + bs * ds * (q1r(ivz) - q1l(ivz))) / dv

! calculate the scalar product of velocity and magnetic field
!
              vb2 = sum(q2(ivx:ivz) * q2(ibx:ibz))

! depending on the contact discontinuity speed chose the right Alfvén state
!
              if (sm > 0.0d0) then

! prepare the conservative variables for the left Alfvén intermediate state
!
                u2(idn) = u1l(idn)
                u2(imx) = u2(idn) * q2(ivx)
                u2(imy) = u2(idn) * q2(ivy)
                u2(imz) = u2(idn) * q2(ivz)
                u2(ibx) = q2(ibx)
                u2(iby) = q2(iby)
                u2(ibz) = q2(ibz)
#ifdef GLM
                u2(iph) = ql(iph,i)
#endif /* GLM */

! calculate the energy for the Alfvén intermediate state (eq. 63)
!
                u2(ien) = u1l(ien) - bs * dlsq * (vb1l - vb2)

! calculate the left Alfvén intermediate flux (eq. 65)
!
                f(:,i) = fl(:,i)                                               &
                            + sml * u2(:) - (sml - sl) * u1l(:) - sl * ul(:,i)

              else if (sm < 0.0d0) then

! prepare the conservative variables for the right Alfvén intermediate state
!
                u2(idn) = u1r(idn)
                u2(imx) = u2(idn) * q2(ivx)
                u2(imy) = u2(idn) * q2(ivy)
                u2(imz) = u2(idn) * q2(ivz)
                u2(ibx) = q2(ibx)
                u2(iby) = q2(iby)
                u2(ibz) = q2(ibz)
#ifdef GLM
                u2(iph) = qr(iph,i)
#endif /* GLM */

! calculate the energy for the Alfvén intermediate state (eq. 63)
!
                u2(ien) = u1r(ien) + bs * drsq * (vb1r - vb2)

! calculate the right Alfvén intermediate flux (eq. 65)
!
                f(:,i) = fr(:,i)                                               &
                            + smr * u2(:) - (smr - sr) * u1r(:) - sr * ur(:,i)

              else ! sm = 0

! prepare the conservative variables for the left Alfvén intermediate state
!
                u2(idn) = u1l(idn)
                u2(imx) = u2(idn) * q2(ivx)
                u2(imy) = u2(idn) * q2(ivy)
                u2(imz) = u2(idn) * q2(ivz)
                u2(ibx) = q2(ibx)
                u2(iby) = q2(iby)
                u2(ibz) = q2(ibz)
#ifdef GLM
                u2(iph) = ql(iph,i)
#endif /* GLM */

! calculate the energy for the Alfvén intermediate state (eq. 63)
!
                u2(ien) = u1l(ien) - bs * dlsq * (vb1l - vb2)

! calculate the left Alfvén intermediate flux (eq. 65)
!
                f(:,i) = fl(:,i)                                               &
                            + sml * u2(:) - (sml - sl) * u1l(:) - sl * ul(:,i)

! prepare the conservative variables for the right Alfvén intermediate state
!
                u2(idn) = u1r(idn)
                u2(imx) = u2(idn) * q2(ivx)
                u2(imy) = u2(idn) * q2(ivy)
                u2(imz) = u2(idn) * q2(ivz)
                u2(ibx) = q2(ibx)
                u2(iby) = q2(iby)
                u2(ibz) = q2(ibz)
#ifdef GLM
                u2(iph) = qr(iph,i)
#endif /* GLM */

! calculate the energy for the Alfvén intermediate state (eq. 63)
!
                u2(ien) = u1r(ien) + bs * drsq * (vb1r - vb2)

! calculate the right Alfvén intermediate flux (eq. 65)
!
                f(:,i) = 0.5d0 * (f(:,i) + (fr(:,i)                            &
                          + smr * u2(:) - (smr - sr) * u1r(:) - sr * ur(:,i)))

              end if ! sm = 0

            end if ! sml ≤ 0 and smr ≥ 0

          else ! Bx = 0

! intermediate state for the case of Bx = 0
!
            if (sm > 0.0d0) then

! calculate the left intermediate flux, eq. (64)
!
              f(:,i) = fl(:,i) + sl * (u1l(:) - ul(:,i))

            else if (sm < 0.0d0) then

! calculate the right intermediate flux, eq. (64)
!
              f(:,i) = fr(:,i) + sr * (u1r(:) - ur(:,i))

            else ! sm = 0

! average the left and right fluxes if both Sm = 0, and Bx = 0
!
              f(:,i) = 0.5d0 * ((fl(:,i) + sl * (u1l(:) - ul(:,i)))            &
                             +  (fr(:,i) + sr * (u1r(:) - ur(:,i))))

            end if ! sm = 0

          end if ! Bx = 0

#ifdef SHOCK_DETECTION
        end if ! shock vicinity
#endif /* SHOCK_DETECTION */

      end if ! sl < 0 < sr

    end do

!-------------------------------------------------------------------------------
!
  end subroutine riemann
#endif /* ADI */
#ifdef ISO
!
!===============================================================================
!
! subroutine RIEMANN:
! ------------------
!
!   Subroutine solves one dimensional Riemann problem using the isothermal HLLD
!   method described by Mignone.
!
!   Arguments:
!
!     n - the length of input vectors;
!     h - the spatial step;
!     q - the input array of primitive variables;
!     b - the input vector of the normal magnetic field component;
!     f - the output array of fluxes;
!     s - the input array of shock indicators;
!
!   References:
!
!     [1] Mignone, A.,
!         "A simple and accurate Riemann solver for isothermal MHD",
!         Journal of Computational Physics, 2007, 225, pp. 1427-1441
!
!===============================================================================
!
  subroutine riemann(n, h, q, b, f, s)

! include external procedures and variables
!
    use equations     , only : csnd
    use equations     , only : prim2cons, fluxspeed
    use interpolations, only : reconstruct, fix_positivity
    use variables     , only : nt
    use variables     , only : idn, imx, imy, imz
    use variables     , only : ivx, ivy, ivz
    use variables     , only : ibx, iby, ibz
#ifdef GLM
    use variables     , only : iph
    use variables     , only : cmax
#endif /* GLM */

! local variables are not implicit by default
!
    implicit none

! input / output arguments
!
    integer                 , intent(in)  :: n
    real                    , intent(in)  :: h
    real   , dimension(nt,n), intent(in)  :: q
    real   , dimension(   n), intent(in)  :: b
    real   , dimension(nt,n), intent(out) :: f
    logical, dimension(n)   , intent(in)  :: s

! local variables
!
    integer :: p, i
#ifdef SHOCK_DETECTION
    integer :: ip1
#endif /* SHOCK_DETECTION */
    real    :: dv, dn, mx, sq, fc, bs, b2, ca, dnl, dnr
    real    :: sl, sr, srl, srml, sm, sml, smr, slmv, srmv, slmm, srmm

! local arrays to store the states
!
    real, dimension(nt,n) :: ul, ur, ql, qr, fl, fr
    real, dimension(n)    :: cl, cr
    real, dimension(nt)   :: u1l, u1r, u2
!
!-------------------------------------------------------------------------------
!
#ifdef GLM
! reconstruct the left and right states from primitive variables
!
    do p = 1, nt
      call reconstruct(n, h, q(p,:), ql(p,:), qr(p,:), s(:))
    end do

! obtain the state values for Bx and Psi for the GLM-MHD equations
!
    cl(:) = 0.5d+00 * ((qr(ibx,:) + ql(ibx,:)) - (qr(iph,:) - ql(iph,:)) / cmax)
    cr(:) = 0.5d+00 * ((qr(iph,:) + ql(iph,:)) - (qr(ibx,:) - ql(ibx,:)) * cmax)
    ql(ibx,:) = cl(:)
    qr(ibx,:) = cl(:)
    ql(iph,:) = cr(:)
    qr(iph,:) = cr(:)
#else /* GLM */
! reconstruct the left and right states from primitive variables
!
    do p = 1, ibx - 1
      call reconstruct(n, h, q(p,:), ql(p,:), qr(p,:), s(:))
    end do
    do p = ibx + 1, nt
      call reconstruct(n, h, q(p,:), ql(p,:), qr(p,:), s(:))
    end do

! copy the normal component of magnetic field to the left and right states
!
    ql(ibx,:) = b(:)
    qr(ibx,:) = b(:)
#endif /* GLM */

#ifdef FIX_POSITIVITY
! check if the reconstruction doesn't give negative densities, if so, correct
! the states
!
    call fix_positivity(n, q(idn,:), ql(idn,:), qr(idn,:))
#endif /* FIX_POSITIVITY */

! prepare conservative variables of both states
!
    call prim2cons(n, ql(:,:), ul(:,:))
    call prim2cons(n, qr(:,:), ur(:,:))

! calculate physical fluxes and speeds for the states
!
    call fluxspeed(n, ql(:,:), ul(:,:), fl(:,:), cl(:))
    call fluxspeed(n, qr(:,:), ur(:,:), fr(:,:), cr(:))

! iterate over all points
!
    do i = 1, n

! estimate the minimum and maximum speeds
!
      sl = min(ql(ivx,i) - cl(i), qr(ivx,i) - cr(i))
      sr = max(ql(ivx,i) + cl(i), qr(ivx,i) + cr(i))

! calculate the HLLD flux
!
      if (sl >= 0.0d0) then

        f(:,i) = fl(:,i)

      else if (sr <= 0.0d0) then

        f(:,i) = fr(:,i)

      else ! sl < 0 < sr

! prepare some coefficients
!
        srl  = sr * sl
        srml = sr - sl

#ifdef SHOCK_DETECTION
! calculate the next index
!
        ip1 = min(n, i + 1)

! check if we are in the vicinity of a shock
!
        if (s(i) .or. s(ip1)) then

! calculate the fluxes for the intermediate state
!
          f(:,i) = ((sr * fl(:,i) - sl * fr(:,i))                              &
                                           + srl * (ur(:,i) - ul(:,i))) / srml

        else
#endif /* SHOCK_DETECTION */

! density and normal momentum are constant acros the intermediate states
! (eq. 20 and 21)
!
          dn = ((sr * ur(idn,i) - sl * ul(idn,i))                              &
                                             - (fr(idn,i) - fl(idn,i))) / srml
          mx = ((sr * ur(imx,i) - sl * ul(imx,i))                              &
                                             - (fr(imx,i) - fl(imx,i))) / srml

! calculate corresponding fluxes for density and normal momentum for
! the intermediate states (eq. 22 and 23)
!
          f(idn,i) = ((sr * fl(idn,i) - sl * fr(idn,i))                        &
                                       + srl * (ur(idn,i) - ul(idn,i))) / srml
          f(imx,i) = ((sr * fl(imx,i) - sl * fr(imx,i))                        &
                                       + srl * (ur(imx,i) - ul(imx,i))) / srml
          f(ibx,i) = ((sr * fl(ibx,i) - sl * fr(ibx,i))                        &
                                       + srl * (ur(ibx,i) - ul(ibx,i))) / srml
#ifdef GLM
          f(iph,i) = ((sr * fl(iph,i) - sl * fr(iph,i))                        &
                                       + srl * (ur(iph,i) - ul(iph,i))) / srml
#endif /* GLM */

! the speed of advection in the intermediate states (calculated from eq. 15
! and eq. 17)
!
          sm   = f(idn,i) / dn

! calculate speed differences
!
          slmv = sl - ql(ivx,i)
          srmv = sr - qr(ivx,i)
          slmm = sl - sm
          srmm = sr - sm

! prepare some coefficients
!
          dnl  = ql(idn,i) * slmv
          dnr  = qr(idn,i) * srmv
          sq   = sqrt(dn)
          b2   = ql(ibx,i) * qr(ibx,i)

! calculate Alfvén speed (eq. 29)
!
          ca   = abs(ql(ibx,i)) / sq

! calculate speeds of left and right going Alfvén waves
!
          sml  = sm - ca
          smr  = sm + ca

! calculate the density and normal component of the momentum and magnetic
! field for the left and right intermediate states
!
          u1l(idn) = dn
          u1l(imx) = mx
          u1l(ibx) = ql(ibx,i)
          u1r(idn) = dn
          u1r(imx) = mx
          u1r(ibx) = qr(ibx,i)

! calculate tangential components of the magnetic field for the left and right
! intermediate states (like eq. 32-33, but use expressions as in the adabatic
! case and in the degeneracy takes place, revert tangential components to
! the HLL averages)
!
          dv = dnl * slmm - b2
          if (dv /= 0.0d0 .and. ca <= csnd) then
            fc       = (dnl * slmv - b2) / dv
            u1l(iby) = ul(iby,i) * fc
            u1l(ibz) = ul(ibz,i) * fc
          else
            u1l(iby) = ((sr * ur(iby,i) - sl * ul(iby,i))                      &
                                             - (fr(iby,i) - fl(iby,i))) / srml
            u1l(ibz) = ((sr * ur(ibz,i) - sl * ul(ibz,i))                      &
                                             - (fr(ibz,i) - fl(ibz,i))) / srml
          end if

          dv = dnr * srmm - b2
          if (dv /= 0.0d0 .and. ca <= csnd) then
            fc       = (dnr * srmv - b2) / dv
            u1r(iby) = ur(iby,i) * fc
            u1r(ibz) = ur(ibz,i) * fc
          else
            u1r(iby) = ((sr * ur(iby,i) - sl * ul(iby,i))                      &
                                             - (fr(iby,i) - fl(iby,i))) / srml
            u1r(ibz) = ((sr * ur(ibz,i) - sl * ul(ibz,i))                      &
                                             - (fr(ibz,i) - fl(ibz,i))) / srml
          end if

! calculate tangential components of the momentum for the left and right
! intermediate states (like eqs. 30-31, but using already calculated magnetic
! field components for these states)
!
          u1l(imy) = (ul(imy,i) * slmv                                         &
                                  + ql(ibx,i) * (ul(iby,i) - u1l(iby))) / slmm
          u1l(imz) = (ul(imz,i) * slmv                                         &
                                  + ql(ibx,i) * (ul(ibz,i) - u1l(ibz))) / slmm
          u1r(imy) = (ur(imy,i) * srmv                                         &
                                  + qr(ibx,i) * (ur(iby,i) - u1r(iby))) / srmm
          u1r(imz) = (ur(imz,i) * srmv                                         &
                                  + qr(ibx,i) * (ur(ibz,i) - u1r(ibz))) / srmm

! calculate fluxes for the intermediate states
!
          if (sml > 0.0d0) then

! calculate the left intermediate flux, eq. (38)
!
            f(imy,i) = fl(imy,i) + sl * (u1l(imy) - ul(imy,i))
            f(imz,i) = fl(imz,i) + sl * (u1l(imz) - ul(imz,i))
            f(iby,i) = fl(iby,i) + sl * (u1l(iby) - ul(iby,i))
            f(ibz,i) = fl(ibz,i) + sl * (u1l(ibz) - ul(ibz,i))

          else if (smr < 0.0d0) then

! calculate the right intermediate flux, eq. (38)
!
            f(imy,i) = fr(imy,i) + sr * (u1r(imy) - ur(imy,i))
            f(imz,i) = fr(imz,i) + sr * (u1r(imz) - ur(imz,i))
            f(iby,i) = fr(iby,i) + sr * (u1r(iby) - ur(iby,i))
            f(ibz,i) = fr(ibz,i) + sr * (u1r(ibz) - ur(ibz,i))

          else ! sml ≤ 0 ≤ smr

! calculate the intermediate state (eq. 34-37)
!
            if (ql(ibx,i) > 0.0d0) then

              u2(imy) = 0.5d0 * ((u1l(imy) + u1r(imy))                         &
                                                 + sq * (u1r(iby) - u1l(iby)))
              u2(imz) = 0.5d0 * ((u1l(imz) + u1r(imz))                         &
                                                 + sq * (u1r(ibz) - u1l(ibz)))

              u2(iby) = 0.5d0 * ((u1l(iby) + u1r(iby))                         &
                                                 + (u1r(imy) - u1l(imy)) / sq)
              u2(ibz) = 0.5d0 * ((u1l(ibz) + u1r(ibz))                         &
                                                 + (u1r(imz) - u1l(imz)) / sq)

            else if (ql(ibx,i) < 0.0d0) then

              u2(imy) = 0.5d0 * ((u1l(imy) + u1r(imy))                         &
                                                 - sq * (u1r(iby) - u1l(iby)))
              u2(imz) = 0.5d0 * ((u1l(imz) + u1r(imz))                         &
                                                 - sq * (u1r(ibz) - u1l(ibz)))

              u2(iby) = 0.5d0 * ((u1l(iby) + u1r(iby))                         &
                                                 - (u1r(imy) - u1l(imy)) / sq)
              u2(ibz) = 0.5d0 * ((u1l(ibz) + u1r(ibz))                         &
                                                 - (u1r(imz) - u1l(imz)) / sq)
            else

              u2(imy) = 0.5d0 * (u1l(imy) + u1r(imy))
              u2(imz) = 0.5d0 * (u1l(imz) + u1r(imz))

              u2(iby) = 0.5d0 * (u1l(iby) + u1r(iby))
              u2(ibz) = 0.5d0 * (u1l(ibz) + u1r(ibz))

            end if

! calculate the intermediate flux (eq. 24)
!
            f(imy,i) = sm * u2(imy) - ql(ibx,i) * u2(iby)
            f(imz,i) = sm * u2(imz) - ql(ibx,i) * u2(ibz)
            f(iby,i) = sm * u2(iby) - ql(ibx,i) * u2(imy) / dn
            f(ibz,i) = sm * u2(ibz) - ql(ibx,i) * u2(imz) / dn

          end if ! sml ≤ 0 ≤ smr

#ifdef SHOCK_DETECTION
        end if ! shock vicinity
#endif /* SHOCK_DETECTION */

      end if ! sl < 0 < sr

    end do

!-------------------------------------------------------------------------------
!
  end subroutine riemann
#endif /* ISO */
#endif /* HLLD */
#endif /* MHD */

!===============================================================================
!
end module schemes
